/*
 * Copyright 1991, 1992, 1993, 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

/* 
 *	Kurt Akeley
 *	February 1991
 *	Simple capping demonstration
 *
 *	  When a polygonal surface representation of a solid object is
 *	  clipped, the inside of the "solid" is revealed.  This
 *	  program demonstrates a technique to "cap" the solid, so that
 *	  it appears solid even when clipped.  In order to do this, it
 *	  draws each object twice, once to render the visible surfaces,
 *	  and again counting the number of accesses at each pixel.
 *	  Finally, the clipping planes themselves are drawn, with pixels
 *	  actually modified only where the pixel access count in the
 *	  stencil buffer is odd (i.e. where the inside of a solid was
 *	  revealed).
 *
 *	  When invoked with no arguments, three internally defined models
 *	  are displayed (hit the space bar to toggle between models).
 *	  If an argument is given, it is expected to be the name of a
 *	  flip-format .bin file.  If this file is found in either the
 *	  local directory or in /usr/demos/data/models, it replaces the
 *	  first internally defined model.
 *
 *	  For this code to work correctly, the model must be built of
 *	  proper polygonal surfaces.  Such surfaces separate space into
 *	  inside and outside volumes (perhaps more than one of each).
 *	  They are constructed of polygons whose adjacent edges share
 *	  vertexes and do not overlap.  The martini glass (martini.bin)
 *	  is an example of such a model.  None of the other models
 *	  currently in the models directory is a proper solid.
 *
 *	  It is acceptable for the model to include intersecting proper
 *	  polygon solids, as demonstrated by the third internal model.
 *	  When intersections are viewed normally, the first solid drawn
 *	  'wins'.  In orange-caps mode, however, intersection regions
 *	  are hilighted in yellow, rather than in the standard orange.
 *
 *	6 November 1991
 *	Added interference checking in "orange cap" mode.
 *
 *	  Allow stencil count to increment after each cap is rendered.
 *	  Render in a different color if the stencil value indicates
 *	  double coverage.
 *
 *	8 November 1991
 *	Enabled interference checking at all times, not just when
 *	in orange cap mode.
 */

#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/device.h>
#include "flip.h"
#include "awk.h"
#include "interp.h"

#define INITDEPTH -3.0
#define FOV 450

#define MAXPLANE 6
#define MAXTRANSFORM (MAXPLANE+2)
#define EDIT_VIEW (MAXTRANSFORM-1)
#define EDIT_OBJECT (MAXTRANSFORM-2)
#define MAXMODEL 4

#define STEN_SIZE 4
#define STEN_MASK ((1 << STEN_SIZE) - 1)

int tracking();
int shifted();
void limitrot();
void reshapewindow();
void drawscene();
void initgraphics();
void initparams();
void drawcube(long outside, float s);
void drawcylinder(long outside, float s);
void drawsquare();
void fillsquare();
void setmatrix(long i);
void setinvmat(long i);
void farclip(long b);
void drawobj(void (*func)(), long mat);
void makemenus();
void changetitle();
float *quadmalloc(long quads);
void winrect();

void drawfobj();
void drawiobj();

void bigcube();
void middlecube();
void littlecube();

void bigcylinder();
void middlecylinder();
void littlecylinder();

void longcylinder();
void topcylinder();
void centercylinder();
void bottomcylinder();

long xdim,ydim;
float aspect;

float xr[MAXTRANSFORM] = {0,0,0,0,0,0,0,0};
float yr[MAXTRANSFORM] = {0,0,0,0,0,0,0,0};
float depth[MAXTRANSFORM] = {0,0,0,0,0,0,0,INITDEPTH};

float boink_xr;
float boink_yr;
float boink_depth;

long penable[MAXPLANE] = {0,0,0,0,0,0};
long pface[MAXPLANE] = {0,0,0,0,0,0};

float boink_rot_scale;
float boink_depth_scale;
short boink_screen_x;
short boink_screen_y;

int model;
int docap;
int motion;
int tselect;
int mselect;
int twoside;
int orangecaps;
int blocked = FALSE;
int ctrlkey = FALSE;
int leftshiftkey = FALSE;
int rightshiftkey = FALSE;
int leftmouse = FALSE;
int middlemouse = FALSE;
int menu,movemenu,disablemenu;
flipobj *fobj = 0;
element_t *iobj = 0;

main(int argc, char *argv[]) {
    int i;
    short dev, val, screen_x, screen_y;

    /* check for correct machine support */
    if (getgdesc(GD_CLIPPLANES) == 0) {
	fprintf(stderr,"no user-defined clipping planes available.  abort.\n");
	exit(1);
    }
    if (getgdesc(GD_BITS_STENCIL) < 4) {
	fprintf(stderr,"requires 4 stencil planes.  abort.\n");
	exit(2);
    }

    /* read model */
    interpinit(1,0);
    if (argc > 1) {
	if (awkStrIndex(argv[1],".bin"))
	    fobj = readflipobj(argv[1]);
	else {
	    iobj = getinterpobj(argv[1]);
	}
    }

    /* initialize the graphics */
    winopen("cap");
    initparams();
    initgraphics();

    qdevice(ESCKEY);
    qdevice(REDRAW);
    qdevice(CTRLKEY);
    qdevice(LEFTSHIFTKEY);
    qdevice(RIGHTSHIFTKEY);
    qdevice(CKEY);
    qdevice(DKEY);
    qdevice(LKEY);
    qdevice(MKEY);
    qdevice(OKEY);
    qdevice(RKEY);
    qdevice(VKEY);
    qdevice(ZKEY);
    qdevice(ZEROKEY);
    qdevice(ONEKEY);
    qdevice(TWOKEY);
    qdevice(THREEKEY);
    qdevice(FOURKEY);
    qdevice(FIVEKEY);
    qdevice(SPACEKEY);
    qdevice(LEFTMOUSE);
    qdevice(MIDDLEMOUSE);
    qdevice(RIGHTMOUSE);
    tie(LEFTMOUSE,MOUSEX,MOUSEY);
    tie(MIDDLEMOUSE,MOUSEX,MOUSEY);
    tie(RIGHTMOUSE,MOUSEX,MOUSEY);

    qenter(REDRAW,1);
    qenter(ZEROKEY,1);
    while(TRUE) {
	while (qtest() || blocked) {
	    dev = qread(&val);
		switch(dev) {
		    case ESCKEY:
			if (!val)
			    exit(0);
			break;
		    case REDRAW:
			reshapewindow();
			blocked = FALSE;
			break;
		    case LEFTSHIFTKEY:
			leftshiftkey = val;
			break;
		    case RIGHTSHIFTKEY:
			rightshiftkey = val;
			break;
		    case CTRLKEY:
			ctrlkey = val;
			break;
		    case ZEROKEY:
			if (val) {
			    if (shifted())
				penable[0] = FALSE;
			    else if (!tracking()) {
				penable[0] = TRUE;
				tselect = 0;
			    }
			    blocked = FALSE;
			}
			break;
		    case ONEKEY:
			if (val) {
			    if (shifted())
				penable[1] = FALSE;
			    else if (!tracking()) {
				penable[1] = TRUE;
				tselect = 1;
			    }
			    blocked = FALSE;
			}
			break;
		    case TWOKEY:
			if (val) {
			    if (shifted())
				penable[2] = FALSE;
			    else if (!tracking()) {
				penable[2] = TRUE;
				tselect = 2;
			    }
			    blocked = FALSE;
			}
			break;
		    case THREEKEY:
			if (val) {
			    if (shifted())
				penable[3] = FALSE;
			    else if (!tracking()) {
				penable[3] = TRUE;
				tselect = 3;
			    }
			    blocked = FALSE;
			}
			break;
		    case FOURKEY:
			if (val) {
			    if (shifted())
				penable[4] = FALSE;
			    else if (!tracking()) {
				penable[4] = TRUE;
				tselect = 4;
			    }
			    blocked = FALSE;
			}
			break;
		    case FIVEKEY:
			if (val) {
			    if (shifted())
				penable[5] = FALSE;
			    else if (!tracking()) {
				penable[5] = TRUE;
				tselect = 5;
			    }
			    blocked = FALSE;
			}
			break;
		    case CKEY:
			if (val) {
			    docap = !docap;
			    blocked = FALSE;
			}
			break;
		    case DKEY:
			if (val) {
			    for (i=0; i<=MAXTRANSFORM; i++)
				printf("xr=%g yr=%g depth=%g\n",
				    xr[i],yr[i],depth[i]);
			}
			break;
		    case LKEY:
			if (val) {
			    twoside = !twoside;
			    if (twoside)
				lmbind(LMODEL,2);
			    else
				lmbind(LMODEL,1);
			    blocked = FALSE;
			}
			break;
		    case MKEY:
			if (val) {
			    motion = !motion;
			    mselect = tselect;
			    blocked = FALSE;
			}
			break;
		    case OKEY:
			if (val && !tracking()) {
			    tselect = EDIT_OBJECT;
			    blocked = FALSE;
			    }
			break;
		    case RKEY:
			if (val) {
			    initparams();
			    blocked = FALSE;
			}
			break;
		    case VKEY:
			if (val && !tracking()) {
			    tselect = EDIT_VIEW;
			    blocked = FALSE;
			    }
			break;
		    case ZKEY:
			if (val) {
			    orangecaps = !orangecaps;
			    blocked = FALSE;
			    }
			break;
		    case SPACEKEY:
			if (val) {
			    model += 1;
			    if (model >= MAXMODEL)
				model = 0;
			    blocked = FALSE;
			}
			break;
		    case LEFTMOUSE:
			if (shifted() && ctrlkey && val) {
			    for (i=0; i<MAXTRANSFORM; i++) {
				depth[i] = (i==EDIT_VIEW) ? INITDEPTH : 0;
			    }
			    tselect = EDIT_VIEW;
			    blocked = FALSE;
			}
			else if (val) {
			    leftmouse = TRUE;
			    dev = qread(&boink_screen_x);
			    dev = qread(&boink_screen_y);
			    boink_depth = depth[tselect];
			    boink_depth_scale = 0.002;
			    if (shifted())
				boink_depth_scale /= 10.0;
			    else if (ctrlkey)
				boink_depth_scale /= 100.0;
			    blocked = FALSE;
			}
			else if (leftmouse) {
			    leftmouse = FALSE;
			    dev = qread(&screen_x);
			    dev = qread(&screen_y);
			    depth[tselect] = boink_depth - boink_depth_scale *
				(float)(boink_screen_x - screen_x);
			    blocked = FALSE;
			}
			break;
		    case MIDDLEMOUSE:
			if (shifted() && ctrlkey && val) {
			    for (i=0; i<MAXTRANSFORM; i++) {
				xr[i] = 0.0;
				yr[i] = 0.0;
			    }
			    tselect = EDIT_VIEW;
			    blocked = FALSE;
			}
			else if (val) {
			    middlemouse = TRUE;
			    dev = qread(&boink_screen_x);
			    dev = qread(&boink_screen_y);
			    boink_yr = yr[tselect];
			    boink_xr = xr[tselect];
			    boink_rot_scale = 0.333;
			    if (shifted())
				boink_rot_scale /= 10.0;
			    else if (ctrlkey)
				boink_rot_scale /= 100.0;
			    blocked = FALSE;
			}
			else if (middlemouse) {
			    middlemouse = FALSE;
			    dev = qread(&screen_x);
			    dev = qread(&screen_y);
			    yr[tselect] = boink_yr + boink_rot_scale * 
				(float)(boink_screen_x - screen_x);
			    xr[tselect] = boink_xr + boink_rot_scale *
				(float)(boink_screen_y - screen_y);
			    limitrot();
			    blocked = FALSE;
			}
			break;
		    case RIGHTMOUSE:
			makemenus();
			if (val && !tracking()) {
			    switch(dopup(menu)) {
				case 100:
				    penable[0] = TRUE;
				    tselect = 0;
				    break;
				case 101:
				    penable[1] = TRUE;
				    tselect = 1;
				    break;
				case 102:
				    penable[2] = TRUE;
				    tselect = 2;
				    break;
				case 103:
				    penable[3] = TRUE;
				    tselect = 3;
				    break;
				case 104:
				    penable[4] = TRUE;
				    tselect = 4;
				    break;
				case 105:
				    penable[5] = TRUE;
				    tselect = 5;
				    break;
				case 110:
				    penable[0] = FALSE;
				    break;
				case 111:
				    penable[1] = FALSE;
				    break;
				case 112:
				    penable[2] = FALSE;
				    break;
				case 113:
				    penable[3] = FALSE;
				    break;
				case 114:
				    penable[4] = FALSE;
				    break;
				case 115:
				    penable[5] = FALSE;
				    break;
				case 116:
				    for (i=0; i<MAXPLANE; i++)
					penable[i] = FALSE;
				    break;
				case 200:
				    tselect = EDIT_VIEW;
				    break;
				case 201:
				    tselect = EDIT_OBJECT;
				    break;
				case 202:
				    motion = !motion;
				    mselect = tselect;
				    break;
				case 203:
				    docap = !docap;
				    break;
				case 204:
				    twoside = !twoside;
				    if (twoside)
					lmbind(LMODEL,2);
				    else
					lmbind(LMODEL,1);
				    break;
				case 205:
				    model += 1;
				    if (model >= MAXMODEL)
					model = 0;
				    break;
				case 206:
				    orangecaps = !orangecaps;
				    break;
				case 207:
				    initparams();
				    break;
				case 300:
				    exit(0);
				    break;
			    }
			    blocked = FALSE;
			}
			break;
		}
	    }
	if (leftmouse) {
	    depth[tselect] = boink_depth - boink_depth_scale *
		(float)(boink_screen_x - getvaluator(MOUSEX));
	}
	if (middlemouse) {
	    yr[tselect] = boink_yr + boink_rot_scale * 
		(float)(boink_screen_x - getvaluator(MOUSEX));
	    xr[tselect] = boink_xr + boink_rot_scale *
		(float)(boink_screen_y - getvaluator(MOUSEY));
	    limitrot();
	}
	if (motion && (!middlemouse || (mselect != tselect))) {
	    yr[mselect] += (getvaluator(MOUSEX)-640)/50.0;
	    xr[mselect] -= (getvaluator(MOUSEY)-512)/50.0;
	    limitrot();
	}
	drawscene();
	blocked = !tracking() && !motion;
    }
}

int tracking() {
    return leftmouse || middlemouse;
}

int shifted() {
    return leftshiftkey || rightshiftkey;
}

void limitrot() {
    while (xr[tselect] > 360) xr[tselect] -= 360;
    while (xr[tselect] < 0) xr[tselect] += 360;
    while (yr[tselect] > 360) yr[tselect] -= 360;
    while (yr[tselect] < 0) yr[tselect] += 360;
    if (motion && (mselect != tselect)) {
	while (xr[mselect] > 360) xr[mselect] -= 360;
	while (xr[mselect] < 0) xr[mselect] += 360;
	while (yr[mselect] > 360) yr[mselect] -= 360;
	while (yr[mselect] < 0) yr[mselect] += 360;
    }
}

void reshapewindow() {
    reshapeviewport();
    getsize(&xdim,&ydim);
    aspect = (float)xdim / (float)ydim;
    mmode(MVIEWING); 
    farclip(TRUE);		/* set perspective projection */
}

void farclip(long b) {
    if (b)
	perspective(FOV,aspect,0.1,5.0);
    else
	perspective(FOV,aspect,0.1,10000.0);
}

void setmatrix(long i) {
    /* do viewing transformation */
    translate(0.0,0.0,depth[EDIT_VIEW]);
    rot(-yr[EDIT_VIEW],'y');
    rot(xr[EDIT_VIEW],'x');
    /* do model transformation */
    rot(-yr[i],'y');
    rot(xr[i],'x');
    translate(0.0,0.0,depth[i]);
}

void setinvmat(long i) {
    /* compute inverse of model transformation */
    translate(0.0,0.0,-depth[i]);
    rot(-xr[i],'x');
    rot(yr[i],'y');
    /* compute inverse of viewing transformation */
    rot(-xr[EDIT_VIEW],'x');
    rot(yr[EDIT_VIEW],'y');
    translate(0.0,0.0,-depth[EDIT_VIEW]);
}

float cp[4] = {0.0, 0.0, 1.0, 0.0};

void drawscene() {
    register i;

    /* update title bar */
    changetitle();

    /* clear the screen */
    swritemask(0xff);
    czclear(0,0);

    /* modify the location of the current clipplane */
    for (i=0; i<MAXPLANE; i++) {
	if ((tselect == i) || (tselect == EDIT_VIEW) ||
		(motion && (mselect == i))) {
	    /* determine which way the clipping plane faces */
	    Matrix mat;
	    pushmatrix();
	    setinvmat(i);
	    getmatrix(mat);
	    pface[i] = (mat[3][2] < 0) ? 1 : 0;
	    popmatrix();
	    /* relocate the clipping plane */
	    pushmatrix();
	    setmatrix(i);
	    clipplane(i,CP_DEFINE,cp);
	    popmatrix();
	}
    }

    switch (model) {
    case 0:
	if (fobj)
	    drawobj(drawfobj,2);
	else if (iobj)
	    drawobj(drawiobj,2);
	else {
	    drawobj(bigcube,1);
	    drawobj(middlecube,3);
	    drawobj(littlecube,5);
	}
	break;
    case 1:
	drawobj(bigcylinder,1);
	drawobj(middlecylinder,3);
	drawobj(littlecylinder,5);
	break;
    case 2:
	drawobj(longcylinder,1);
	drawobj(topcylinder,3);
	drawobj(centercylinder,4);
	drawobj(bottomcylinder,5);
	break;
    case 3:
	if (fobj)
	    drawobj(drawfobj,2);
	else if (iobj)
	    drawobj(drawiobj,2);
	else {
	    drawobj(bigcube,1);
	    drawobj(middlecube,3);
	    drawobj(littlecube,5);
	}
	drawobj(bigcylinder,1);
	drawobj(middlecylinder,3);
	drawobj(littlecylinder,5);
	break;
    }

    /* draw intersections and clear stencil buffer */
    for (i=0; i<MAXPLANE; i++)
	clipplane(i,CP_OFF,0);
    stencil(TRUE,2,SF_LESS,STEN_MASK,ST_ZERO,ST_ZERO,ST_ZERO);
    zfunction(ZF_ALWAYS);
    cpack(0x00ffff);
    winrect();
    zfunction(ZF_GEQUAL);
    stencil(FALSE,0,0,0,0,0,0);

    /* disable all clipping planes */
    for (i=0; i<MAXPLANE; i++) {
	clipplane(i,CP_OFF,0);
    }

    /* draw the current clipping plane */
    if ((tselect < MAXPLANE) && penable[tselect]) {
	if (pface[tselect])
	    cpack(0xffffffff);
	else
	    cpack(0xff00ff00);
	pushmatrix();
	setmatrix(tselect);
	drawsquare();
	popmatrix();
    }

    swapbuffers();
}

void drawobj(void (*func)(), long mat) {
    register i;

    /* setup the matrix for drawing the object */
    pushmatrix(); 
    setmatrix(EDIT_OBJECT);

    /* enable selected clipping planes - draw object */
    for (i=0; i<MAXPLANE; i++) {
	if (penable[i])
	    clipplane(i,CP_ON,0);
	else
	    clipplane(i,CP_OFF,0);
    }
    if (docap)
	backface(TRUE);
    lmbind(MATERIAL,mat);
    (*func)();
    lmbind(MATERIAL,0);
    if (docap)
	backface(FALSE);

    /* done if capping is not enabled */
    if (!docap) {
	popmatrix();
	return;
    }

    /* enable only front clipping planes - draw into stencil bitplanes */
    for (i=0; i<MAXPLANE; i++) {
	if (penable[i] && pface[i])
	    clipplane(i,CP_ON,0);
	else
	    clipplane(i,CP_OFF,0);
    }
    farclip(FALSE);
    zbuffer(FALSE);
    wmpack(0x00000000);
    swritemask(0x01);
    stencil(TRUE,0,SF_ALWAYS,0x1,ST_INVERT,ST_INVERT,ST_INVERT);
    (*func)();
    farclip(TRUE);
    zbuffer(TRUE);
    wmpack(0xffffffff);
    swritemask(0xff);
    popmatrix();

    /* draw caps */
    for (i=0; i<MAXPLANE; i++) {
	if (penable[i])
	    clipplane(i,CP_ON,0);
	else
	    clipplane(i,CP_OFF,0);
    }
    zfunction(ZF_ALWAYS);
    stencil(TRUE,1,SF_EQUAL,STEN_MASK,ST_KEEP,ST_KEEP,ST_INCR);
    for (i=0; i<MAXPLANE; i++) {
	if (penable[i] && pface[i]) {
	    clipplane(i,CP_OFF,0);
	    pushmatrix();
	    setmatrix(i);
	    if (orangecaps) {
		cpack(0x7fff);
		fillsquare();
	    } else {
		lmbind(MATERIAL,mat);
		fillsquare();
		lmbind(MATERIAL,0);
	    }
	    popmatrix();
	    clipplane(i,CP_ON,0);
	}
    }

    zfunction(ZF_GEQUAL);
    stencil(FALSE,0,0,0,0,0,0);
}

void drawfobj() {
    drawflipobj(fobj);
}

void drawiobj() {
    drawinterpobj(iobj);
}

void bigcube() {
    drawcube(1, 0.5);
    drawcube(0, 0.45);
}

void middlecube() {
    drawcube(1, 0.3);
    drawcube(0, 0.25);
}

void littlecube() {
    drawcube(1, 0.1);
    drawcube(0, 0.05);
}

void bigcylinder() {
    drawcylinder(1, 0.5);
    drawcylinder(0, 0.45);
}

void middlecylinder() {
    drawcylinder(1, 0.3);
    drawcylinder(0, 0.25);
}

void littlecylinder() {
    drawcylinder(1, 0.1);
    drawcylinder(0, 0.05);
}

void longcylinder() {
    drawcylinder(1, 0.5);
}

void topcylinder() {
    pushmatrix();
    translate(0.0, 0.0, 0.4);
    rot(90.0,'x');
    drawcylinder(1, 0.3);
    popmatrix();
}

void centercylinder() {
    pushmatrix();
    rot(90.0,'x');
    rot(45.0,'y');
    drawcylinder(1, 0.3);
    popmatrix();
}

void bottomcylinder() {
    pushmatrix();
    translate(0.0, 0.0, -0.4);
    rot(90.0,'x');
    rot(90.0,'y');
    drawcylinder(1, 0.3);
    popmatrix();
}

float cubevertex[8][4] = {
    {-1.0, -1.0, -1.0, 1.0},
    { 1.0, -1.0, -1.0, 1.0},
    {-1.0,  1.0, -1.0, 1.0},
    { 1.0,  1.0, -1.0, 1.0},
    {-1.0, -1.0,  1.0, 1.0},
    { 1.0, -1.0,  1.0, 1.0},
    {-1.0,  1.0,  1.0, 1.0},
    { 1.0,  1.0,  1.0, 1.0},
};

float cubenormal[6][4] = {
    { 0.0,  0.0, -1.0},
    { 1.0,  0.0,  0.0},
    { 0.0,  0.0,  1.0},
    {-1.0,  0.0,  0.0},
    { 0.0,  1.0,  0.0},
    { 0.0, -1.0,  0.0},
};

void drawcube(long outside, float s) {
    pushmatrix();
    scale(s,s,s);
    if (outside) {
	bgnpolygon();
	n3f(cubenormal[0]);
	v3f(cubevertex[2]);
	v3f(cubevertex[3]);
	v3f(cubevertex[1]);
	v3f(cubevertex[0]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[1]);
	v3f(cubevertex[3]);
	v3f(cubevertex[7]);
	v3f(cubevertex[5]);
	v3f(cubevertex[1]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[2]);
	v3f(cubevertex[7]);
	v3f(cubevertex[6]);
	v3f(cubevertex[4]);
	v3f(cubevertex[5]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[3]);
	v3f(cubevertex[6]);
	v3f(cubevertex[2]);
	v3f(cubevertex[0]);
	v3f(cubevertex[4]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[4]);
	v3f(cubevertex[6]);
	v3f(cubevertex[7]);
	v3f(cubevertex[3]);
	v3f(cubevertex[2]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[5]);
	v3f(cubevertex[0]);
	v3f(cubevertex[1]);
	v3f(cubevertex[5]);
	v3f(cubevertex[4]);
	endpolygon();
    }
    else {
	bgnpolygon();
	n3f(cubenormal[2]);
	v3f(cubevertex[0]);
	v3f(cubevertex[1]);
	v3f(cubevertex[3]);
	v3f(cubevertex[2]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[3]);
	v3f(cubevertex[1]);
	v3f(cubevertex[5]);
	v3f(cubevertex[7]);
	v3f(cubevertex[3]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[0]);
	v3f(cubevertex[5]);
	v3f(cubevertex[4]);
	v3f(cubevertex[6]);
	v3f(cubevertex[7]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[1]);
	v3f(cubevertex[4]);
	v3f(cubevertex[0]);
	v3f(cubevertex[2]);
	v3f(cubevertex[6]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[5]);
	v3f(cubevertex[2]);
	v3f(cubevertex[3]);
	v3f(cubevertex[7]);
	v3f(cubevertex[6]);
	endpolygon();
	bgnpolygon();
	n3f(cubenormal[4]);
	v3f(cubevertex[4]);
	v3f(cubevertex[5]);
	v3f(cubevertex[1]);
	v3f(cubevertex[0]);
	endpolygon();
    }
    popmatrix();
}

#define SLICES 80
#define LENGTH 1
#define WIDTH 0.5
#define PI (3.141592654)

void drawcylinder(long outside, float s) {
    static init = 0;
    register i;
    float angle;
    static float *vtop,*vbot,*noutside,*ninside;
    static float ntop[3] = {0,0,1};
    static float nbot[3] = {0,0,-1};
    static float vtopcenter[3] = {0,0,LENGTH};
    static float vbotcenter[3] = {0,0,-LENGTH};

    if (!init) {
	init = 1;
	vtop = quadmalloc(SLICES+1);
	vbot = quadmalloc(SLICES+1);
	noutside = quadmalloc(SLICES+1);
	ninside = quadmalloc(SLICES+1);
	for (i=0; i<=SLICES; i++) {
	    angle = ((float)i / (float)SLICES) * PI * 2.0;
	    vtop[4*i+0] = vbot[4*i+0] = fcos(angle) * WIDTH;
	    vtop[4*i+1] = vbot[4*i+1] = fsin(angle) * WIDTH;
	    vtop[4*i+2] = LENGTH;
	    vbot[4*i+2] = -LENGTH;
	    noutside[4*i+0] = fcos(angle);
	    noutside[4*i+1] = fsin(angle);
	    ninside[4*i+0] = -fcos(angle);
	    ninside[4*i+1] = -fsin(angle);
	    noutside[4*i+2] = 0;
	    ninside[4*i+2] = 0;
	}
    }

    pushmatrix();
    scale(s,s,s);

    /* draw top of cylinder */
    bgntmesh();
    n3f(outside ? ntop : nbot);
    v3f(vtopcenter);
    for (i=0; i<=SLICES; i++) {
	v3f(&vtop[outside ? (4*i) : (4*(SLICES-i))]);
	swaptmesh();
    }
    endtmesh();

    /* draw bottom of cylinder */
    bgntmesh();
    n3f(outside ? nbot : ntop);
    v3f(vbotcenter);
    for (i=0; i<=SLICES; i++) {
	v3f(&vbot[outside ? (4*(SLICES-i)) : (4*i)]);
	swaptmesh();
    }
    endtmesh();

    /* draw sides of cylinder */
    for (i=0; i<SLICES; i++) {
	bgnpolygon();
	if (outside) {
	    n3f(&noutside[4*i]);
	    v3f(&vtop[4*i]);
	    v3f(&vbot[4*i]);
	    n3f(&noutside[4*(i+1)]);
	    v3f(&vbot[4*(i+1)]);
	    v3f(&vtop[4*(i+1)]);
	}
	else {
	    n3f(&ninside[4*i]);
	    v3f(&vbot[4*i]);
	    v3f(&vtop[4*i]);
	    n3f(&ninside[4*(i+1)]);
	    v3f(&vtop[4*(i+1)]);
	    v3f(&vbot[4*(i+1)]);
	}
	endpolygon();
    }

    popmatrix();
}

float squarevertex[4][4] = {
    {-1.0, -1.0, 0.0, 1.0},
    { 1.0, -1.0, 0.0, 1.0},
    {-1.0,  1.0, 0.0, 1.0},
    { 1.0,  1.0, 0.0, 1.0},
};

float squarenormal[4] = {0.0, 0.0, -1.0, 0.0};

void drawsquare() {
    blendfunction(BF_SA,BF_MSA);
    bgnclosedline();
    v3f(squarevertex[0]);
    v3f(squarevertex[1]);
    v3f(squarevertex[3]);
    v3f(squarevertex[2]);
    endclosedline();
    blendfunction(BF_ONE,BF_ZERO);
}

void fillsquare() {
    pushmatrix();
    scale(100.0,100.0,100.0);
    bgnpolygon();
    n3f(squarenormal);
    v3f(squarevertex[0]);
    v3f(squarevertex[2]);
    v3f(squarevertex[3]);
    v3f(squarevertex[1]);
    endpolygon();
    popmatrix();
}

float *quadmalloc(long quads) {
    float *f;
    f = (float*)malloc((4*quads+3)*4);
    return (float*)(((long)f + 0xc) & 0xfffffff0);
}

void makemenus() {
    char s[1000];

    sprintf(s,"Options%%t|move view (v)%%x200|move object (o)%%x201|move/enable plane%%m|disable plane%%m|%s motion (m) %%x202|%s capping (c)%%x203|%s lighting (l)%%x204|change model (space) %%x205|%s caps (z) %%x206|reset (r)%%x207|quit (esc)%%x300",motion?"disable":"enable",docap?"disable":"enable",twoside?"oneside":"twoside",orangecaps?"lighted":"orange");

    movemenu = defpup("0%x100|1%x101|2%x102|3%x103|4%x104|5%x105");
    disablemenu = defpup("0%x110|1%x111|2%x112|3%x113|4%x114|5%x115|all%x116");
    menu = defpup(s,movemenu,disablemenu);
}

void changetitle() {
    static char oldtitle[100] = "";
    char title[100];
    if (tselect < MAXPLANE)
	sprintf(title,"cap: move plane %d",tselect);
    else if (tselect == EDIT_OBJECT)
	sprintf(title,"cap: move object");
    else if (tselect == EDIT_VIEW)
	sprintf(title,"cap: move view");
    else
	sprintf(title,"cap:");
    if (strcmp(title,oldtitle)) {
	strcpy(oldtitle,title);
	wintitle(title);
    }
}

float brass[] = {
    AMBIENT, 0.65, 0.5, 0.35,
    DIFFUSE, 0.65, 0.5, 0.35,
    SPECULAR, 0.0, 0.0, 0.0,
    SHININESS, 5.0,
    LMNULL
};

float pewter[] = {
    AMBIENT, 0.6, 0.55, 0.65,
    DIFFUSE, 0.6, 0.55, 0.65,
    SPECULAR, 0.9, 0.9, 0.95,
    SHININESS, 10.0,
    LMNULL
};

float redplastic[] = {
    AMBIENT, 0.5, 0.1, 0.1,
    DIFFUSE, 0.5, 0.1, 0.1,
    SPECULAR, 0.45, 0.45, 0.45,
    SHININESS, 30.0,
    LMNULL
};

float greenplastic[] = {
    AMBIENT, 0.1, 0.5, 0.1,
    DIFFUSE, 0.1, 0.5, 0.1,
    SPECULAR, 0.45, 0.45, 0.45,
    SHININESS, 30.0,
    LMNULL
};

float blueplastic[] = {
    AMBIENT, 0.1, 0.1, 0.5,
    DIFFUSE, 0.1, 0.1, 0.5,
    SPECULAR, 0.45, 0.45, 0.45,
    SHININESS, 30.0,
    LMNULL
};

float whitelight[] = {
    AMBIENT, 0.0, 0.0, 0.0, 
    LCOLOR, 1.0, 1.0, 1.0, 
    POSITION, 0.0, 0.0, 1.0, 0.0,
    LMNULL
};
		    
float infinite[] = {
    AMBIENT, 0.3, 0.3, 0.3, 
    LOCALVIEWER, 0.0, 
    LMNULL
};
		    
float infinite2[] = {
    AMBIENT, 0.3, 0.3, 0.3, 
    LOCALVIEWER, 0.0, 
    TWOSIDE, 1.0,
    LMNULL
};

void initparams() {
    register i;
    for (i=0; i<MAXPLANE; i++)
	penable[i] = FALSE;
    for (i=0; i<MAXTRANSFORM; i++) {
	xr[i] = 0.0;
	yr[i] = 0.0;
	depth[i] = 0.0;
    }
    depth[EDIT_VIEW] = INITDEPTH;
    tselect = EDIT_VIEW;
    mselect = EDIT_VIEW;
    motion = FALSE;
    docap = TRUE;
    twoside = TRUE;
    model = 0;
    orangecaps = 0;
}

void initgraphics() {
    RGBmode();
    doublebuffer();
    stensize(STEN_SIZE);
    gconfig();
    subpixel(TRUE);
    reshapewindow();
    zbuffer(TRUE);
    zfunction(ZF_GEQUAL);
    lsetdepth(0x7fffff,0);
    lmdef(DEFMATERIAL, 1, 0, brass);
    lmdef(DEFMATERIAL, 2, 0, pewter);
    lmdef(DEFMATERIAL, 3, 0, redplastic);
    lmdef(DEFMATERIAL, 4, 0, greenplastic);
    lmdef(DEFMATERIAL, 5, 0, blueplastic);
    lmdef(DEFLIGHT,    1, 0, whitelight);
    lmdef(DEFLMODEL,   1, 0, infinite);
    lmdef(DEFLMODEL,   2, 0, infinite2);
    lmbind(LIGHT1, 1);
    if (twoside)
	lmbind(LMODEL, 2);
    else
	lmbind(LMODEL, 1);
    sclear(0);
    linesmooth(SML_SMOOTHER);
}

float identity[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};

void winrect() {
    /* draw a rectangle the size of the window */
    ortho2(0,1,0,1);
    pushmatrix();
    loadmatrix(identity);
    rectfi(0,0,1,1);
    popmatrix();
    farclip(TRUE);
}
