/*------------------------------------------------------------------------
 * $Id: Devices.c++,v 2.7 1994/02/28 22:36:34 carlson Exp $
 *
 * Methods for the Devices classes.
 *
 * Revision History:
 *	$Log: Devices.c++,v $
// Revision 2.7  1994/02/28  22:36:34  carlson
// Added keeping track of errors with I/O.
// Changed the way that I create the temporary file name because tempnam must
//   make its name in /usr/var/tmp which is frequently on a different file
//   system and won't allow a simple rename.
// Added code to change the mode of the file created to be the same as the
//   original.
//
// Revision 2.6  1994/02/10  22:54:29  chrisc
// Removed package access functions and made them inline.
//
// Revision 2.5  1994/01/18  04:02:59  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.4  1993/11/16  02:29:01  carlson
// Added code to handle modified flag.
// Added method for deleting an entry.  This was kinda done by mistake but
//   I thought I'd leave it, since it existed in the System.c++ file.
// Added methods for checking and clearing modified flag.
//
// Revision 2.3  1993/11/05  19:55:13  carlson
// Changed all references to LinkList to use SmartList.
//
// Revision 2.2  1993/10/06  17:41:50  chris
// Modified all parameters for PathName to use references.
// Added two methods for saving the file.
//
// Revision 2.1  1993/07/27  23:59:52  chris
// Updated to only include class for Devices.
// Made to follow new format, using PathName as input.
//
// Revision 1.3  1993/05/17  06:18:28  chris
// No longer need parameters to instantiate Devices class.  Now we call
//   the method readFile and makeNew to set up internally.
//
// Revision 1.2  1993/05/07  00:22:33  chris
// No longer pass or keep track of Dialers.
//
// Revision 1.1  1993/05/05  19:35:10  chris
// Initial revision
//
 *------------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <errno.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "Devices.h"
#include "Device.h"

extern char *read_line (ifstream *file);
extern char *get_line (ifstream *file);

/*========================================================================
 * Devices class
 *========================================================================*/

/*------------------------------------------------------------------------
 * Devices::Devices (PathName&)
 *------------------------------------------------------------------------*/

Devices::Devices (PathName& filename) 
: LinkData (), SmartList ()
{
    create (filename);
    readFile (filename);
    modified = 0;
}

/*------------------------------------------------------------------------
 * Devices::create (PathName&)
 *------------------------------------------------------------------------*/

void Devices::create (PathName& filename)
{
    if (top)
	clean ();

    setKey ((char *)filename.name ());
}

/*------------------------------------------------------------------------
 * Devices::readFile (PathName&)
 *------------------------------------------------------------------------*/

void Devices::readFile (PathName& filename)
{
    LinkData	*device;
    ifstream	*device_file;
    char	*cptr;

    _errno = 0;
    device_file = new ifstream (filename.name ());
    if (device_file->fail ())
    {
	_errno = errno;
	return;
    }

    while (!device_file->eof ())
    {
	cptr = read_line (device_file);
	if (*cptr)
	{
	    if ((*cptr == ' ') || (*cptr == '\t') || (*cptr == '\n'))
		continue;
	    device = makeNewDevice (cptr);
	    add (device);
	}
    }

    delete device_file;
}

/*------------------------------------------------------------------------
 * Devices::save (PathName& )
 *------------------------------------------------------------------------*/

void Devices::save (PathName& filename)
{
    ifstream		*ifile;
    ofstream		*ofile;
    char		*tmpfname, *cptr, *tptr;
    Device		*devline, *device;
    int			flag;
    struct stat		statBuf;

    _errno = 0;
    if (!wasModified ())
	return;

    tmpfname = new char [(filename.directory ()).length () + 14];
    strcpy (tmpfname, (filename.directory ()).name ());
    strcat (tmpfname, "/DevicesXXXXXX");
    mktemp (tmpfname);
    ofile = new ofstream (tmpfname);
    if (ofile->fail ())
    {
	_errno = errno;
	return;
    }

    ifile = new ifstream (filename.name ());

    /*----
     * Scan through the input file copying comments and blank lines
     * until we come to a legitimate device entry.
     *----*/

    while (!ifile->eof ())
    {
	cptr = get_line (ifile);
	if (ifile->eof () && (*cptr == '\0'))
	    break;

	if ((*cptr == '#') || (*cptr == '\0'))
	    *ofile << cptr << endl;
	else
	{
	    /*----
	     * Create an instance of a Device class using the entry
	     * found so we can use it for comparisons.
	     *----*/

	    devline = new Device (cptr);

	    /*----
	     * Search our list to see if we have an entry with the
	     * same device name.  If we do, put *all* of our entries
	     * with the same device name out here.
	     *
	     * Note:  We use the ListEntry flag to indicate those
	     * entries that we\'ve written.  We assume all the
	     * entries are currently set to 0.  We also assume that
	     * if an entry matches the name but the flag is set,
	     * then all entries we have internally have been written.
	     *----*/

	    device = (Device *)find (devline->theKey ());
	    if (device)
	    {
		if (lastFlag ())
		    continue;
		tptr = device->line ();
		*ofile << tptr << endl;
		setLastFlag (1);
		while ((device = (Device *)next ()) &&
		       (strcmp (device->theKey (), devline->theKey ()) == 0))
		{
		    tptr = device->line ();
		    *ofile << tptr << endl;
		    setLastFlag (1);
		}
	    }

	    delete devline;
	}
    }
    ifile->close ();

    /*----
     * Scan through the list of devices we have and write any that
     * have not been written yet.  While we're scanning, clear all
     * the flags.
     *----*/

    for (device = (Device *)first (); device; device = (Device *)next ())
    {
	flag = lastFlag ();
	setLastFlag (0);
	if (flag)
	    continue;
	tptr = device->line ();
	*ofile << tptr << endl;
    }
    ofile->close ();

    /*----
     * Now we're done.
     * Find out the mode settings of the original file.
     * Make sure we have write access to the file.
     * Rename the new file over the old file.
     * Then set the mode settings of the new file back to what they were.
     *----*/

    if (stat (filename.name (), &statBuf))
	_errno = errno;

    if (!_errno)
	if (chmod (filename.name (),
		   statBuf.st_mode | (S_IWUSR | S_IWGRP | S_IWOTH)))
	    _errno = errno;

    if (!_errno)
	if (rename (tmpfname, filename.name ()))
	    _errno = errno;

    if (!_errno)
    {
	if (chmod (filename.name (), statBuf.st_mode))
	    _errno = errno;
	clearModified ();
    }

    delete [] tmpfname;
}

/*------------------------------------------------------------------------
 * Devices::save (const char *)
 *------------------------------------------------------------------------*/

void Devices::save (const char *fname)
{
    PathName		filename;

    if (fname)
	filename = (char *)fname;
    else
	filename = (char *)theKey ();

    save (filename);
}

/*------------------------------------------------------------------------
 * Devices::makeNewDevice (const char *)
 *------------------------------------------------------------------------*/

LinkData* Devices::makeNewDevice (const char *line)
{
    Device	*device;

    device = new Device (line);

    return device;
}

/*------------------------------------------------------------------------
 * Devices::deleteEntry (LinkData *)
 *------------------------------------------------------------------------*/

void Devices::deleteEntry (LinkData *e)
{
    delete e;
}

/*------------------------------------------------------------------------
 * Devices::wasModified (void)
 *------------------------------------------------------------------------*/

int Devices::wasModified (void)
{
    Device		*d;

    if (modified == 0)
    {
	for (d = (Device *)first (); d; d = (Device *)next ())
	{
	    if (d->wasModified ())
	    {
		modified = 1;
		break;
	    }
	}
    }

    return modified;
}

/*------------------------------------------------------------------------
 * Devices::clearModified (void)
 *------------------------------------------------------------------------*/

void Devices::clearModified (void)
{
    Device		*d;

    for (d = (Device *)first (); d; d = (Device *)next ())
	d->clearModified ();

    modified = 0;
}
