/*------------------------------------------------------------------------
 * $Id: DevicesWindow.c++,v 2.18 1994/03/03 00:28:12 carlson Exp $
 *
 * Methods for DevicesWindow class.
 *
 * Revision History:
 *	$Log: DevicesWindow.c++,v $
// Revision 2.18  1994/03/03  00:28:12  carlson
// Modified so that it will compile on 5.1 OS versions.  This required taking
//   out the parameters for postAndWait methods that caused dialog window
//   to appear over the parent widget.
//
// Revision 2.17  1994/02/28  22:30:07  carlson
// When doing a save, check if an error occurred and generate a dialog if it
//   does.
//
// Revision 2.16  1994/02/26  03:58:33  carlson
// Modified all dialogs to place dialog window over the current parent widget.
// Removed code to add new device to device list since this is done
//   automatically in the callback for when the list of devices is changed.
// Added code to method that handles a change in the list of devices to select
//   the current device, if there is a current device selected.
//
// Revision 2.15  1994/02/24  21:30:33  carlson
// Fixed a bug where it wouldn't find selected devices because the device was
//   set for "Any" baud.
//
// Revision 2.14  1994/02/18  18:41:23  chrisc
// Added method 'getSelected' to do the work of finding the entry in the
//   list for the selected item.
// Cleaned up useless code in selectCallback.
// Improved the way select callback functions.
// Added delete button and delete callback to delete a file.
// Moved all the loops that fill the package list and selection list to
//   be in a method together.  This way, they can be updated by callbacks.
// Added callback for when the items in the list change.
//
// Revision 2.13  1994/01/20  20:17:40  carlson
// Modified to support new kind of help facility.
//
// Revision 2.12  1994/01/18  00:38:14  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.11  1994/01/04  20:47:15  carlson
// Changed some error printouts to use an error dialog.
// Instead of having a <new> item in list, created a New button.
// Fixed all \( \) to just ( ).
//
// Revision 2.10  1993/11/16  02:37:35  carlson
// Added code to support a default line when creating a new device.  This
//   was necessary to avoid a situation where it would core dump.
//
// Revision 2.9  1993/11/15  22:03:29  carlson
// Whoops.  Deleted a line that shouldn't have been deleted.
//
// Revision 2.8  1993/11/15  21:35:30  carlson
// Keep the current device as well as the current device window in our
//   structure so it is available to us.
// Set up callbacks if a device changes so we can update the device list.
// Update device list if new device is created.
// Print 'Any' when any baud is OK in the device list.
// When a package is added, add it to the devices also.
//
// Revision 2.7  1993/11/05  19:55:12  carlson
// Changed all references to LinkList to use SmartList.
//
// Revision 2.6  1993/11/05  00:41:07  carlson
// Removed some of the error print statements and added calls to
//   theErrorDialog ViewKit widget.
// Modified to allow user to create a new device.
// Added help for buttons and labels.
//
// Revision 2.5  1993/10/14  23:32:31  carlson
// Added debug function 'dump' which dumps out everything about an instance
//   of a class to stdout.
//
// Revision 2.4  1993/10/06  17:39:16  chris
// Added 'save' method.
//
// Revision 2.3  1993/09/28  13:25:51  chris
// Changed the name of the encapsulating widget to hostWindow, instead of
//   just host.
// Removed all the definitions for the labels titles so that it is done in
//   a resource file now.  Also, removed the extraneous sizing info.
//
// Revision 2.2  1993/08/11  00:10:28  chris
// Display all information in the device list that makes a device unique
//   so that each unique device can be selected individually.
// When selecting a device to display, verify unique info to select it.
//
// Revision 2.1  1993/07/29  01:25:55  chris
// Initial revision.
//
 *------------------------------------------------------------------------*/

#ifndef OSVERS
#define OSVERS	5100
#endif

#include "DevicesWindow.h"
#include "Devices.h"
#include "DeviceWindow.h"
#include "HelpDialog.h"

#include <Vk/VkPromptDialog.h>
#include <Vk/VkErrorDialog.h>
#include <Vk/VkQuestionDialog.h>

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>

#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/SeparatoG.h>

static char	*default_line = " - - 9600";

/*------------------------------------------------------------------------
 * DevicesWindow::getSelected (char *)
 *------------------------------------------------------------------------*/

Device *DevicesWindow::getSelected (char *text)
{
    Device		*aDev;
    int			found = 0, b;
    char		*newtext, *data, *dialer, *bc;

    /*----
     * Figure out which device matches the one they selected.
     * Note that we have to check the device name, device, dialer and
     * baud rate for a match.
     *----*/

    newtext = new char[strlen (text) + 1];
    strcpy (newtext, text);

    data = strchr (newtext, ' ');
    if (data)
    {
	*data++ = '\0';

	dialer = strchr (data, ' ');
	if (dialer)
	{
	    *dialer++ = '\0';

	    bc = strchr (dialer, ' ');
	    if (bc)
	    {
		*bc++ = '\0';
		b = (strcmp (bc, "Any")) ? atoi (bc) : ANYBAUD;
	    }
	}
    }

    for (aDev = (Device *)devices->find (newtext);
	 aDev && (strcmp (newtext, aDev->theKey ()) == 0);
	 aDev = (Device *)devices->next ())
    {
	if ((strcmp (data, aDev->data_port ()) == 0) &&
	    (strcmp (dialer, aDev->dialer_port ()) == 0) &&
	    (b == aDev->baud ()))
	{
	    found = 1;
	    break;
	}
    }

    delete [] newtext;

    if (!found)
	aDev = NULL;

    return aDev;
}

/*------------------------------------------------------------------------
 * DevicesWindow::selectCallback (Widget, XtPointer, XtPointer)
 *------------------------------------------------------------------------*/

void DevicesWindow::selectCallback (Widget, XtPointer clientData,
				   XtPointer callData)
{
    XmListCallbackStruct*	cbStruct;
    DevicesWindow*		sw;
    char			*text;

    cbStruct = (XmListCallbackStruct *)callData;
    sw = (DevicesWindow *)clientData;

    XmStringGetLtoR (cbStruct->item, XmFONTLIST_DEFAULT_TAG, &text);

    sw->select_cb (text);
}

/*------------------------------------------------------------------------
 * DevicesWindow::select_cb (char *)
 *------------------------------------------------------------------------*/

void DevicesWindow::select_cb (char *text)
{
    char			line[80];
    Device			*newDev;

    newDev = getSelected (text);
    if (!newDev)
    {
	strcpy (line, "Unable to find device ");
	strcat (line, text);
	strcat (line, ".\nReport this to SGI.");
	theErrorDialog->postAndWait (line
#if OSVERS >= 5200
				     , True, False, False, NULL, NULL, _parent
#endif
				     );
	return;
    }

    /*----
     * If a device is already up, bring it down.
     *----*/

    if (current)
    {
	c_dev->removeChangeCallback (devChangedCallback, this);
	delete current;
    }

    c_dev = newDev;
    c_dev->addChangeCallback (devChangedCallback, this);
    current = new DeviceWindow (parent_for_device, c_dev, dlrList);
    current->show ();
}

/*------------------------------------------------------------------------
 * DevicesWindow::newCallback (Widget, XtPointer, XtPointer)
 *------------------------------------------------------------------------*/

void DevicesWindow::newCallback (Widget, XtPointer clientData, XtPointer)
{
    DevicesWindow*		dw;

    dw = (DevicesWindow *)clientData;
    dw->new_cb ();
}

/*------------------------------------------------------------------------
 * DevicesWindow::new_cb ()
 *------------------------------------------------------------------------*/

void DevicesWindow::new_cb ()
{
    char				dev_string[80];

    if (thePromptDialog->postAndWait ("Enter name of device to add:"
#if OSVERS >= 5200
				      , True, True, False, NULL, NULL, _parent
#endif
				      ) == VkPromptDialog::OK)
    {
	if (current)
	{
	    c_dev->removeChangeCallback (devChangedCallback, this);
	    delete current;
	}
    
	strcpy (dev_string, thePromptDialog->text ());
	strcat (dev_string, default_line);
	c_dev = new Device (dev_string);
	c_dev->addChangeCallback (devChangedCallback, this);
	current = new DeviceWindow (parent_for_device, c_dev, dlrList);
	devices->add (c_dev);		// Must be after current defined.
	current->show ();
    }
}

/*------------------------------------------------------------------------
 * DevicesWindow::deleteCallback (Widget, XtPointer, XtPointer)
 *------------------------------------------------------------------------*/

void DevicesWindow::deleteCallback (Widget, XtPointer clientData, XtPointer)
{
    DevicesWindow*		dw;

    dw = (DevicesWindow *)clientData;
    dw->delete_cb ();
}

/*------------------------------------------------------------------------
 * DevicesWindow::delete_cb ()
 *------------------------------------------------------------------------*/

void DevicesWindow::delete_cb ()
{
    int			*position_list, position_count;
    XmString		*xmstr;
    char		*text, line[80];
    Device		*aDev;

    if (!XmListGetSelectedPos (device_list, &position_list, &position_count))
    {
	theErrorDialog->postAndWait ("You need to select the\ndevice "
				     "to delete."
#if OSVERS >= 5200
				     , True, False, False, NULL, NULL, _parent
#endif
				     );
	return;
    }
    XFree (position_list);

    XtVaGetValues (device_list, XmNselectedItems, &xmstr, NULL);
    XmStringGetLtoR (xmstr[0], XmFONTLIST_DEFAULT_TAG, &text);
    aDev = getSelected (text);

    if (!aDev)
    {
	strcpy (line, "Unable to find device ");
	strcat (line, text);
	strcat (line, ".\nReport this to SGI.");
	theErrorDialog->postAndWait (line
#if OSVERS >= 5200
				     , True, False, False, NULL, NULL, _parent
#endif
				     );
	return;
    }

    strcpy (line, "Are you sure you want\nto remove device ");
    strcat (line, text);
    strcat (line, "?");
    if (theQuestionDialog->postAndWait (line
#if OSVERS >= 5200
					, True, True, False, NULL, NULL,
					_parent
#endif
					) == VkDialogManager::OK)
    {
	if (aDev == c_dev)
	{
	    delete current;
	    current = NULL;
	}

	/*----
	 * Removing the entry will automatically remove it from
	 * the list window because of callbacks we have set
	 * up.
	 *----*/

	devices->removeEntry (aDev);
    }
}

/*------------------------------------------------------------------------
 * DevicesWindow::~DevicesWindow ()
 *------------------------------------------------------------------------*/

DevicesWindow::~DevicesWindow ()
{
    if (current)
    {
	c_dev->removeChangeCallback (devChangedCallback, this);
	delete current;
    }
    devices->removeChangeCallback (listChangedCallback, this);
}

/*------------------------------------------------------------------------
 * DevicesWindow::DevicesWindow (PathName)
 *------------------------------------------------------------------------*/

DevicesWindow::DevicesWindow (Widget parent, Widget device_parent,
			      Devices *devs, SmartList *dlrs)
: VkComponent (devs->theKey ())
{
    Arg			wargs[10];
    int			n;
    XmString		xmstr;
    Widget		temp, label, pkg_list_label, device_list_label;
    Widget		rc1, form1, form2;

    _parent = parent;
    devices = devs;
    dlrList = dlrs;
    parent_for_device = device_parent;
    current = NULL;

    /*----
     * Create a form into which to put all the stuff.  We have to
     * specify its size, for some reason, or it never gets set.
     *----*/
    
    n = 0;
    XtSetArg (wargs[n], XmNwidth, 425); n++;
    XtSetArg (wargs[n], XmNheight, 160); n++;
    XtSetArg (wargs[n], XmNallowOverlap, False); n++;
    _baseWidget = XmCreateForm (parent, (char *)devices->theKey (), wargs, n);
    installDestroyHandler ();

    /*----
     * Create a label to indicate what Devices we are showing here.
     *----*/
    
    n = 0;
    xmstr = XmStringCreateSimple ((char *)devices->theKey ());
    XtSetArg (wargs[n], XmNlabelString, xmstr); n++;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    label = XmCreateLabel (_baseWidget, "label", wargs, n);
    XtManageChild (label);
    XmStringFree (xmstr);

    XtAddCallback (label, XmNhelpCallback, aHelpCallback, "devw_help");

    /*----
     * Because of display problems, I was hoping that I could get
     * it to look better by creating a row/column to contain the
     * rest of the information in.
     *----*/

    n = 0;
    XtSetArg (wargs[n], XmNorientation, XmHORIZONTAL); n++;
    XtSetArg (wargs[n], XmNpacking, XmPACK_TIGHT); n++;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNtopWidget, label); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    rc1 = XmCreateRowColumn (_baseWidget, "lists", wargs, n);
    XtManageChild (rc1);

    n = 0;
    form1 = XmCreateForm (rc1, "plForm", wargs, n);
    XtManageChild (form1);

    /*----
     * List packages that use this file.
     *----*/
    
    n = 0;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    pkg_list_label = XmCreateLabel (form1, "pkgListLabel", wargs, n);
    XtManageChild (pkg_list_label);
    
    XtAddCallback (pkg_list_label, XmNhelpCallback, aHelpCallback,
		   "devw_help");

    n = 0;
    XtSetArg (wargs[n], XmNadjustLast, False); n++;
    XtSetArg (wargs[n], XmNpacking, XmPACK_COLUMN); n++;
    XtSetArg (wargs[n], XmNorientation, XmHORIZONTAL); n++;
    XtSetArg (wargs[n], XmNnumColumns, 4); n++;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNtopWidget, pkg_list_label); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNentryAlignment, XmALIGNMENT_CENTER); n++;
    XtSetArg (wargs[n], XmNborderWidth, 1); n++;
    pkg_list = XmCreateRowColumn (form1, "pkgList", wargs, n);
    XtManageChild (pkg_list);
    
    XtAddCallback (pkg_list, XmNhelpCallback, aHelpCallback, "devw_help");

    /*----
     * Put a separator between the two areas.
     *----*/

    n = 0;
    XtSetArg (wargs[n], XmNorientation, XmVERTICAL); n++;
    temp = XmCreateSeparatorGadget (rc1, "separator", wargs, n);
    XtManageChild (temp);

    n = 0;
    form2 = XmCreateForm (rc1, "dlForm", wargs, n);
    XtManageChild (form2);

    /*----
     * Provide a list of devices to select from.
     *----*/
    
    n = 0;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    device_list_label = XmCreateLabel (form2, "deviceListLabel", wargs, n);
    XtManageChild (device_list_label);

    XtAddCallback (device_list_label, XmNhelpCallback, aHelpCallback,
		   "devw_help");

    n = 0;
    XtSetArg (wargs[n], XmNlistSizePolicy, XmVARIABLE); n++;
    XtSetArg (wargs[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
    XtSetArg (wargs[n], XmNvisibleItemCount, 4); n++;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNtopWidget, device_list_label); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    device_list = XmCreateScrolledList (form2, "deviceList", wargs, n);
    XtManageChild (device_list);

    updateLists ();

#if 0					// Used for single click
    XtAddCallback (device_list, XmNsingleSelectionCallback, selectCallback,
		   this);
#else					// Used for double click
    XtAddCallback (device_list, XmNdefaultActionCallback, selectCallback,
		   this);
#endif
    XtAddCallback (device_list, XmNhelpCallback, aHelpCallback, "devw_help");
    devices->addChangeCallback (listChangedCallback, this);

    n = 0;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNtopWidget, device_list); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    temp = XmCreatePushButton (form2, "new", wargs, n);
    XtManageChild (temp);

    XtAddCallback (temp, XmNactivateCallback, newCallback, this);
    XtAddCallback (temp, XmNhelpCallback, aHelpCallback, "devw_new");

    n = 0;
    XtSetArg (wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNtopWidget, device_list); n++;
    XtSetArg (wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (wargs[n], XmNleftWidget, temp); n++;
    XtSetArg (wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg (wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    temp = XmCreatePushButton (form2, "delete", wargs, n);
    XtManageChild (temp);

    XtAddCallback (temp, XmNactivateCallback, deleteCallback, this);
    XtAddCallback (temp, XmNhelpCallback, aHelpCallback, "devw_delete");
}

/*------------------------------------------------------------------------
 * DevicesWindow::addPackage (LinkData *)
 *------------------------------------------------------------------------*/

void DevicesWindow::addPackage (LinkData *le)
{
    Arg		wargs[5];
    int		n;
    XmString	xmstr;
    Widget	temp;

    if (_baseWidget)
    {
	n = 0;
	xmstr = XmStringCreateSimple ((char *)le->theKey ());
	XtSetArg (wargs[n], XmNlabelString, xmstr); n++;
	XtSetArg (wargs[n], XmNalignment, XmALIGNMENT_CENTER); n++;
	temp = XmCreateLabel (pkg_list, (char *)le->theKey (), wargs, n);
	XtManageChild (temp);
	XmStringFree (xmstr);
    }

    devices->addPackage (le);
}

#ifdef DEBUG
/*------------------------------------------------------------------------
 * DevicesWindow::dump (char *)
 *------------------------------------------------------------------------*/

void DevicesWindow::dump (char *indent)
{
    char		*nextIndent;
    int			i;

    i = strlen (indent);
    nextIndent = new char [i + 5];
    strcpy (nextIndent, indent);
    strcat (nextIndent, "    ");

    cout << indent << "DevicesWindow dump (" << this << "):" << endl;
    cout << indent << "  devices = " << devices << endl;
    cout << indent << "  dlrList = " << dlrList << endl;
    cout << indent << "  current = " << current << endl;

    if (current)
	current->dump (nextIndent);
}
#endif

/*------------------------------------------------------------------------
 * DevicesWindow::save (void)
 *------------------------------------------------------------------------*/

void DevicesWindow::save (void)
{
    char		line[128];

    devices->save ();
    if (devices->error ())
    {
	strcpy (line, "*** Error saving file:\n");
	strcat (line, strerror (devices->error ()));
	theErrorDialog->postAndWait (line);
    }
}

/*------------------------------------------------------------------------
 * DevicesWindow::devChangedCallback (Device *, Device::Field, void *)
 *------------------------------------------------------------------------*/

void DevicesWindow::devChangedCallback (Device *dev, Device::Field field,
					void *client_data)
{
    DevicesWindow		*dw;

    dw = (DevicesWindow *)client_data;
    dw->devChanged (dev, field);
}

/*------------------------------------------------------------------------
 * DevicesWindow::devChanged (Device *, Device::Field)
 *------------------------------------------------------------------------*/

void DevicesWindow::devChanged (Device *dev, Device::Field)
{
    int				*pos_list, pos;
    char			dev_string[80];
    XmString			xmstr;

    if (XmListGetSelectedPos (device_list, &pos_list, &pos))
    {
	if (dev->baud () == ANYBAUD)
	    sprintf (dev_string, "%s %s %s Any", dev->theKey (),
		     dev->data_port (), dev->dialer_port ());
	else
	    sprintf (dev_string, "%s %s %s %d", dev->theKey (),
		     dev->data_port (), dev->dialer_port (), dev->baud ());
	xmstr = XmStringCreateSimple (dev_string);
	XmListReplaceItemsPos (device_list, &xmstr, 1, pos_list[0]);
	XmStringFree (xmstr);
	XmListSelectPos (device_list, pos_list[0], False);
    }
}

/*------------------------------------------------------------------------
 * DevicesWindow::listChangedCallback (SmartList *, SmartList::ACTION, void *)
 *------------------------------------------------------------------------*/

void DevicesWindow::listChangedCallback (SmartList *, SmartList::ACTION,
					void *client_data)
{
    DevicesWindow		*dw;

    dw = (DevicesWindow *)client_data;
    dw->updateLists ();
}

/*------------------------------------------------------------------------
 * DevicesWindow::updateLists (void)
 *------------------------------------------------------------------------*/

void DevicesWindow::updateLists (void)
{
    Device		*aDev;
    char		dev_string[80];
    XmString		xmstr;
    WidgetList		wPackages;
    int			nPackages, i, n;
    LinkData		*ld;
    Widget		temp;
    Arg			wargs[10];

    /*----
     * Update the package list.
     *	1. Find out what children exist in the package list.
     *	2. Loop through the packages and add any that are missing.
     *----*/

    XtVaGetValues (pkg_list,
		   XmNchildren, &wPackages,
		   XmNnumChildren, &nPackages, NULL);

    for (ld = devices->firstPackage (); ld; ld = devices->nextPackage ())
    {
	for (i = 0; i < nPackages; i++)
	    if (strcmp (ld->theKey (), XtName (wPackages[i])) == 0)
		break;
	if (i != nPackages)
	    continue;
	n = 0;
	XtSetArg (wargs[n], XmNalignment, XmALIGNMENT_CENTER); n++;
	temp = XmCreateLabel (pkg_list, (char *)ld->theKey (), wargs, n);
	XtManageChild (temp);

	XtAddCallback (temp, XmNhelpCallback, aHelpCallback, "devw_help");
    }

    /*----
     * Now loop through the children and remove any that are
     * not in the package list.
     *----*/

    for (i = 0; i < nPackages; i++)
    {
	ld = devices->findPackage (XtName (wPackages[i]));
	if (!ld)
	    XtDestroyWidget (wPackages[i]);
    }

    /*----
     * Update the device list.
     *	1. Empty it out.
     *	2. Fill it.
     *----*/

    XmListDeleteAllItems (device_list);

    for (aDev = (Device *)devices->first (); aDev;
	 aDev = (Device *)devices->next ())
    {
	if (aDev->baud () == ANYBAUD)
	    sprintf (dev_string, "%s %s %s Any", aDev->theKey (),
		     aDev->data_port (), aDev->dialer_port ());
	else
	    sprintf (dev_string, "%s %s %s %d", aDev->theKey (),
		     aDev->data_port (), aDev->dialer_port (), aDev->baud ());
	xmstr = XmStringCreateSimple (dev_string);
	XmListAddItemUnselected (device_list, xmstr, 0);
	XmStringFree (xmstr);
    }

    if (current)
    {
	if (c_dev->baud () == ANYBAUD)
	    sprintf (dev_string, "%s %s %s Any", c_dev->theKey (),
		     c_dev->data_port (), c_dev->dialer_port ());
	else
	    sprintf (dev_string, "%s %s %s %d", c_dev->theKey (),
		     c_dev->data_port (), c_dev->dialer_port (),
		     c_dev->baud ());
	xmstr = XmStringCreateSimple (dev_string);
	XmListSelectItem (device_list, xmstr, False);
	XmStringFree (xmstr);
    }
}
