/*------------------------------------------------------------------------
 * $Id: Dialer.c++,v 2.6 1993/11/16 17:56:46 carlson Exp $
 *
 * Methods for handling dialer class.
 *
 * Revision History:
 *	$Log: Dialer.c++,v $
// Revision 2.6  1993/11/16  17:56:46  carlson
// Include systemfiles.h and we get to remove prototype for gettoken.
// Replaced COPY macro with better one used in Host and Device.
// Added support for modified flag.
//
// Revision 2.5  1993/10/22  00:25:55  carlson
// Added removeChangeCallback method.
//
// Revision 2.4  1993/10/18  19:58:45  carlson
// Added debug function 'dump' which dumps out everything about an instance
//   of a class to stdout.
//
// Revision 2.3  1993/10/06  18:49:36  chris
// Added method for generating a line to be put back into the Dialers file.
// Modified the way of determining the pause and w4carrier characters
//   because of a bug that I found in the Dialers file.  This fix should
//   not be permanent.
//
// Revision 2.2  1993/08/11  16:20:43  chris
// Added ability to generate callbacks for when elements are changed.
// Renamed the wait character variables to be more descriptive.
// Added destructor.
// Added methods for changing the elements (which then calls the callbacks).
// Modified the method 'script' to be 'expandChatScript' which is more
//   descriptive of what it does.  Also, it returns either the internal
//   chat script expanded or the script passed to it.
//
// Revision 2.1  1993/07/28  16:20:24  chris
// Initial revision.
// Broken out from Dialers.c++.
//
 *------------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <iostream.h>

#include "Dialer.h"
#include "systemfiles.h"

#define COPY(from,to)	(to) = Copy (from)
#define DELETE(field)	if (field) delete [] field

/*========================================================================
 * Dialer class
 *========================================================================*/

/*------------------------------------------------------------------------
 * Dialer
 *------------------------------------------------------------------------*/

Dialer::Dialer (const char *line)
{
    char	*cptr, *tptr;
    char	*newline, *nexttoken;

    callbacks = NULL;
    modified = 0;

    /*----
     * Make a duplicate of the line so we can edit it without screwing
     * up the original.
     *----*/

    newline = new char [strlen (line) + 1];
    nexttoken = newline;
    strcpy (newline, line);

    /*----
     * First token is the dialer name.
     *----*/

    cptr = gettoken (&nexttoken, " \t\n");
    setKey (cptr);

    /*----
     * Second token is the translation characters.  If they aren\'t
     * provided, use default.
     *----*/

    tptr = cptr = gettoken (&nexttoken, "\n");
    if ((cptr == NULL) || ((*cptr != '=') && (*cptr != '-')))
	tptr = "==--";
    else
	cptr += 4;

    w4carrier = pause = '\0';
    while (*tptr && !isspace (*tptr))
    {
	if (*tptr == '=')
	{
	    if (w4carrier)
		pause = *(++tptr);
	    else
		w4carrier = *(++tptr);
	}
	else if (*tptr == '-')
	{
	    if (pause)
		w4carrier = pause;
	    pause = *(++tptr);
	}
	else
	{
	    cerr << "Dialer: Unrecognized translation character in:\n";
	    cerr << "  " << line << "\n";
	}
	tptr++;
    }

    /*----
     * Third token is the chat script to use.
     *----*/

    if (cptr)
	while (*cptr && isspace (*cptr))
	    cptr++;

    if (cptr && *cptr)
	COPY (cptr, chat);
    else
	chat = NULL;

    delete [] newline;
}

/*------------------------------------------------------------------------
 * Dialer::~Dialer \(\)
 *------------------------------------------------------------------------*/

Dialer::~Dialer ()
{
    DELETE (chat);
}

/*------------------------------------------------------------------------
 * Dialer::line \(void\)
 *------------------------------------------------------------------------*/

char *Dialer::line (void)
{
    static char		buffer[256];

    if (chat)
	sprintf (buffer, "%s =%c-%c %s", theKey (), w4carrier, pause, chat);
    else
	sprintf (buffer, "%s =%c-%c", theKey (), w4carrier, pause);
    return buffer;
}

/*------------------------------------------------------------------------
 * Dialer::setWait4CarrierChar \(char\)
 *------------------------------------------------------------------------*/

void Dialer::setWait4CarrierChar (char nw4)
{
    callbackStruct	*cbs;

    w4carrier = nw4;
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, W4CARRIERCHAR,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Dialer::setPauseChar \(char\)
 *------------------------------------------------------------------------*/

void Dialer::setPauseChar (char np)
{
    callbackStruct	*cbs;

    pause = np;
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, PAUSECHAR,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Dialer::setChatScript \(char *\)
 *------------------------------------------------------------------------*/

void Dialer::setChatScript (char *nc)
{
    callbackStruct	*cbs;

    if (chat)
	DELETE (chat);

    COPY (nc, chat);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, CHATSCRIPT,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Dialer::addChangeCallback \(callbackFunction *, void *\)
 *------------------------------------------------------------------------*/

void Dialer::addChangeCallback (callbackFunction cbf, void *cd)
{
    callbackStruct	*cbs, *cbstemp;

    cbs = new callbackStruct;
    cbs->cbf = cbf;
    cbs->clientData = cd;
    cbs->next = NULL;

    if (callbacks)
    {
	for (cbstemp = callbacks; cbstemp->next; cbstemp = cbstemp->next)
	    ;
	cbstemp->next = cbs;
    }
    else
	callbacks = cbs;
}

/*------------------------------------------------------------------------
 * Dialer::removeChangeCallback \(callbackFunction, void *\)
 *------------------------------------------------------------------------*/

void Dialer::removeChangeCallback (callbackFunction cbf, void *cd)
{
    callbackStruct	*cbs, *cbstemp;

    if (callbacks)
    {
	for (cbs = NULL, cbstemp = callbacks; cbstemp; cbstemp = cbstemp->next)
	{
	    if ((cbstemp->cbf == cbf) && (cbstemp->clientData == cd))
	    {
		if (cbs)
		    cbs->next = cbstemp->next;
		else
		    callbacks = cbstemp->next;
		delete cbstemp;
		break;
	    }
	    cbs = cbstemp;
	}
    }
}

/*------------------------------------------------------------------------
 * Dialer::expandChatScript \(char *\)
 *------------------------------------------------------------------------*/

char *Dialer::expandChatScript (char *script)
{
    static char		buff[256];
    char		*c, *b;

    buff[0] = '\0';
    c = (script) ? script : chat;

    if (c)
    {
	for (b = buff; (*b = *c); b++, c++)
	{
	    if ((*c == '=') && (*(c-1) != '\\'))
		*b = w4carrier;
	    else if ((*c == '-') && (*(c-1) != '\\'))
		*b = pause;
	}
    }

    return buff;
}

#ifdef DEBUG
/*------------------------------------------------------------------------
 * Dialer::dump \(char *\)
 *------------------------------------------------------------------------*/

void Dialer::dump (char *indent)
{
    char		*nextIndent;
    int			i;
    callbackStruct	*cb;

    i = strlen (indent);
    nextIndent = new char [i + 5];
    strcpy (nextIndent, indent);
    strcat (nextIndent, "    ");

    cout << indent << "Dialer dump (" << this << "):" << endl;
    cout << indent << "  dialer    = " << theKey () << endl;
    cout << indent << "  w4carrier = " << w4carrier << endl;
    cout << indent << "  pause     = " << pause << endl;
    cout << indent << "  chat      = " << chat << endl;
    cout << endl;
    cout << indent << "  Callbacks:" << endl;
    for (cb = callbacks; cb; cb = cb->next)
    {
	cout << indent << "  . function = " << cb->cbf << endl;
	cout << indent << "    data     = " << cb->clientData << endl;
    }
}
#endif
