/*------------------------------------------------------------------------
 * $Id: Host.c++,v 2.9 1993/11/16 17:56:46 carlson Exp $
 *
 * Methods for handling the Host class.
 *
 * Revision History:
 *	$Log: Host.c++,v $
// Revision 2.9  1993/11/16  17:56:46  carlson
// Include systemfiles.h and we get to remove prototype for gettoken and
//   remove the static function 'Copy'.
//
// Revision 2.8  1993/11/16  01:34:08  carlson
// Added code to keep track of whether we've been modified or not.
// Removed code to find illegal entry of baud.
//
// Revision 2.7  1993/10/23  00:36:19  carlson
// Had to make the COPY macro a function.  It clearly wasn't doing what
//   I wanted it to do - if parameter is not given, insert NULL.
//
// Revision 2.6  1993/10/19  23:53:07  carlson
// Modified addChangeCallback to only allow one copy of a callback.  Thus,
//   it won't call the same callback twice.
// Added a removeChangeCallback method.
//
// Revision 2.5  1993/10/18  18:32:39  carlson
// Added debug function 'dump' which dumps out everything about an instance
//   of a class to stdout.
//
// Revision 2.4  1993/10/06  02:02:01  chris
// Added the method 'line' which rebuilds the line from the Systems file.
// Added two versions of isMatch which will return true if this host
//   is a match for another host or data passed.
//
// Revision 2.3  1993/07/30  23:45:51  chris
// Added the ability to define callbacks to be called when something
//   changes in the structure.
// Changed the names of the structure elements so that they can be accessed
//   by methods of their original names.
// Added methods to change the values of the structure elements.  This way
//   each method can call the callbacks when they're called.
//
// Revision 2.2  1993/07/28  16:14:03  chris
// Changed all calls to strtok to use gettoken.
// Removed redundant include of linkList.h.
//
// Revision 2.1  1993/06/30  04:34:33  chris
// Initial revision.
//
 *------------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>

#include "Host.h"
#include "systemfiles.h"

#define COPY(from,to)	(to) = Copy (from)
#define DELETE(field)	if (field) delete [] field

/*========================================================================
 * Host class
 *========================================================================*/

/*------------------------------------------------------------------------
 * Host::Host (const char *)
 *------------------------------------------------------------------------*/

Host::Host (const char *line) 
: LinkData ()
{
    char	*cptr, *newline, *nexttoken;

    callbacks = NULL;
    modified = 0;

    /*----
     * Make a duplicate of the line so we can edit it without screwing
     * up the original.
     *----*/

    newline = new char [strlen (line) + 1];
    nexttoken = newline;
    strcpy (newline, line);

    /*----
     * Pull off first token and give it to LinkData initialization code.
     *----*/

    cptr = gettoken (&nexttoken, " \t\n");
    setKey (cptr);

    /*----
     * Now, pull off all other fields and put them in their respective places.
     *----*/

    cptr = gettoken (&nexttoken, " \t\n");
    COPY (cptr, s);

    cptr = gettoken (&nexttoken, " \t\n");
    COPY (cptr, d);

    b = 0;
    cptr = gettoken (&nexttoken, " \t\n");
    if (cptr)
	b = (strcmp (cptr, "Any") == 0) ? ANYBAUD : atoi (cptr);

    cptr = gettoken (&nexttoken, " \t\n");
    COPY (cptr, p);

    COPY (nexttoken, c);

    delete [] newline;
}

/*------------------------------------------------------------------------
 * Host::~Host ()
 *------------------------------------------------------------------------*/

Host::~Host ()
{
    DELETE (c);
    DELETE (p);
    DELETE (d);
    DELETE (s);
}

/*------------------------------------------------------------------------
 * Host::line \(void\)
 *------------------------------------------------------------------------*/

char *Host::line (void)
{
    static char		buffer[256];

    if (b == ANYBAUD)
	sprintf (buffer, "%s %s %s %s %s %s", theKey (), s, d, "Any", p, c);
    else
	sprintf (buffer, "%s %s %s %d %s %s", theKey (), s, d, b, p, c);
    return buffer;
}

/*------------------------------------------------------------------------
 * Host::setSchedule \(char *\)
 *------------------------------------------------------------------------*/

void Host::setSchedule (char *ns)
{
    callbackStruct	*cbs;

    if (s)
	DELETE (s);

    COPY (ns, s);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, SCHEDULE,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::setDevice \(char *\)
 *------------------------------------------------------------------------*/

void Host::setDevice (char *nd)
{
    callbackStruct	*cbs;

    if (d)
	DELETE (d);

    COPY (nd, d);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, DEVICE,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::setBaud \(int\)
 *------------------------------------------------------------------------*/

void Host::setBaud (int nb)
{
    callbackStruct	*cbs;

    b = nb;
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, BAUD, cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::setBaud \(char *\)
 *------------------------------------------------------------------------*/

void Host::setBaud (char *nb)
{
    callbackStruct	*cbs;
    int			old_baud = b;

    b = (strcmp (nb, "Any") == 0) ? ANYBAUD : atoi (nb);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, BAUD, cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::setPhone \(char *\)
 *------------------------------------------------------------------------*/

void Host::setPhone (char *np)
{
    callbackStruct	*cbs;

    if (p)
	DELETE (p);

    COPY (np, p);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, PHONE, cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::setChatScript \(char *\)
 *------------------------------------------------------------------------*/

void Host::setChatScript (char *nc)
{
    callbackStruct	*cbs;

    if (c)
	DELETE (c);

    COPY (nc, c);
    modified = 1;

    if (callbacks)
	for (cbs = callbacks; cbs; cbs = cbs->next)
	    if (cbs->cbf)
		(*((callbackFunction)cbs->cbf))(this, CHATSCRIPT,
						cbs->clientData);
}

/*------------------------------------------------------------------------
 * Host::isMatch \(char *, char *, int\)
 *------------------------------------------------------------------------*/

int Host::isMatch (char *name, char *device, int baud)
{
    int		retval = 1;			// Assume true

    if (strcmp (name, theKey ()))
	retval = 0;
    if (strcmp (device, d))
	retval = 0;
    if (baud != b)
	retval = 0;

    return retval;
}

/*------------------------------------------------------------------------
 * Host::isMatch \(Host *\)
 *------------------------------------------------------------------------*/

int Host::isMatch (Host *host)
{
    int		retval = 1;			// Assume true

    if (strcmp (host->theKey (), theKey ()))
	retval = 0;
    if (strcmp (host->d, d))
	retval = 0;
    if (host->b != b)
	retval = 0;

    return retval;
}

/*------------------------------------------------------------------------
 * Host::addChangeCallback \(callbackFunction, void *\)
 *------------------------------------------------------------------------*/

void Host::addChangeCallback (callbackFunction cbf, void *cd)
{
    callbackStruct	*cbs, *cbstemp;

    cbs = new callbackStruct;
    cbs->cbf = cbf;
    cbs->clientData = cd;
    cbs->next = NULL;

    if (callbacks)
    {
	for (cbstemp = callbacks; cbstemp->next; cbstemp = cbstemp->next)
	{
	    if ((cbstemp->cbf == cbf) && (cbstemp->clientData == cd))
	    {
		delete cbs;
		return;
	    }
	}
	cbstemp->next = cbs;
    }
    else
	callbacks = cbs;
}

/*------------------------------------------------------------------------
 * Host::removeChangeCallback \(callbackFunction, void *\)
 *------------------------------------------------------------------------*/

void Host::removeChangeCallback (callbackFunction cbf, void *cd)
{
    callbackStruct	*cbs, *cbstemp;

    if (callbacks)
    {
	for (cbs = NULL, cbstemp = callbacks; cbstemp; cbstemp = cbstemp->next)
	{
	    if ((cbstemp->cbf == cbf) && (cbstemp->clientData == cd))
	    {
		if (cbs)
		    cbs->next = cbstemp->next;
		else
		    callbacks = cbstemp->next;
		delete cbstemp;
		break;
	    }
	    cbs = cbstemp;
	}
    }
}

#ifdef DEBUG
/*------------------------------------------------------------------------
 * Host::dump \(char *\)
 *------------------------------------------------------------------------*/

void Host::dump (char *indent)
{
    char		*nextIndent;
    int			i;
    callbackStruct	*cb;

    i = strlen (indent);
    nextIndent = new char [i + 5];
    strcpy (nextIndent, indent);
    strcat (nextIndent, "    ");

    cout << indent << "Host dump (" << this << "):" << endl;
    cout << indent << "  host     = " << theKey () << endl;
    cout << indent << "  schedule = " << s << endl;
    cout << indent << "  device   = " << d << endl;
    cout << indent << "  baud     = " << b << endl;
    cout << indent << "  phone #  = " << p << endl;
    cout << indent << "  chat     = " << c << endl;
    cout << endl;
    cout << indent << "  Callbacks:" << endl;
    for (cb = callbacks; cb; cb = cb->next)
    {
	cout << indent << "  . function = " << cb->cbf << endl;
	cout << indent << "    data     = " << cb->clientData << endl;
    }
}
#endif
