/*------------------------------------------------------------------------
 * $Id: ModemFile.c++,v 2.9 1994/02/25 20:50:43 carlson Exp $
 *
 * Methods for ModemFile class.
 *
 * Revision History:
 *	$Log: ModemFile.c++,v $
// Revision 2.9  1994/02/25  20:50:43  carlson
// Modified readFile and writeFile to take a parameter telling it which file to read
//   or write.
// Added methods to insert a file into the current class without changing the name of
//   the file this class represents.
//
// Revision 2.8  1994/02/03  04:07:12  chrisc
// Added methods to delete registers.
// Added methods to support setting the answer ability of a modem.
//
// Revision 2.7  1994/01/19  00:20:00  carlson
// Modified getSetRSend and getSetFSend to return NULL if no value is set
//   for this field *and* if none matches the register passed (if passed).
//
// Revision 2.6  1994/01/18  21:49:21  carlson
// Added some comments to readFile.
// Modified how we write file.
//   Found an error with always finding documentation on registers and seqs.
//   Found an error with there always being at least two lines of docs.
//   Fixed error where lines with docs were written twice before a newline.
// Fixed error where an unrecognized baud still returns the speed string.
//
// Revision 2.5  1994/01/18  03:31:31  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.4  1994/01/17  20:12:23  carlson
// Added methods to get first and next bauds.
//
// Revision 2.3  1994/01/14  02:17:32  carlson
// Modified to change strings of unprintable characters to printable when
//   writing to file and undoes the process when reading.
// Fixed some bugs with skipping characters when expanding.
//
// Revision 2.2  1994/01/13  20:08:57  carlson
// Removed function dupstr since the Copy function in systemfiles.c++ does
//   the same thing.
// Call readFile on instantiation only if the file exists.
// Can't use gettoken to parse sequences since there are times when there
//   is nothing to send (or expect) and we need to recognize this.
// Added a number of methods to retrieve records.  These methods will expand
//   the strings where needed, if requested.
//
// Revision 2.1  1994/01/11  22:17:48  carlson
// *** empty log message ***
//
 *------------------------------------------------------------------------*/

#include "ModemFile.h"
#include "systemfiles.h"

#include <Vk/VkErrorDialog.h>

#include <unistd.h>
#include <stdlib.h>
#include <fstream.h>
#include <strstream.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

static char	buff[256];

/*------------------------------------------------------------------------
 * ModemFile::ModemFile (PathName &)
 *------------------------------------------------------------------------*/

ModemFile::ModemFile (PathName &name)
{
    fileName = name;
    _error = 0;

    registers.setOrdering (LinkList::ASCENDING);
    sequences.setOrdering (LinkList::ASCENDING);

    if (access (fileName.name (), R_OK) == 0)
	readFile (fileName);
}

/*------------------------------------------------------------------------
 * ModemFile::ModemFile (const char *)
 *------------------------------------------------------------------------*/

ModemFile::ModemFile (const char *name)
{
    fileName = name;
    _error = 0;

    registers.setOrdering (LinkList::ASCENDING);
    sequences.setOrdering (LinkList::ASCENDING);

    if (access (name, R_OK) == 0)
	readFile (fileName);
}

/*------------------------------------------------------------------------
 * ModemFile::insertFile (PathName &)
 *------------------------------------------------------------------------*/

int ModemFile::insertFile (PathName &name)
{
    if (access (name.name (), R_OK) != 0)
    {
	_error = errno;
	return -1;
    }

    readFile (name);

    return (_error) ? -1 : 0;
}

/*------------------------------------------------------------------------
 * ModemFile::insertFile (const char *)
 *------------------------------------------------------------------------*/

int ModemFile::insertFile (const char *name)
{
    PathName		Name ((char *)name);

    return insertFile (Name);
}

/*------------------------------------------------------------------------
 * ModemFile::firstRegister (void)
 *------------------------------------------------------------------------*/

Registers *ModemFile::firstRegister (void)
{
    return (Registers *)registers.first ();
}

/*------------------------------------------------------------------------
 * ModemFile::nextRegister (void)
 *------------------------------------------------------------------------*/

Registers *ModemFile::nextRegister (void)
{
    return (Registers *)registers.next ();
}

/*------------------------------------------------------------------------
 * ModemFile::findRegister (const char *)
 *------------------------------------------------------------------------*/

Registers *ModemFile::findRegister (const char *reg)
{
    return (Registers *)registers.find (reg);
}

/*------------------------------------------------------------------------
 * ModemFile::deleteRegister (const char *)
 *------------------------------------------------------------------------*/

void ModemFile::deleteRegister (const char *reg)
{
    Registers		*found;

    found = (Registers *)registers.find (reg);
    if (found)
    {
	registers.removeEntry (found);
	delete found;
    }
}

/*------------------------------------------------------------------------
 * ModemFile::deleteRegister (Registers *)
 *------------------------------------------------------------------------*/

void ModemFile::deleteRegister (Registers *reg)
{
    Registers		*found;

    found = (Registers *)registers.find (reg);
    if (found)
    {
	registers.removeEntry (found);
	delete found;
    }
}

/*------------------------------------------------------------------------
 * ModemFile::firstSequence (void)
 *------------------------------------------------------------------------*/

Sequences *ModemFile::firstSequence (void)
{
    return (Sequences *)sequences.first ();
}

/*------------------------------------------------------------------------
 * ModemFile::nextSequence (void)
 *------------------------------------------------------------------------*/

Sequences *ModemFile::nextSequence (void)
{
    return (Sequences *)sequences.next ();
}

/*------------------------------------------------------------------------
 * ModemFile::findSequence (const char *)
 *------------------------------------------------------------------------*/

Sequences *ModemFile::findSequence (const char *seq)
{
    return (Sequences *)sequences.find (seq);
}

/*------------------------------------------------------------------------
 * ModemFile::firstBaud (void)
 *------------------------------------------------------------------------*/

char *ModemFile::firstBaud (void)
{
    lastBaud = (Sequences *)sequences.find ("baud");
    if (lastBaud)
	return lastBaud->getSend ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::nextBaud (void)
 *------------------------------------------------------------------------*/

char *ModemFile::nextBaud (void)
{
    Sequences		*seq;

    if (lastBaud)
    {
	for (seq = (Sequences *)sequences.find ("baud"); seq;
	     seq = (Sequences *)sequences.next ())
	    if (seq == lastBaud)
		break;

	lastBaud = NULL;
	if (seq)
	{
	    seq = (Sequences *)sequences.next ();
	    if (seq)
		if (strcmp (seq->theKey (), "baud") == 0)
		    lastBaud = seq;
	}
    }

    if (lastBaud)
	return lastBaud->getSend ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::addRegister (Registers *)
 *------------------------------------------------------------------------*/

void ModemFile::addRegister (Registers *reg)
{
    registers.add (reg);
}

/*------------------------------------------------------------------------
 * ModemFile::addSequence (Sequences *)
 *------------------------------------------------------------------------*/

void ModemFile::addSequence (Sequences *reg)
{
    sequences.add (reg);
}

/*------------------------------------------------------------------------
 * ModemFile::readFile (PathName &)
 *
 * Format of file:
 *	- All lines at the beginning of the file that have any white
 *	  space in the first column are considered a comment and are
 *	  ignored.
 *	- Blank lines are ignored
 *	- All lines prior to first \f (^L) is a register definition.
 *	- All lines after the first \f (^L) is a sequence definition.
 *	- Register definition is of the form:
 *		<number> <type> <value> [<documentation>]
 *	  The <number> *must* start in column 1.
 *	- All lines *after* a register definition that begin with a
 *	  white space are continuations of the documentation.
 *	- Sequence definitions are of the form:
 *		<seq>\t<send>\t<expect>\t<comment>
 *	  The <seq> *must* start in column 1.
 *	  <seq> is one of the following:
 *		check		How to check if modem is there.
 *		speed		How to set the interface speed of the modem.
 *		setR		How to set a register.
 *		setF		How to set a flag.
 *		save		How to save the configuration.
 *		baud		Each baud rate supported and sequence used
 *				to set it.
 *	- Baud is somewhat different.  Many 'baud' sequences can be given.
 *	  The <send> portion identifies the baud rate and the <expect>
 *	  portion identifies the symbol that is used to set this baud.
 *------------------------------------------------------------------------*/

void ModemFile::readFile (PathName &filename)
{
    ifstream		*file;
    char		line[128], *cptr;
    ostrstream		*ostrng;
    Registers		*reg = NULL;
    char		*regNum, *regTyp, *regVal, *regDoc;
    Registers::Type	type;
    char		*seqNam, *seqSend, *seqExpect, *seqDoc;
    Sequences		*seq = NULL;

    /*----
     * Open file and error if the file doesn't exist or can't be opened.
     *----*/

    file = new ifstream (filename.name ());
    if (file->fail ())
    {
	ostrng = new ostrstream (line, sizeof (line));
	*ostrng << "Unable to open file " << filename.name () << endl;
	*ostrng << "  " << strerror (errno) << endl;
	theErrorDialog->postAndWait (line);
	delete ostrng;
	return;
    }

    /*----
     * Read the register settings.
     *	[Registers are defined up to ^L]
     *----*/

    while (!file->eof ())
    {
	file->getline (line, sizeof (line));

	/*----
	 * Check if end of register settings.  If yes, quit.
	 *----*/

	if (line[0] == '\f')
	    break;

	/*----
	 * Check if this line is a continuation of the previous line's
	 * comments.  If yes, append to the comments.
	 *----*/

	if (isspace (line[0]))
	{
	    if (reg)
	    {
		for (cptr = line; isspace (*cptr); cptr++);
		reg->appendDocumentation (cptr);
	    }
	    continue;
	}

	/*----
	 * Parse the fields of the line.
	 *	<number> <type> <value> <documentation>
	 *----*/

	cptr = line;
	regNum = Copy (gettoken (&cptr, " \t"));
	regTyp = Copy (gettoken (&cptr, " \t"));
	type = Registers::NONE;
	if (strcmp (regTyp, "register") == 0)
	    type = Registers::REGISTER;
	else if (strcmp (regTyp, "flag") == 0)
	    type = Registers::FLAG;
	regVal = Copy (gettoken (&cptr, " \t"));
	regDoc = Copy (cptr);

	reg = new Registers (regNum, type, regVal, regDoc);
	registers.add (reg);
    }

    /*----
     * Read the special sequences.
     *----*/

    while (!file->eof ())
    {
	file->getline (line, sizeof (line));

	if (line[0] == '\f')		// Added because there's something
	    break;			//   wrong with file->eof ()

	if (isspace (line[0]))
	{
	    if (seq)
	    {
		for (cptr = line; isspace (*cptr); cptr++);
		seq->appendDocumentation (cptr);
	    }
	    continue;
	}

	/*----
	 * Get sequence name.
	 *----*/

	cptr = line;
	while (*cptr && isspace (*cptr))
	    cptr++;
	seqSend = cptr;
	while (*cptr && (*cptr != '\t'))
	    cptr++;
	if (*cptr)
	    *cptr++ = '\0';
	seqNam = Copy (seqSend);

	/*----
	 * Get send sequence.
	 *----*/

	seqExpect = cptr;
	while (*cptr && (*cptr != '\t'))
	    cptr++;
	if (*cptr)
	    *cptr++ = '\0';
	seqSend = Copy (parse_string (seqExpect));

	/*----
	 * Get expect sequence.
	 *----*/

	seqDoc = cptr;
	while (*cptr && (*cptr != '\t'))
	    cptr++;
	if (*cptr)
	    *cptr++ = '\0';
	seqExpect = Copy (seqDoc);

	/*----
	 * The rest is the documentation sequence.
	 *----*/

	seqDoc = Copy (cptr);

	seq = new Sequences (seqNam, seqSend, seqExpect, seqDoc);
	sequences.add (seq);
    }

    file->close ();
}

/*------------------------------------------------------------------------
 * ModemFile::writeFile (PathName &)
 *------------------------------------------------------------------------*/

void ModemFile::writeFile (PathName &filename)
{
    ofstream		*file;
    char		line[128], *cptr, *dptr;
    int			n;
    ostrstream		*ostrng;
    istrstream		*istrng;
    Registers		*reg;
    Sequences		*seq;

    file = new ofstream (filename.name ());
    if (file->fail ())
    {
	ostrng = new ostrstream (line, sizeof (line));
	*ostrng << "Unable to open file " << filename.name () << endl;
	*ostrng << "  " << strerror (errno) << endl;
	theErrorDialog->postAndWait (line);
	delete ostrng;
	return;
    }

    for (reg = (Registers *)registers.first (); reg;
	 reg = (Registers *)registers.next ())
    {
	strcpy (line, reg->theKey ());
	strcat (line, " ");
	cptr = "register";
	if (reg->getType () == Registers::FLAG)
	    cptr = "flag";
	strcat (line, cptr);
	strcat (line, " ");
	strcat (line, reg->getValue ());
	*file << line;
	cptr = reg->getDocumentation ();
	if (cptr && strlen (cptr))
	{
	    dptr = strchr (cptr, '\n');
	    n = dptr ? (dptr - cptr) : strlen (cptr);
	    if ((strlen (line) + n) >= sizeof (line))
		*file << endl;
	    line[0] = '\0';
	    istrng = new istrstream (cptr);
	    while (!istrng->eof ())
	    {
		istrng->getline (line, sizeof (line));
		if (strlen (line))
		    *file << " " << line << endl;
	    }
	    delete istrng;
	}
	else
	    *file << endl;
    }

    *file << "\f" << endl;

    for (seq = (Sequences *)sequences.first (); seq;
	 seq = (Sequences *)sequences.next ())
    {
	strcpy (line, seq->theKey ());
	strcat (line, "\t");
	strcat (line, unparse_string (seq->getSend ()));
	strcat (line, "\t");
	strcat (line, seq->getExpect ());
	*file << line;
	cptr = seq->getDocumentation ();
	if (cptr && strlen (cptr))
	{
	    dptr = strchr (cptr, '\n');
	    n = dptr ? (dptr - cptr) : strlen (cptr);
	    if ((strlen (line) + n) >= sizeof (line))
		*file << endl;
	    line[0] = '\0';
	    istrng = new istrstream (cptr);
	    while (!istrng->eof ())
	    {
		istrng->getline (line, sizeof (line));
		if (strlen (line))
		    *file << "\t" << line << endl;
	    }
	    delete istrng;
	}
	else
	    *file << endl;
    }

    *file << "\f" << endl;

    file->close ();
}

/*------------------------------------------------------------------------
 * ModemFile::getBaudSequence (char *)
 *------------------------------------------------------------------------*/

Sequences *ModemFile::getBaudSequence (char *baud)
{
    Sequences			*seq = NULL;

    if (baud)
    {
	for (seq = findSequence ("baud"); seq; seq = nextSequence ())
	{
	    if (strcmp (seq->theKey (), "baud"))
	    {
		seq = NULL;
		break;
	    }

	    if (strcmp (seq->getSend (), baud) == 0)
		break;
	}
    }

    return seq;
}

/*------------------------------------------------------------------------
 * ModemFile::getBaudSequence (int)
 *------------------------------------------------------------------------*/

Sequences *ModemFile::getBaudSequence (int baud)
{
    char			baudStr[10];

    if (baud >= 0)
    {
	sprintf (baudStr, "%d", baud);
	return getBaudSequence (baudStr);
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getCheckSend (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getCheckSend (void)
{
    Sequences			*seq;

    seq = findSequence ("check");
    if (seq)
	return seq->getSend ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getCheckExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getCheckExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("check");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSetRSend (Registers *)
 *------------------------------------------------------------------------*/

char *ModemFile::getSetRSend (Registers *reg)
{
    Sequences			*seq;

    seq = findSequence ("setR");
    if (seq)
    {
	if (reg)
	{
	    if (reg->getType () == Registers::REGISTER)
		return expandRegister (seq->getSend (), (char *)reg->theKey (),
				       reg->getValue ());
	}
	else
	    return seq->getSend ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSetRExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getSetRExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("setR");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSetFSend (Registers *)
 *------------------------------------------------------------------------*/

char *ModemFile::getSetFSend (Registers *reg)
{
    Sequences			*seq;

    seq = findSequence ("setF");
    if (seq)
    {
	if (reg)
	{
	    if (reg->getType () == Registers::FLAG)
		return expandFlag (seq->getSend (), (char *)reg->theKey (),
				   reg->getValue ());
	}
	else
	    return seq->getSend ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSetFExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getSetFExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("setF");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSaveSend (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getSaveSend (void)
{
    Sequences			*seq;

    seq = findSequence ("save");
    if (seq)
	return seq->getSend ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSaveExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getSaveExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("save");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSpeedSend (char *)
 *------------------------------------------------------------------------*/

char *ModemFile::getSpeedSend (char *baud)
{
    Sequences			*seq;
    char			*baudCode;

    seq = findSequence ("speed");
    if (seq)
    {
	if (baud)
	{
	    baudCode = getBaudCode (baud);
	    if (baudCode)
		return expandBaud (seq->getSend (), baudCode);
	}
	else
	    return seq->getSend ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSpeedSend (int)
 *------------------------------------------------------------------------*/

char *ModemFile::getSpeedSend (int baud)
{
    Sequences			*seq;
    char			*baudCode;

    seq = findSequence ("speed");
    if (seq)
    {
	if (baud >= 0)
	{
	    baudCode = getBaudCode (baud);
	    if (baudCode)
		return expandBaud (seq->getSend (), baudCode);
	}
	else
	    return seq->getSend ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getSpeedExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getSpeedExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("speed");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getPrefix (char *)
 *------------------------------------------------------------------------*/

char *ModemFile::getPrefix (char *str)
{
    Sequences			*seq;
    static char			buff[256];
    char			*cptr, *dptr, *pfx;

    pfx = NULL;
    seq = findSequence ("prefix");
    if (seq)
    {
	pfx = seq->getSend ();

	if (str)
	{
	    for (cptr = buff; (*cptr = *pfx); pfx++)
	    {
		if (*pfx == '%')
		{
		    if (*(++pfx) == 'c')
		    {
			for (dptr = str; (*cptr = *dptr); dptr++)
			    cptr++;
			cptr--;
		    }
		    else if (*pfx == '%')
			pfx++;
		    else
			pfx--;
		}

		cptr++;
	    }

	    pfx = buff;
	}
    }

    return pfx;
}

/*------------------------------------------------------------------------
 * ModemFile::getMinimumSend (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getMinimumSend (void)
{
    Sequences			*seq;

    seq = findSequence ("minimum");
    if (seq)
	return seq->getSend ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getMinimumExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getMinimumExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("minimum");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getBaudCode (char *)
 *------------------------------------------------------------------------*/

char *ModemFile::getBaudCode (char *baud)
{
    Sequences			*seq;

    if (baud)
    {
	seq = getBaudSequence (baud);
	if (seq)
	    return seq->getExpect ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getBaudCode (int)
 *------------------------------------------------------------------------*/

char *ModemFile::getBaudCode (int baud)
{
    Sequences			*seq;

    seq = getBaudSequence (baud);
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getAnswerSend (char *)
 *------------------------------------------------------------------------*/

char *ModemFile::getAnswerSend (char *rings)
{
    if (rings)
	return getAnswerSend (atoi (rings));

    return getAnswerSend (-1);
}

/*------------------------------------------------------------------------
 * ModemFile::getAnswerSend (int)
 *------------------------------------------------------------------------*/

char *ModemFile::getAnswerSend (int rings)
{
    Sequences		*seq;

    seq = findSequence ("answer");
    if (seq)
    {
	if (rings >= 0)
	    return expandAnswer (seq->getSend (), rings);

	return seq->getSend ();
    }

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::getAnswerExpect (void)
 *------------------------------------------------------------------------*/

char *ModemFile::getAnswerExpect (void)
{
    Sequences			*seq;

    seq = findSequence ("answer");
    if (seq)
	return seq->getExpect ();

    return NULL;
}

/*------------------------------------------------------------------------
 * ModemFile::expandRegister (char *, char *, char *)
 *------------------------------------------------------------------------*/

char *ModemFile::expandRegister (char *str, char *rnum, char *rval)
{
    char		*cptr, *dptr;

    for (cptr = buff; (*cptr = *str); str++)
    {
	if (*str == '%')
	{
	    switch (*(++str))
	    {
	      case '%':
		break;

	      case 'r':
		for (dptr = rnum; (*cptr = *dptr); dptr++)
		    cptr++;
		cptr--;
		break;

	      case 'v':
		for (dptr = rval; (*cptr = *dptr); dptr++)
		    cptr++;
		cptr--;
		break;

	      default:
		str--;
		break;
	    }
	}

	cptr++;
    }

    return buff;
}

/*------------------------------------------------------------------------
 * ModemFile::expandFlag (char *, char *, char *)
 *------------------------------------------------------------------------*/

char *ModemFile::expandFlag (char *str, char *fstr, char *fval)
{
    char		*cptr, *dptr;

    for (cptr = buff; (*cptr = *str); str++)
    {
	if (*str == '%')
	{
	    switch (*(++str))
	    {
	      case '%':
		break;

	      case 'f':
		for (dptr = fstr; (*cptr = *dptr); dptr++)
		    cptr++;
		cptr--;
		break;

	      case 'v':
		for (dptr = fval; (*cptr = *dptr); dptr++)
		    cptr++;
		cptr--;
		break;

	      default:
		str--;
		break;
	    }
	}

	cptr++;
    }

    return buff;
}

/*------------------------------------------------------------------------
 * ModemFile::expandBaud (char *, char *)
 *------------------------------------------------------------------------*/

char *ModemFile::expandBaud (char *str, char *baud)
{
    char		*cptr, *dptr;

    for (cptr = buff; (*cptr = *str); str++)
    {
	if (*str == '%')
	{
	    switch (*(++str))
	    {
	      case '%':
		break;

	      case 'b':
		for (dptr = baud; (*cptr = *dptr); dptr++)
		    cptr++;
		cptr--;
		break;

	      default:
		str--;
		break;
	    }
	}

	cptr++;
    }

    return buff;
}

/*------------------------------------------------------------------------
 * ModemFile::expandAnswer (char *, int)
 *------------------------------------------------------------------------*/

char *ModemFile::expandAnswer (char *str, int rings)
{
    char		*cptr;

    for (cptr = buff; (*cptr = *str); str++)
    {
	if (*str == '%')
	{
	    switch (*(++str))
	    {
	      case '%':
		break;

	      case 'n':
		sprintf (cptr, "%d", rings);
		cptr += strlen (cptr) - 1;
		break;

	      case 'b':
		*cptr = (rings) ? '1' : '0';
		break;

	      default:
		str--;
		break;
	    }
	}

	cptr++;
    }

    return buff;
}
