/*------------------------------------------------------------------------
 * $Id: Package.c++,v 2.16 1994/02/17 04:41:10 chrisc Exp $
 *
 * Methods for Package class.
 *
 * Revision History:
 *	$Log: Package.c++,v $
// Revision 2.16  1994/02/17  04:41:10  chrisc
// Include systemfiles.h instead of prototyping gettoken.
// Added support for a flag that indicates if the Package has been modified.
// Added method for building a line to be output to Sysfiles.
//
// Revision 2.15  1994/02/15  04:21:14  chrisc
// Removed all references to dialcodes.  It is now done separately from
//   the main window.
//
// Revision 2.14  1994/01/18  05:41:54  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.13  1993/11/05  20:11:49  carlson
// Changed all references to LinkList to use SmartList.
//
// Revision 2.12  1993/10/22  02:01:52  carlson
// Made sure the ordering for the systems, devices and dialers are FIFO
//   even if the other instantiation occurs.
// Be sure to expand file names before working with them.
//   This fixed a bug when we sometime referred to the same file but one
//   had a fully qualified name and the other didn't.
//
// Revision 2.11  1993/10/19  22:52:01  carlson
// Make the lists of files FIFO ordered.
// Added method to dump information to stdout.
//
// Revision 2.10  1993/10/08  18:14:36  carlson
// Made separate methods for adding a file to one of the lists of files
//   that belong to a package.  These methods are used both when first
//   initializing the package from the file and when a new file is to
//   be included.
//
// Revision 2.9  1993/10/07  22:46:28  carlson
// No longer compute the path to use here.  Expect it to be passed down.
// Need to provide path to expand_files.
// Now PathName doesn't return a NULL but a null string.  Need to call the
//   length function to see if we have hit the last entry from expand_files.
//
// Revision 2.8  1993/08/12  20:02:07  chris
// Added method for setting the default package after creation.
//
// Revision 2.7  1993/07/29  00:32:25  chris
// Commented out definition of DEBUG1.
//
// Revision 2.6  1993/07/28  19:29:30  chris
// Get default Package when creating (if one exists).
// Remove commented out code.
// Added method for adding Dialcodes after instantiation.
//
// Revision 2.5  1993/07/28  15:26:34  chris
// Added code to read and handle Devices, Dialers and Dialcodes classes.
// Changed all calls to strtok to use gettoken.
//
// Revision 2.4  1993/07/23  17:14:20  chris
// Added some documentation.
// Added code to add the package to the list of known packages for each
//   system processed.
//
// Revision 2.3  1993/07/07  18:09:24  chris
// Commented out print statements.
// Added symbols for indexing into all_files array.
// Added method for making a new System class so that PackageWindow can
//   call something different.
// Provided a separate method for filling the info in Package class so it
//   can be called at a more appropriate time in PackageWindow.
// Completely rewrote portion which breaks out line.
//
// Revision 2.2  1993/07/01  01:32:51  chris
// Use new PathName::operator+= operator.
// Error with checking the wrong string for 'service'.
// Modified way system is stored in various lists.
//
// Revision 2.1  1993/06/30  04:34:33  chris
// Initial revision.
//
 *------------------------------------------------------------------------*/

#include <string.h>
#include <errno.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include "Package.h"
#include "System.h"
#include "Devices.h"
#include "Dialers.h"
#include "Dialcodes.h"
#include "systemfiles.h"

#define SYSTEMS		0
#define DEVICES		1
#define DIALERS		2

/*========================================================================
 * Static functions required
 *========================================================================*/

/*------------------------------------------------------------------------
 * expand_files
 *	Given a list of the form "name1:name2:name3...", this routine
 *	returns an array of PathNames each having the format of
 *	"/etc/uucp/nameN".  If there is already a path in front of
 *	"nameN", this routine will not change it.
 *------------------------------------------------------------------------*/

static PathName *expand_files (char *line, PathName& path)
{
    int		file_count;
    char	*cptr, *f;
    PathName	*file_names;
    char	file[256];

    /*----
     * First, count how many files are going to be defined and
     * create an area for an array of pointers (+1).
     *----*/

    for (file_count = 1, cptr = line; *cptr; cptr++)
	if (*cptr == ':')
	    file_count++;

    file_names = new PathName[file_count + 1];

    /*----
     * Now collect the file names and allocate memory for them,
     * placing the pointers in the array just created.
     *----*/

    for (file_count = 0, cptr = line, f = file; 1; cptr++)
    {
	if ((*cptr == '\0') || (*cptr == ':'))
	{
	    *f = '\0';
	    if (file[0] == '/')
		file_names[file_count] =  file;
	    else
	    {
		file_names[file_count] = path;
		file_names[file_count] += file;
	    }
	    file_count++;
	    f = file;

	    if (*cptr == '\0')
		break;
	}
	else
	    *f++ = *cptr;
    }

    return file_names;
}

/*========================================================================
 * Package class
 *========================================================================*/

/*------------------------------------------------------------------------
 * Package::makeNewSystem (PathName)
 *------------------------------------------------------------------------*/

System *Package::makeNewSystem (PathName& filename)
{
#ifdef DEBUG1
    cout << "Package::makeNewSystem(" << this
	 << ") - Filename = " << (const char *)filename
	 << endl;
#endif

    System *s = new System (filename);
    return s;
}

/*------------------------------------------------------------------------
 * Package::makeNewDevice (PathName)
 *------------------------------------------------------------------------*/

Devices *Package::makeNewDevice (PathName& filename)
{
#ifdef DEBUG1
    cout << "Package::makeNewDevice(" << this
	 << ") - Filename = " << (const char *)filename
	 << endl;
#endif

    Devices *s = new Devices (filename);
    return s;
}

/*------------------------------------------------------------------------
 * Package::makeNewDialer (PathName)
 *------------------------------------------------------------------------*/

Dialers *Package::makeNewDialer (PathName& filename)
{
#ifdef DEBUG1
    cout << "Package::makeNewDialer(" << this
	 << ") - Filename = " << (const char *)filename
	 << endl;
#endif

    Dialers *s = new Dialers (filename);
    return s;
}

/*------------------------------------------------------------------------
 * Package::Package (const char *line, SmartList *all_files)
 *	This is the proper way to instantiate this class.  It will not
 *	only create an instance but will also fill in all the data and
 *	instantiate subclasses.
 *------------------------------------------------------------------------*/

Package::Package (const char *line, SmartList *all_files, Package *dflt,
		  PathName* dir)
: LinkData ()
{
    if (dir)
	dirPath = *dir;

    systems.setOrdering (LinkList::FIFO);
    devices.setOrdering (LinkList::FIFO);
    dialers.setOrdering (LinkList::FIFO);

    create (line, all_files, dflt);
    modified = 0;
}

/*------------------------------------------------------------------------
 * Package::Package (const char *)
 *	This is a shortened constructor that doesn\'t fill in any of the
 *	data, just creates a name that can be put in a linked list.
 *------------------------------------------------------------------------*/

Package::Package (const char *line)
{
    char		*newline, *nexttoken;
    char		*token, *value;

    newline = new char [strlen (line) + 1];
    nexttoken = newline;
    strcpy (newline, line);

    token = gettoken (&nexttoken, " \t\n");

    value = strchr (token, '=');
    if (value == NULL)
    {
	cerr << "Package: Malformed line:" << endl;
	cerr << "    " << line << endl;
	return;
    }
    *value++ = '\0';

    if (strcmp (token, "service") != 0)
    {
	cerr << "Package: Expect 'service=' in:" << endl;
	cerr << "    " << line << endl;
	return;
    }
    setKey (value);				// Name of this package

    delete [] newline;

    systems.setOrdering (LinkList::FIFO);
    devices.setOrdering (LinkList::FIFO);
    dialers.setOrdering (LinkList::FIFO);

    modified = 0;
}

/*------------------------------------------------------------------------
 * Package::create (const char *, SmartList *, Package *)
 *	This is the function that actually does all the work of
 *	constructing a member of this class.  It creates instances of
 *	any other classes below it and fills in all fields.
 *------------------------------------------------------------------------*/

void Package::create (const char *line, SmartList *all_files,
		      Package *dflt)
{
    char		*newline, *nexttoken;
    char		*token, *value;
    PathName		*tempPaths, *aPath;
    LinkData		*ld;

    defaults = dflt;
    allFiles = all_files;

    /*----
     * Break line apart into tokens and values.
     * First token is the service=, value is what is to the right of
     * the '='.
     *----*/

    newline = new char [strlen (line) + 1];
    nexttoken = newline;
    strcpy (newline, line);
    token = gettoken (&nexttoken, " \t\n");

    value = strchr (token, '=');
    if (value == NULL)
    {
	cerr << "Package: Malformed line:" << endl;
	cerr << "    " << line << endl;
	return;
    }
    *value++ = '\0';

    if (strcmp (token, "service") != 0)
    {
	cerr << "Package: Expect 'service=' in:" << endl;
	cerr << "    " << line << endl;
	return;
    }
    setKey (value);				// Name of this package

    /*----
     * Now find other token=value pairs.  Call expand_files to break
     * the file list apart into PathNames.  Allow multiple references
     * to the same service to be given.
     *----*/

    while (token = gettoken (&nexttoken, " \t\n"))
    {
	value = strchr (token, '=');
	if (value == NULL)
	{
	    cerr << "Package: Malformed line:" << endl;
	    cerr << "    " << line << endl;
	    continue;
	}
	*value++ = '\0';

	tempPaths = expand_files (value, dirPath);
	if (strcmp (token, "systems") == 0)
	{
	    for (aPath = tempPaths; aPath->length (); aPath++)
		ld = addSystemFile (*aPath);
	}
	else if (strcmp (token, "devices") == 0)
	{
	    for (aPath = tempPaths; aPath->length (); aPath++)
		ld = addDeviceFile (*aPath);
	}
	else if (strcmp (token, "dialers") == 0)
	{
	    for (aPath = tempPaths; aPath->length (); aPath++)
		ld = addDialerFile (*aPath);
	}
	else
	{
	    cerr << "SysFile: Unrecognized token '" << token << "' in:"
		 << endl;
	    cerr << "  " << line << endl;
	}
    }

    delete [] newline;
}

/*------------------------------------------------------------------------
 * Package::line (void)
 *------------------------------------------------------------------------*/

char *Package::line (void)
{
    static char		buffer[256];
    ostrstream		*buffstr;
    int			i = 0;
    FileName		*fn;
    PathName		fName;

    buffstr = new ostrstream (buffer, sizeof (buffer));

    *buffstr << "service=" << theKey ();

    fn = (FileName *)systems.first ();
    if (fn)
    {
	i = 1;
	fName = fn->theKey ();
	*buffstr << "\tsystems=" << fName.tail ();
	while ((fn = (FileName *)systems.next ()))
	{
	    fName = fn->theKey ();
	    *buffstr << ":" << fName.tail ();
	}
    }

    fn = (FileName *)devices.first ();
    if (fn)
    {
	if (i)
	    *buffstr << " \\\n\t";
	i = 1;
	fName = fn->theKey ();
	*buffstr << "\tdevices=" << fName.tail ();
	while (fn = (FileName *)devices.next ())
	{
	    fName = fn->theKey ();
	    *buffstr << ":" << fName.tail ();
	}
    }

    fn = (FileName *)dialers.first ();
    if (fn)
    {
	if (i)
	    *buffstr << " \\\n\t";
	i = 1;
	fName = fn->theKey ();
	*buffstr << "\tdialers=" << fName.tail ();
	while (fn = (FileName *)dialers.next ())
	{
	    fName = fn->theKey ();
	    *buffstr << ":" << fName.tail ();
	}
    }

    *buffstr << ends;
    delete buffstr;
    return buffer;
}

/*------------------------------------------------------------------------
 * Package::setDefaults (Package *)
 *------------------------------------------------------------------------*/

void Package::setDefaults (Package *dflts)
{
    defaults = dflts;
}

/*------------------------------------------------------------------------
 * Package::addSystemFile (PathName&)
 *------------------------------------------------------------------------*/

LinkData* Package::addSystemFile (PathName& file_name)
{
    LinkData		*ld;
    FileName		*fn = NULL;
    System		*system;

    file_name.expand ();

    /*----
     * Find out if we have already installed this file.
     * If we have, then we're done.  Otherwise, we need
     * to create a new FileName structure and add it.
     *----*/

    system = (System *)systems.find (file_name);
    if (!system)
    {
	fn = new FileName (file_name);

	/*----
	 * Find out if this System class has been built
	 * up yet.  If yes, use the one already created,
	 * otherwise create a new one.
	 *----*/

	system = (System *)allFiles[SYSTEMS].find (file_name.name ());
	if (system)
	    fn->data = system;
	else
	{
	    system = makeNewSystem (file_name);
	    allFiles[SYSTEMS].add (system);

	    fn->data = system;
	}
	systems.add (fn);

	/*----
	 * Find out if this package is in the list of
	 * known packages for the system.  If not, add
	 * it.
	 *----*/

	ld = system->findPackage (theKey ());
	if (!ld)
	    system->addPackage (this);
    }

    modified = 1;
    return fn;
}

/*------------------------------------------------------------------------
 * Package::addDeviceFile (PathName&)
 *------------------------------------------------------------------------*/

LinkData* Package::addDeviceFile (PathName& file_name)
{
    LinkData		*ld;
    FileName		*fn = NULL;
    Devices		*device;
    
    file_name.expand ();

    /*----
     * Find out if we have already installed this file.
     * If we have, then we're done.  Otherwise, we need
     * to create a new FileName structure and add it.
     *----*/
    
    device = (Devices *)devices.find (file_name);
    if (!device)
    {
	fn = new FileName (file_name);
	
	/*----
	 * Find out if this Device class has been built
	 * up yet.  If yes, use the one already created,
	 * otherwise create a new one.
	 *----*/
	
	device = (Devices *)allFiles[DEVICES].find (file_name.name ());
	if (device)
	    fn->data = device;
	else
	{
	    device = makeNewDevice (file_name);
	    allFiles[DEVICES].add (device);
	    
	    fn->data = device;
	}
	devices.add (fn);
	
	/*----
	 * Find out if this package is in the list of
	 * known packages for the device.  If not, add
	 * it.
	 *----*/
	
	ld = device->findPackage (theKey ());
	if (!ld)
	    device->addPackage (this);
    }

    modified = 1;
    return fn;
}

/*------------------------------------------------------------------------
 * Package::addDialerFile (PathName&)
 *------------------------------------------------------------------------*/

LinkData* Package::addDialerFile (PathName& file_name)
{ 
    LinkData		*ld;
    FileName		*fn = NULL;
    Dialers		*dialer;

    file_name.expand ();

    /*----
     * Find out if we have already installed this file.
     * If we have, then we're done.  Otherwise, we need
     * to create a new FileName structure and add it.
     *----*/
    
    dialer = (Dialers *)dialers.find (file_name);
    if (!dialer)
    {
	fn = new FileName (file_name);
	
	/*----
	 * Find out if this Dialer class has been built
	 * up yet.  If yes, use the one already created,
	 * otherwise create a new one.
	 *----*/
	
	dialer = (Dialers *)allFiles[DIALERS].find (file_name.name ());
	if (dialer)
	    fn->data = dialer;
	else
	{
	    dialer = makeNewDialer (file_name);
	    allFiles[DIALERS].add (dialer);
	    
	    fn->data = dialer;
	}
	dialers.add (fn);
	
	/*----
	 * Find out if this package is in the list of
	 * known packages for the dialer.  If not, add
	 * it.
	 *----*/
	
	ld = dialer->findPackage (theKey ());
	if (!ld)
	    dialer->addPackage (this);
    }

    modified = 1;
    return fn;
}

#ifdef DEBUG
/*------------------------------------------------------------------------
 * Package::dump (char *)
 *------------------------------------------------------------------------*/

void Package::dump (char *indent)
{
    char		*nextIndent;
    int			i;
    LinkData		*ld;

    i = strlen (indent);
    nextIndent = new char [i + 5];
    strcpy (nextIndent, indent);
    strcat (nextIndent, "    ");
    
    cout << indent << "Package (" << this << "):" << endl;
    cout << indent << "  systems (" << &systems << ")" << endl;
    for (ld = systems.first (); ld; ld = systems.next ())
	cout << nextIndent << ld->theKey () << " (" << ld << ")" << endl;
    cout << indent << "  devices (" << &devices << ")" << endl;
    for (ld = devices.first (); ld; ld = devices.next ())
	cout << nextIndent << ld->theKey () << " (" << ld << ")" << endl;
    cout << indent << "  dialers (" << &dialers << ")" << endl;
    for (ld = dialers.first (); ld; ld = dialers.next ())
	cout << nextIndent << ld->theKey () << " (" << ld << ")" << endl;
    cout << "  dirPath = " << dirPath.name () << endl;
    cout << "  defaults = " << defaults << endl;
    cout << "  allFiles:" << endl;
    cout << nextIndent << "Systems = " << &allFiles[0] << endl;
    cout << nextIndent << "Devices = " << &allFiles[1] << endl;
    cout << nextIndent << "Dialers = " << &allFiles[2] << endl;
}
#endif
