/*---*- C++ -*------------------------------------------------------------
 * $Id: Package.h,v 2.14 1994/02/17 04:42:44 chrisc Exp $
 *
 * Header file for Package class.
 *
 * Revision History:
 *	$Log: Package.h,v $
 * Revision 2.14  1994/02/17  04:42:44  chrisc
 * Added prototype for method to build an output line.
 * Added functions to keep track of whether this package was modified.
 *
 * Revision 2.13  1994/02/15  04:22:01  chrisc
 * No longer keep track of dialcodes.  It is done in the main window.
 *
 * Revision 2.12  1993/11/05  20:09:29  carlson
 * Changed all references to LinkList to use SmartList.
 *
 * Revision 2.11  1993/10/19  21:25:33  carlson
 * Added method to dump information.
 *
 * Revision 2.10  1993/10/08  18:14:36  carlson
 * No longer provide 'all_files' list to addDialcodes.
 * Added virtual methods for to add files to those used by the package.
 * Added variable for holding the all_files list.
 *
 * Revision 2.9  1993/10/07  21:34:31  carlson
 * Add parameter allowing directory path to be provided.
 * Save directory path where files were found.
 *
 * Revision 2.8  1993/10/06  00:17:46  chris
 * Just moved the create method to a different part of the structure.
 *
 * Revision 2.7  1993/08/12  21:07:23  chris
 * Added prototype for virtual function for setting the default package
 *   after creation.
 *
 * Revision 2.6  1993/07/28  19:37:32  chris
 * Get default Package when constructed.
 * Made a variable to save default package.
 *
 * Revision 2.5  1993/07/28  00:02:42  chris
 * Added code to handle Devices, Dialers and Dialcodes.
 *
 * Revision 2.4  1993/07/22  00:38:27  chris
 * Added documentation.
 *
 * Revision 2.3  1993/07/07  18:09:24  chris
 * Changed FileName to a struct, since we don't need to hide anything.
 * Got rid of 'path' since it is the key.
 * Added a LinkData so that we can have just one linked list for the
 *   various files containing the file name and the structure.
 *
 * Revision 2.2  1993/06/30  21:51:59  chris
 * Don't get pathname from conversion method.  Instead, use specific method.
 *
 * Revision 2.1  1993/06/30  04:02:21  chris
 * Initial revision.
 *
 *------------------------------------------------------------------------*/

#ifndef _PACKAGE_H_
#define _PACKAGE_H_

#include "SmartList.h"
#include "PathName.h"

class System;
class Devices;
class Dialers;
class Dialcodes;

/*------------------------------------------------------------------------
 * Package class
 *	Class containing all information known for a package described in
 *	the "Sysfiles" file.
 *
 *	Internally, this class creates the structure FileName.  This
 *	structure is used to keep track of information about each
 *	file in each of the paths.  So, the linked lists "systems",
 *	"devices", "dialers" and "dialcodes" are lists of "FileName"
 *	structures.
 *
 *	The key for FileName is the file name and path.
 *	The data in FileName is a pointer to the "System", "Devices",
 *	"Dialers" or "Dialcodes" class that is referred to by the name.
 *------------------------------------------------------------------------*/

class Package : public LinkData
{
  public:
    struct FileName : public LinkData
    {
	LinkData	*data;

	FileName (PathName& pn) : LinkData (pn.name ()) { }
    };

    Package (const char *line);
    Package (const char *line, SmartList *all_files, Package *dflt = NULL,
	     PathName* dir = NULL);
    ~Package () { }

    char*		line (void);
    int			wasModified (void) { return modified; }
    void		clearModified (void) { modified = 0; }

    virtual void	setDefaults (Package *dflt);
    virtual LinkData*	addSystemFile (PathName& file_name);
    virtual LinkData*	addDeviceFile (PathName& file_name);
    virtual LinkData*	addDialerFile (PathName& file_name);

#ifdef DEBUG
    virtual void	dump (char *indent);
#endif

    SmartList		systems;
    SmartList		devices;
    SmartList		dialers;

  protected:
    int			modified;
    PathName		dirPath;
    Package		*defaults;
    SmartList		*allFiles;

    void		create (const char *line, SmartList *all_files,
				Package *dflt = NULL);

    virtual System*	makeNewSystem (PathName& pn);
    virtual Devices*	makeNewDevice (PathName& pn);
    virtual Dialers*	makeNewDialer (PathName& pn);
};

#endif
