/*------------------------------------------------------------------------
 * $Id: Sysfiles.c++,v 2.11 1994/02/17 04:44:54 chrisc Exp $
 *
 * Methods for SysFiles class.
 *
 * Revision History:
 *	$Log: Sysfiles.c++,v $
// Revision 2.11  1994/02/17  04:44:54  chrisc
// Include systemfiles.h instead of prototyping read_line.
// Added support for a flag to indicate if this class (or its children) have
//   been modified.
// Put the file name in a variable instead of losing it.
// Created save methods.
// Added methods to check modified flag and clear it.
//
// Revision 2.10  1994/02/15  04:26:11  chrisc
// No longer create the default dialcodes package.  This is taken care of
//   separately in the main window.
//
// Revision 2.9  1994/01/18  00:38:15  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.8  1993/11/12  01:10:34  carlson
// Modified to only load dirPath if it isn't already filled.  This way, it
//   can be filled by a derived class.\
//
// Revision 2.7  1993/11/05  20:09:29  carlson
// Changed all references to LinkList to use SmartList.
//
// Revision 2.6  1993/10/08  01:23:11  carlson
// No longer need to pass 'all_files' to addDialcodes.  It retains the
//   list from when the package was created.
// Added cast so that warning goes away.
//
// Revision 2.5  1993/10/07  22:56:46  carlson
// No longer figure out the path for files here.  Expect it to be where
//   the Sysfiles file is found.
// Pass this directory down to children.
//
// Revision 2.4  1993/08/12  00:31:36  chris
// After setting up the default files, make this package its own default.
//
// Revision 2.3  1993/07/28  19:07:58  chris
// Added code to add Dialcodes.
// Now keep track of the defaults Package and pass it when creating a Package.
//
// Revision 2.2  1993/06/30  23:23:16  chris
// Fixed a bug where we didn't create a full path to the file.
//
// Revision 2.1  1993/06/30  04:36:57  chris
// Updated to have new format.
//
// Revision 1.5  1993/05/16  20:26:27  chris
// Removed an unused automatic variable.
//
// Revision 1.4  1993/05/14  06:59:56  chris
// Make dialcode_list the same type as the others, for consistency.
//
// Revision 1.3  1993/05/12  17:27:55  chris
// Removed default name for Sysfiles so it can be passed as a parameter.
// Broke the instantiation code up so that there is a method which will
//   read the file and it calls a virtual method which can be redefined
//   by a derived class.
// Commented out code which builds instances of Systems, Devices, Dialers
//   and Dialcodes.  Now just leave the expanded file names in class.
// Found there was no destructor for SysFile.
//
// Revision 1.2  1993/05/10  18:08:47  chris
// Added some comments.
//
// Revision 1.1  1993/05/05  20:04:48  chris
// Initial revision
//
 *------------------------------------------------------------------------*/

#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <iostream.h>
#include <fstream.h>

#include "systemfiles.h"
#include "Sysfiles.h"
#include "Package.h"

static char buff1[256];

/*========================================================================
 * SysFile class
 *========================================================================*/

/*------------------------------------------------------------------------
 * SysFile::Sysfile (const char *)
 *------------------------------------------------------------------------*/

SysFile::SysFile (const char *filename)
: SmartList ()
{
    PathName	file((char *)filename);

    create (file);
    modified = 0;
}

/*------------------------------------------------------------------------
 * SysFile::SysFile (PathName&)
 *------------------------------------------------------------------------*/

SysFile::SysFile (PathName& filename)
: SmartList ()
{
    create (filename);
    modified = 0;
}

/*------------------------------------------------------------------------
 * SysFile::create (PathName&)
 *------------------------------------------------------------------------*/

void SysFile::create (PathName& filename)
{
    char		*cptr;
    Package		*pkg;

    defaults = NULL;

    if (filename.hasPath ())
    {
	fileName = filename;
	if (dirPath.length () == 0)
	    dirPath = (PathName &)filename.directory ();
    }

    /*----
     * Create a default entry for services.
     *----*/

    strcpy (buff1, "service=default ");
    strcat (buff1, "systems=Systems ");
    strcat (buff1, "devices=Devices ");
    strcat (buff1, "dialers=Dialers ");
    pkg = (Package *)makeNewPackage (buff1);
    add (pkg);

    pkg->setDefaults (pkg);
    defaults = pkg;

    /*----
     * Open ~uucp/Sysfiles and create entries for each line found.
     *----*/

    ifstream file ((const char *)fileName);
    if (!file)
    {
        cerr << "SysFile: Unable to open " << (const char *)fileName
	     << " for read." << endl;
        cerr << "  " << strerror (errno) << endl;
	return;
    }
    else
    {
        while (!file.eof ())
        {
            cptr = read_line (&file);
            if (*cptr)
	    {
                pkg = (Package *)makeNewPackage (cptr);
		add (pkg);
	    }
        }

        file.close ();
    }
}

/*------------------------------------------------------------------------
 * ~SysFile
 *------------------------------------------------------------------------*/

SysFile::~SysFile ()
{
}

/*------------------------------------------------------------------------
 * Sysfile::save (const char *)
 *------------------------------------------------------------------------*/

void SysFile::save (const char *fname)
{
    PathName		filename;

    if (fname)
	filename = (char *)fname;
    else
	filename = fileName;

    save (filename);
}

/*------------------------------------------------------------------------
 * SysFile::save (PathName &)
 *------------------------------------------------------------------------*/

void SysFile::save (PathName &filename)
{
    ifstream		*ifile;
    ofstream		*ofile;
    char		*tmpfname, *cptr, *tptr;
    Package		*pkgline, *pkg;
    int			flag;

    if (!wasModified ())
	return;

    tmpfname = tempnam (filename.directory (), "Sysfiles");
    ofile = new ofstream (tmpfname);
    if (ofile->fail ())
    {
	cerr << "Sysfiles: Unable to open file " << filename.name ()
	    << " for output." << endl;
	cerr << "  " << strerror (errno) << endl;
	return;
    }

    ifile = new ifstream (filename.name ());

    /*----
     * Scan through the input file copying comments and blank lines
     * until we come to a legitimate package entry.
     *----*/

    while (!ifile->eof ())
    {
	cptr = get_line (ifile);
	if (ifile->eof () && (*cptr == '\0'))
	    break;

	if ((*cptr == '#') || (*cptr == '\0'))
	    *ofile << cptr << endl;
	else
	{
	    /*----
	     * Create an instance of a Package class using the entry
	     * found so we can use it for comparisons.
	     *----*/

	    pkgline = new Package (cptr);

	    /*----
	     * Search our list to see if we have an entry with the
	     * same package name.  If we do, write it out here.
	     *
	     * Note:  We use the ListEntry flag to indicate those
	     * entries that we\'ve written.  We assume all the
	     * entries are currently set to 0.  We also assume that
	     * if an entry matches the name but the flag is set,
	     * then all entries we have internally have been written.
	     *----*/

	    pkg = (Package *)find (pkgline->theKey ());
	    if (pkg)
	    {
		if (lastFlag ())
		    continue;
		tptr = pkg->line ();
		*ofile << tptr << endl;
		setLastFlag (1);
		while ((pkg = (Package *)next ()) &&
		       (strcmp (pkg->theKey (), pkgline->theKey ()) == 0))
		{
		    tptr = pkg->line ();
		    *ofile << tptr << endl;
		    setLastFlag (1);
		}
	    }

	    delete pkgline;
	}
    }
    ifile->close ();

    /*----
     * Scan through the list of packages we have and write any that
     * have not been written yet.  While we're scanning, clear all
     * the flags.
     *----*/

    for (pkg = (Package *)first (); pkg; pkg = (Package *)next ())
    {
	flag = lastFlag ();
	setLastFlag (0);
	if (flag)
	    continue;
	if (strcmp (pkg->theKey (), "default") == 0)
	    continue;
	tptr = pkg->line ();
	*ofile << tptr << endl;
    }
    ofile->close ();

    rename (tmpfname, filename.name ());

    clearModified ();
}

/*------------------------------------------------------------------------
 * SysFile::makeNewPackage (const char **)
 *------------------------------------------------------------------------*/

LinkData* SysFile::makeNewPackage (const char *line)
{
    Package		*p;

    p = new Package (line, all_files, defaults, &dirPath);

    return p;
}

/*------------------------------------------------------------------------
 * SysFile::wasModified (void)
 *------------------------------------------------------------------------*/

int SysFile::wasModified (void)
{
    Package		*p;

    if (modified == 0)
    {
	for (p = (Package *)first (); p; p = (Package *)next ())
	{
	    if (p->wasModified ())
	    {
		modified = 1;
		break;
	    }
	}
    }

    return modified;
}

/*------------------------------------------------------------------------
 * SysFile::clearModified (void)
 *------------------------------------------------------------------------*/

void SysFile::clearModified (void)
{
    Package		*p;

    for (p = (Package *)first (); p; p = (Package *)next ())
	p->clearModified ();

    modified = 0;
}
