/*------------------------------------------------------------------------
 * $Id: System.c++,v 2.16 1994/02/28 22:30:08 carlson Exp $
 *
 * Methods for System class.
 *
 * Revision History:
 *	$Log: System.c++,v $
// Revision 2.16  1994/02/28  22:30:08  carlson
// Added keeping track of errors with I/O.
// Changed the way that I create the temporary file name because tempnam must
//   make its name in /usr/var/tmp which is frequently on a different file
//   system and won't allow a simple rename.
// Added code to change the mode of the file created to be the same as the
//   original.
//
// Revision 2.15  1994/02/26  01:14:19  carlson
// Removed the 'deleteEntry' method since this is handled by SmartList.
//
// Revision 2.14  1994/02/10  22:55:38  chrisc
// Removed package access functions and made them inline.
//
// Revision 2.13  1994/01/18  00:38:15  carlson
// Removed all references to LinkEntry since it no longer is required by
//   the LinkList class.
//
// Revision 2.12  1993/11/16  18:16:28  carlson
// Include systemfiles.h and we get to remove prototypes for read_line and
//   get_line.
//
// Revision 2.11  1993/11/16  02:29:01  carlson
// Found that the modified flag has to be cleared AFTER initializing class.
//
// Revision 2.10  1993/11/16  01:31:51  carlson
// Added code to indicate we've been modified.
// Added methods for checking if we've been modified and to clear the flag.
//
// Revision 2.9  1993/11/05  20:09:29  carlson
// Changed all references to LinkList to use SmartList.
//
// Revision 2.8  1993/10/06  17:39:17  chris
// Added code to recognize end of line so we don't append a blank line.
//
// Revision 2.7  1993/10/06  02:14:01  chris
// Changed all parameters that were for PathName to be references.
// Added two methods for 'save', which saves a new file.
//
// Revision 2.6  1993/07/24  00:43:48  chris
// Added methods for getting the first and following package entries.
//
// Revision 2.5  1993/07/23  17:14:20  chris
// Added the method findPackage.
//
// Revision 2.4  1993/07/08  00:14:51  chris
// Don't need to pass the file name to LinkData since it is done in 'create'.
// Broke 'create' up so that a separate method is used to read the file.
// Pass const char to makeNewHost.
//
// Revision 2.3  1993/07/07  21:36:04  chris
// Added call to setKey method to set the key value.
//
// Revision 2.2  1993/07/01  01:33:22  chris
// Use specific method to provide file name to LinkData.  Conversion method
//   doesn't work so well.
// Use specific method whereever the file name is needed.
//
// Revision 2.1  1993/06/30  00:24:19  chris
// Initial revision.
//
 *------------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "System.h"
#include "Host.h"
#include "systemfiles.h"

/*========================================================================
 * System class
 *========================================================================*/

/*------------------------------------------------------------------------
 * System::System (Pathname)
 *------------------------------------------------------------------------*/

System::System (PathName& filename) 
: LinkData (), SmartList ()
{
    create (filename);
    readFile (filename);
    modified = 0;
}

/*------------------------------------------------------------------------
 * System::create (PathName&)
 *------------------------------------------------------------------------*/

void System::create (PathName& filename)
{
    if (top)
	clean ();

    setKey ((char *)filename.name ());
}

/*------------------------------------------------------------------------
 * System::readFile (PathName&)
 *------------------------------------------------------------------------*/

void System::readFile (PathName& filename)
{
    LinkData	*host;
    ifstream	*file;
    char	*cptr;

    _errno = 0;
    file = new ifstream (filename.name ());
    if (file->fail ())
    {
	_errno = errno;
	return;
    }

    while (!file->eof ())
    {
	cptr = read_line (file);
	if (*cptr)
	{
	    host = makeNewHost (cptr);
	    add (host);
	}
    }

    delete file;
}

/*------------------------------------------------------------------------
 * System::save (PathName&)
 *------------------------------------------------------------------------*/

void System::save (PathName& filename)
{
    ifstream		*ifile;
    ofstream		*ofile;
    char		*tmpfname, *cptr, *tptr;
    Host		*hostline, *host;
    int			flag;
    struct stat		statBuf;

    _errno = 0;
    if (!wasModified ())
	return;

    tmpfname = new char [(filename.directory ()).length () + 14];
    strcpy (tmpfname, (filename.directory ()).name ());
    strcat (tmpfname, "/SystemsXXXXXX");
    mktemp (tmpfname);
    ofile = new ofstream (tmpfname);
    if (ofile->fail ())
    {
	_errno = errno;
	return;
    }

    ifile = new ifstream (filename.name ());

    /*----
     * Scan through the input file copying comments and blank lines
     * until we come to a legitimate host entry.
     *----*/

    while (!ifile->eof ())
    {
	cptr = get_line (ifile);
	if (ifile->eof () && (*cptr == '\0'))
	    break;

	if ((*cptr == '#') || (*cptr == '\0'))
	    *ofile << cptr << endl;
	else
	{
	    /*----
	     * Create an instance of a Host class using the entry
	     * found so we can use it for comparisons.
	     *----*/

	    hostline = new Host (cptr);

	    /*----
	     * Search our list to see if we have an entry with the
	     * same host name.  If we do, put *all* of our entries
	     * with the same host name out here.
	     *
	     * Note:  We use the ListEntry flag to indicate those
	     * entries that we\'ve written.  We assume all the
	     * entries are currently set to 0.  We also assume that
	     * if an entry matches the name but the flag is set,
	     * then all entries we have internally have been written.
	     *----*/

	    host = (Host *)find (hostline->theKey ());
	    if (host)
	    {
		if (lastFlag ())
		    continue;
		tptr = host->line ();
		*ofile << tptr << endl;
		setLastFlag (1);
		while ((host = (Host *)next ()) &&
		       (strcmp (host->theKey (), hostline->theKey ()) == 0))
		{
		    tptr = host->line ();
		    *ofile << tptr << endl;
		    setLastFlag (1);
		}
	    }

	    delete hostline;
	}
    }
    ifile->close ();

    /*----
     * Scan through the list of hosts we have and write any that
     * have not been written yet.  While we're scanning, clear all
     * the flags.
     *----*/

    for (host = (Host *)first (); host; host = (Host *)next ())
    {
	flag = lastFlag ();
	setLastFlag (0);
	if (flag)
	    continue;
	tptr = host->line ();
	*ofile << tptr << endl;
    }
    ofile->close ();

    /*----
     * Now we're done.
     * Find out the mode settings of the original file.
     * Make sure we have write access to the file.
     * Rename the new file over the old file.
     * Then set the mode settings of the new file back to what they were.
     *----*/

    if (stat (filename.name (), &statBuf))
	_errno = errno;

    if (!_errno)
	if (chmod (filename.name (),
		   statBuf.st_mode | (S_IWUSR | S_IWGRP | S_IWOTH)))
	    _errno = errno;

    if (!_errno)
	if (rename (tmpfname, filename.name ()))
	    _errno = errno;

    if (!_errno)
    {
	if (chmod (filename.name (), statBuf.st_mode))
	    _errno = errno;
	clearModified ();
    }

    delete [] tmpfname;
}

/*------------------------------------------------------------------------
 * System::save (const char*)
 *------------------------------------------------------------------------*/

void System::save (const char *fname)
{
    PathName		filename;

    if (fname)
	filename = (char *)fname;
    else
	filename = (char *)theKey ();

    save (filename);
}

/*------------------------------------------------------------------------
 * System::makeNewHost (const char *)
 *------------------------------------------------------------------------*/

LinkData* System::makeNewHost (const char *c)
{
    Host	*host;

    host = new Host (c);

    return host;
}

/*------------------------------------------------------------------------
 * System::wasModified (void)
 *------------------------------------------------------------------------*/

int System::wasModified (void)
{
    Host		*h;

    if (modified == 0)
    {
	for (h = (Host *)first (); h; h = (Host *)next ())
	{
	    if (h->wasModified ())
	    {
		modified = 1;
		break;
	    }
	}
    }

    return modified;
}

/*------------------------------------------------------------------------
 * System::clearModified (void)
 *------------------------------------------------------------------------*/

void System::clearModified (void)
{
    Host		*h;

    for (h = (Host *)first (); h; h = (Host *)next ())
	h->clearModified ();

    modified = 0;
}
