/*------------------------------------------------------------------------
 * $Id: Terminal.c++,v 2.1 1994/01/19 01:51:34 carlson Exp $
 *
 * Methods for Terminal class.
 *
 * Revision History:
 *	$Log: Terminal.c++,v $
// Revision 2.1  1994/01/19  01:51:34  carlson
// Initial revision.
//
 *------------------------------------------------------------------------*/

#include "Terminal.h"
#include "systemfiles.h"

#include <unistd.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#ifndef NULL
#define NULL 0
#endif

#define ALARM_TIME		10	// Seconds to alarm
#define CHAR_TIME		10	// .1 seconds per character

#define ALARM_OCCURRED		-1
#define WRITE_INCOMPLETE	-2
#define NO_MATCH		-3
#define ILLEGAL_BAUD		-4

static int		alrm_set;

/*-------------------------------------------------------------------------
 * alrm		Function that will be called when an alarm occurs.
 *-------------------------------------------------------------------------*/

void alrm (int)
{
    alrm_set++;
}

/*------------------------------------------------------------------------
 * Terminal::Terminal (const char *)
 *------------------------------------------------------------------------*/

Terminal::Terminal (const char *device)
{
    doInitialization ();
    setDevice (device);
}

/*------------------------------------------------------------------------
 * Terminal::Terminal (void)
 *------------------------------------------------------------------------*/

Terminal::Terminal (void)
{
    doInitialization ();
}

/*------------------------------------------------------------------------
 * Terminal::~Terminal ()
 *------------------------------------------------------------------------*/

Terminal::~Terminal ()
{
    if (tty_fd >= 0)
	::close (tty_fd);
}

/*------------------------------------------------------------------------
 * Terminal::currentState (void)
 *------------------------------------------------------------------------*/

enum Terminal::State Terminal::currentState (void)
{
    if (_error == FAILURE)
	return ERROR;
    if (savErrno)
	return ERROR;

    if (devString)
    {
	if (tty_fd >= 0)
	    return OPENED;
	return CLOSED;
    }

    return NOTSETUP;
}

/*------------------------------------------------------------------------
 * Terminal::errorStrings[]
 *------------------------------------------------------------------------*/

char *Terminal::errorStrings[] =
{
    "Alarm timeout occurred.",
    "Write did not send all characters.",
    "No match occurred.",
    "Illegal baud to be set."
};

/*------------------------------------------------------------------------
 * Terminal::strerror (void)
 *------------------------------------------------------------------------*/

char *Terminal::strerror (void)
{
    return strerror (savErrno);
}

/*------------------------------------------------------------------------
 * Terminal::strerror (int)
 *------------------------------------------------------------------------*/

char *Terminal::strerror (int err)
{
    if (err < 0)
	return errorStrings[-1 - err];
    else if (err > 0)
	return ::strerror (err);
    else
	return "";
}

/*------------------------------------------------------------------------
 * Terminal::setDevice (const char *)
 *------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::setDevice (const char *device)
{
    if (!device)
	return FAILURE;

    devString = NULL;
    savErrno = 0;
    if (access (device, R_OK | W_OK))
    {
	savErrno = errno;
	return FAILURE;
    }

    devString = Copy (device);
    return SUCCESS;
}

/*------------------------------------------------------------------------
 * Terminal::doInitialization (void)
 *------------------------------------------------------------------------*/

void Terminal::doInitialization (void)
{
    ifstream		*ifile;
    char		line[128], *token, *ctmp, *tmp;

    _error = SUCCESS;
    devString = NULL;
    tty_fd = -1;

    /*----
     * Try to get the "official" list of bauds from the include
     * file.  If we can't, go ahead and set what we think they are.
     *----*/

    ifile = new ifstream ("/usr/include/sys/ttydev.h");
    if (ifile->fail ())
    {
	bauds[0]  = 0;     baudCodes[0]  = 000;
	bauds[1]  = 50;	   baudCodes[1]  = 001;
	bauds[2]  = 75;	   baudCodes[2]  = 002;
	bauds[3]  = 110;   baudCodes[3]  = 003;
	bauds[4]  = 134;   baudCodes[4]  = 004;
	bauds[5]  = 150;   baudCodes[5]  = 005;
	bauds[6]  = 200;   baudCodes[6]  = 006;
	bauds[7]  = 300;   baudCodes[7]  = 007;
	bauds[8]  = 600;   baudCodes[8]  = 010;
	bauds[9]  = 1200;  baudCodes[9]  = 011;
	bauds[10] = 1800;  baudCodes[10] = 012;
	bauds[11] = 2400;  baudCodes[11] = 013;
	bauds[12] = 4800;  baudCodes[12] = 014;
	bauds[13] = 9600;  baudCodes[13] = 015;
	bauds[14] = 19200; baudCodes[14] = 016;
	bauds[15] = 38400; baudCodes[15] = 017;
	bauds[16] = -1;    baudCodes[16] = -1;
	baud_ndx = 16;
    }
    else
    {
	for (baud_ndx = 0; !ifile->eof (); )
	{
	    ifile->getline (line, sizeof (line));

	    tmp = line;
	    token = gettoken (&tmp, " \t");

	    if (token && (strcmp (token, "#define") == 0))
	    {
		token = gettoken (&tmp, " \t");
		if (token[0] == 'B')
		{
		    ctmp = ++token;
		    while (isdigit (*ctmp))
			ctmp++;
		    if (!*ctmp)
		    {
			bauds[baud_ndx] = atoi (token);
			token = gettoken (&tmp, " \t");
			baudCodes[baud_ndx++] =
			    (unsigned char)strtoul (token, NULL, 0);
		    }
		}
	    }
	}

	bauds[baud_ndx] = -1;
	baudCodes[baud_ndx] = (unsigned char)-1;

	ifile->close ();
    }

    /*----
     * Set up the alarm to alert us if we've taken too long.
     *----*/

    alrm_set = 0;
    alarm (0);
    sigset (SIGALRM, (void (*)(...))alrm);
}

/*------------------------------------------------------------------------
 * Terminal::setBaud (int)
 *------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::setBaud (int baud)
{
    for (baud_ndx = 0; bauds[baud_ndx] != baud; baud_ndx++)
    {
	if (bauds[baud_ndx] == -1)
	{
	    savErrno = ILLEGAL_BAUD;
	    return FAILURE;
	}
    }

    /*----
     * This step sets the baud rate.
     *----*/

    alarm (ALARM_TIME);

    /*----
     * Try to flush out whatever might be in the terminal.
     * This is not a fatal error so we can continue if it doesn't
     * work.  In an attempt to clear any errors, close the device
     * and re-open it.
     *----*/

    if (ioctl (tty_fd, TCFLSH, 2) < 0)
    {
	savErrno = errno;
	::close (tty_fd);
	if ((tty_fd = reopen_tty (&new_termio)) < 0)
	{
	    alarm (0);
	    return FAILURE;
	}
    }

    if (alrm_set)
    {
	savErrno = ALARM_OCCURRED;
	alarm (0);
	return FAILURE;
    }

    /*----
     * Set the new baud rate.
     * If we can't set it for some reason, this is a fatal error
     * and we should return.  Again, experience shows that it is
     * a good idea to close the device and re-open it.
     *----*/

    new_termio.c_cflag = (new_termio.c_cflag & ~CBAUD) | baudCodes[baud_ndx];

    if (ioctl (tty_fd, TCSETAW, &new_termio) < 0)
    {
	savErrno = errno;
	::close (tty_fd);
	tty_fd = reopen_tty (&new_termio);
	alarm (0);
	return FAILURE;
    }

    if (alrm_set)
    {
	savErrno = ALARM_OCCURRED;
	alarm (0);
	return FAILURE;
    }

    alarm (0);

    return SUCCESS;
}

/*-------------------------------------------------------------------------
 * Terminal::open (void)
 *
 * Routine opens a tty device and sets it up.
 *
 * Description:
 *	Attempts to open devString (should be fully qualified as
 *	in, /dev/ttyf3).  It opens it with O_NDELAY so that we don't
 *	block waiting for DTR.  If the tty won't open, prints an
 *	error and returns -1.  Otherwise, turns off O_NDELAY so that
 *	subsequent read/writes will block and then returns the file
 *	descriptor opened.
 *-------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::open (void)
{
    int			i;
    char		line[128];

    if (tty_fd >= 0)
	::close (tty_fd);

    savErrno = 0;
    if ((tty_fd = ::open (devString, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    {
	savErrno = errno;
	return FAILURE;
    }

	/*----
	 * Ok, we can turn off NDELAY.
	 *----*/

    i = fcntl (tty_fd, F_GETFL);

    if (i == -1) {
	savErrno = errno;
	::close (tty_fd);
	tty_fd = -1;
	return FAILURE;
    }

    i &= ~FNDELAY;

    if (fcntl (tty_fd, F_SETFL, i) == -1) {
	savErrno = errno;
	::close (tty_fd);
	tty_fd = -1;
	return FAILURE;
    }

    return SUCCESS;
}

/*-------------------------------------------------------------------------
 * Terminal::reopen_tty (struct termio *)
 *
 * Reopen a tty and set up appropriately.
 *
 * Synopsis:
 *	tty_fd = reopen_tty (termio_struct);
 *	int tty_fd;
 *	struct termio *termio_struct;
 *
 * Description:
 *	Opens the tty using open_tty and then resets the line control
 *	stuff with what is in termio_struct.  If it can't set the control
 *	stuff, prints and error and returns -1.  Otherwise, it returns
 *	the file descriptor returned by open_tty;
 *-------------------------------------------------------------------------*/

int Terminal::reopen_tty (struct termio *tmp_tty)
{
    int			fd;
    char		line[128];

    savErrno = 0;
    if ((fd = open ()) >= 0) {
	if (ioctl (fd, TCSETAF, tmp_tty) < 0) {
	    savErrno = errno;
	    ::close (fd);
	    fd = -1;
	}
    }

    return fd;
}

/*------------------------------------------------------------------------
 * Terminal::write (char *)
 *
 * Description:
 *	Writes a string to the openned device.
 *------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::write (char *str)
{
    int			n;

    savErrno = 0;
    alarm (ALARM_TIME);
    n = ::write (tty_fd, str, strlen (str));
    alarm (0);

    if (alrm_set)
    {
	savErrno = ALARM_OCCURRED;
	return FAILURE;
    }

    if (n < 0)
    {
	savErrno = errno;
	return FAILURE;
    }
    else if (n != strlen (str))
    {
	savErrno = WRITE_INCOMPLETE;
	return FAILURE;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------
 * Terminal::read (char *, int)
 *
 * Read characters from input.
 *------------------------------------------------------------------------*/

int Terminal::read (char *line, int size)
{
    int 		cnt;

    /*----
     * Loop until we either have some kind of problem or the characters
     * read match what we expect.  Note that we read characters with
     * both a timeout on the read and an alarm.  The alarm timeout is
     * used as a limit for the entire string.
     *
     * The loop is because we will get characters one at a time from
     * the read.
     *----*/

    alrm_set = 0;
    alarm (ALARM_TIME);
    savErrno = 0;
    cnt = ::read (tty_fd, line, size);
    alarm (0);

    /*----
     * First, handle situation where a read error occurred.
     * If alarm occurred, indicate that.
     *
     * Otherwise, terminate the string received.
     *----*/

    if (cnt < 0) {
	savErrno = errno;
	if ((savErrno == EINTR) && alrm_set)
	    savErrno = ALARM_OCCURRED;
    }
    else
	line[cnt] = '\0';

    return cnt;
}

/*------------------------------------------------------------------------
 * Terminal::init (void)
 *
 * This step initializes the device by setting the parameters
 * to what we want.
 *------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::init (void)
{
    /*----
     * Get the current terminal information and save it.
     *----*/

    if (ioctl (tty_fd, TCGETA, &orig_termio) < 0)
    {
	savErrno = errno;
	return FAILURE;
    }

    new_termio = orig_termio;		// Copy so we can modify it.

    /*----
     * Set the terminal up to the basic minimum.
     * Note, we must provide a legal speed otherwise the device
     * gets hung.
     *----*/

    new_termio.c_iflag &= IXON | IXANY | IXOFF;
    new_termio.c_oflag = 0;
    new_termio.c_cflag &= CLOCAL;
    new_termio.c_cflag |= (CS8 | CREAD | HUPCL | SSPEED);
    new_termio.c_lflag = 0;
    new_termio.c_cc[VMIN] = 0;
    new_termio.c_cc[VTIME] = CHAR_TIME;

    if (ioctl (tty_fd, TCSETAF, &new_termio) < 0)
    {
	savErrno = errno;
	ioctl (tty_fd, TCSETA, &orig_termio);
	return FAILURE;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------
 * Terminal::clear (void)
 *------------------------------------------------------------------------*/

enum Terminal::StatusReturn Terminal::clear (void)
{
    alarm (ALARM_TIME);

    ioctl (tty_fd, TCFLSH, 2);

    new_termio.c_cflag = (new_termio.c_cflag & ~CBAUD) | SSPEED;

    if (ioctl (tty_fd, TCSETA, &new_termio) < 0)
    {
	::close (tty_fd);
	alarm (0);
	if ((tty_fd = reopen_tty (&new_termio)) < 0)
	    return FAILURE;
    }
    alarm (0);
    sleep (1);

    return SUCCESS;
}

/*------------------------------------------------------------------------
 * Terminal::uninit (void)
 *------------------------------------------------------------------------*/

void Terminal::uninit (void)
{
    alarm (ALARM_TIME);
    ioctl (tty_fd, TCSETA, &orig_termio);
    alarm (0);
}

/*------------------------------------------------------------------------
 * Terminal::close (void)
 *------------------------------------------------------------------------*/

void Terminal::close (void)
{
    if (tty_fd >= 0)
    {
	alarm (ALARM_TIME);
	::close (tty_fd);
	tty_fd = -1;
	alarm (0);
    }
    _error = SUCCESS;
    savErrno = 0;
}
