/*------------------------------------------------------------------------
 * $Id: systemfiles.c++,v 2.7 1994/02/17 04:49:12 chrisc Exp $
 *
 * Miscellaneous functions used by the system files classes.
 *
 * Revision History:
 *	$Log: systemfiles.c++,v $
// Revision 2.7  1994/02/17  04:49:12  chrisc
// Cleaned up some of the comments.
// Modified get_line to return comment lines immediately, rather than
//   check them for continuation characters.
//
// Revision 2.6  1994/01/14  02:40:33  carlson
// Added routines convert_string, parse_string and unparse_string.
//
// Revision 2.5  1994/01/12  21:56:45  carlson
// Added a function for returning the class name of a widget.  This is
//   mostly used for debugging.
//
// Revision 2.4  1994/01/12  00:47:40  carlson
// Copy now gets a constant char string.
//
// Revision 2.3  1993/11/16  17:56:46  carlson
// Moved 'Copy' function to here since we needed to duplicate it in Hosts,
//   Devices, Dialers and Dialcodes.
//
// Revision 2.2  1993/10/06  00:17:46  chris
// Added the function get_line which is essentially identical to read_line
//   except that it returns all lines and does not pass over comments or
//   empty lines.
//
// Revision 2.1  1993/07/28  16:14:03  chris
// Added function gettoken which emulates strtok except that it allows
//   for nested calls.
//
// Revision 1.1  1993/05/04  19:59:10  chris
// Initial revision
//
 *------------------------------------------------------------------------*/

#include "systemfiles.h"

#include <ctype.h>
#include <string.h>
#include <malloc.h>

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>

static char buff1[1024];

/*------------------------------------------------------------------------
 * read_line
 *	Reads a line from the stream.
 *	If the line begins with '#', ignores it.
 *	If the line is one or less characters long, ignores it.
 *	Removes any control characters at the end (like linefeed).
 *	If the line ends with backslash-<CR>, removes these characters
 *	  and appends the next line.
 *------------------------------------------------------------------------*/

char *read_line (ifstream *file)
{
    char	*cptr;
    int		c;

    buff1[0] = '\0';
    if (!file->eof ())
    {
	do
	{
	    buff1[0] = '\0';
	    file->getline (buff1, sizeof (buff1));
	} while (((buff1[0] == '#') || (strlen (buff1) < 1)) && !file->eof ());

	c = strlen (buff1);
	if (c)
	{
	    cptr = buff1 + c - 1;
	    while (iscntrl (*cptr))
	    {
		*cptr-- = '\0';
		c--;
	    }
	    while (*cptr == 0134)
	    {
		if (file->eof ())
		{
		    *cptr = '\0';
		}
		else
		{
		    file->getline (cptr, sizeof (buff1) - c);
		    c = strlen (buff1);
		    cptr = buff1 + c - 1;
		    while (iscntrl (*cptr))
		    {
			*cptr-- = '\0';
			c--;
		    }
		}
	    }
	}
    }

    return buff1;
}

/*------------------------------------------------------------------------
 * get_line
 *	If line begins with a '#', returns it immediately.
 *	Reads a line from the stream.
 *	Removes any control characters at the end (like linefeed).
 *	If the line ends with backslash-<CR>, removes these characters
 *	  and appends the next line.
 *------------------------------------------------------------------------*/

char *get_line (ifstream *file)
{
    char	*cptr;
    int		c;

    buff1[0] = '\0';
    if (!file->eof ())
    {
	buff1[0] = '\0';
	file->getline (buff1, sizeof (buff1));

	c = strlen (buff1);
	if (c)
	{
	    cptr = buff1 + c - 1;
	    while (iscntrl (*cptr))
	    {
		*cptr-- = '\0';
		c--;
	    }
	    if (buff1[0] != '#')
	    {
		while (*cptr == 0134)
		{
		    if (file->eof ())
		    {
			*cptr = '\0';
		    }
		    else
		    {
			file->getline (cptr, sizeof (buff1) - c);
			c = strlen (buff1);
			cptr = buff1 + c - 1;
			while (iscntrl (*cptr))
			{
			    *cptr-- = '\0';
			    c--;
			}
		    }
		}
	    }
	}
    }

    return buff1;
}

/*------------------------------------------------------------------------
 * gettoken
 *	A replacement for strtok.  This one allows the calling program
 *	to maintain the pointer into the line so that nested calls to
 *	gettoken can be performed.
 *------------------------------------------------------------------------*/

char *gettoken (char **string, char *delims)
{
    char	*return_pointer;

    while (**string && strchr (delims, **string)) // Skip delims at beginning.
	(*string)++;
    return_pointer = *string;
    if (*return_pointer == '\0')
	return NULL;
    while (**string && !strchr (delims, **string)) // Scan to end of token.
	(*string)++;
    if (**string)				// Position point to start
    {						//   of next token.
	**string = '\0';
	(*string)++;
	while (**string && strchr (delims, **string))
	    (*string)++;
    }

    return return_pointer;
}

/*------------------------------------------------------------------------
 * Copy
 *	Similar to strdup except that it uses the 'new' command to
 *	allocate memory and it returns NULL if there is nothing to dup.
 *------------------------------------------------------------------------*/

char *Copy (const char *from)
{
    char	*to = NULL;

    if (from)
    {
	to = new char [strlen (from) + 1];
	strcpy (to, from);
    }
    return to;
}

/*------------------------------------------------------------------------
 * ClassName
 *	Function for returning the class name string from a widget.
 *------------------------------------------------------------------------*/

char *XtClassName (Widget w)
{
    WidgetClass		wclass;

    wclass = w->core.widget_class;
    return wclass->core_class.class_name;
}

/*------------------------------------------------------------------------
 * convert_string (char *)
 *
 * Description:
 *	Converts a string that has control characters in it so that all
 *	characters are printable by converting the control characters
 *	to '^C' form.
 *------------------------------------------------------------------------*/

char *convert_string (char *str)
{
    static char		newstr[256];
    char		*ptr;

    for (ptr = newstr; (*ptr = *str); str++)
    {
	if (iscntrl (*str))
	{
	    *ptr++ = '^';
	    *ptr++ = *str + '@';
	    if (*str == '\n')
		*ptr++ = *str;
	}
	else if (!isascii (*str))
	{
	    sprintf (ptr, "\\%3.3o", (unsigned int)*str);
	    ptr += 4;
	}
	else
	    ptr++;
    }

    return newstr;
}

/*------------------------------------------------------------------------
 * parse_string (char *)
 *
 * Description:
 *	Parses a string using sprintf to handle special escape sequences
 *	recognized by C.
 *------------------------------------------------------------------------*/

static char	spec_chrs[] =
"\a\b\c\d\e\f\g\h\i\j\k\l\m\n\o\p\q\r\s\t\u\v\w\x78\y\z";

char *parse_string (char *str)
{
    static char			line[256];
    char			*cptr;

    for (cptr = line; (*cptr = *str); str++)
    {
	if (*str == '\\')
	{
	    if (isdigit (*(++str)))
	    {
		*cptr = (*str++ - '0') & 7;
		if (isdigit (*str))
		    *cptr = (*cptr << 3) | ((*str++ - '0') & 7);
		if (isdigit (*str))
		    *cptr = (*cptr << 3) | ((*str++ - '0') & 7);
		str--;
	    }
	    else if (isalpha (*str))
	    {
		*cptr = spec_chrs[tolower (*str) - 'a'];
	    }
	}
	cptr++;
    }

    return line;
}

/*------------------------------------------------------------------------
 * unparse_string (char *)
 *
 * Documentation:
 *	Performs the reverse of parse_string.
 *------------------------------------------------------------------------*/

char *unparse_string (char *str)
{
    static char			line[256];
    char			*cptr, *dptr;

    for (cptr = line; (*cptr = *str); str++)
    {
	if (!isprint (*str))
	{
	    *cptr++ = '\\';
	    dptr = strchr (spec_chrs, *str);
	    if (dptr)
	    {
		*cptr = 'a' + dptr - spec_chrs;
	    }
	    else
	    {
		*cptr++ = (*str >> 6) + '0';
		*cptr++ = ((*str >> 3) & 7) + '0';
		*cptr = (*str & 7) + '0';
	    }
	}
	cptr++;
    }

    return line;
}
