/* uLinMath.c */
/*
 * Copyright (c) 1993, 1994, Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include "uLinMath.h"  /* nested "uSmoke.h" */


float uNormalizeVec3(uVec3 vec)
{
    float l = uLengthVec3(vec);
    float sc = 1.0f / l;

    if(l < _ETA) sc = _HUGEVAL;

    uScaleVec3(vec,sc,vec);

    return(l);
}

void uCrossVec3(uVec3 dst, const uVec3 v1, const uVec3 v2)
{
    float t0, t1, t2;

    t0 = v1[1] * v2[2] - v1[2] * v2[1];
    t1 = v1[2] * v2[0] - v1[0] * v2[2];
    t2 = v1[0] * v2[1] - v1[1] * v2[0];
    dst[0] = t0;
    dst[1] = t1;
    dst[2] = t2;
}


/* dst = v * m, dst may be same vector as v */
void uXformVec3(uVec3 dst, const uVec3 v, const uMatrix m)
{
    register float  t0, t1, t2;

    t0 = v[0];
    t1 = v[1];
    t2 = v[2];

    dst[0] = (t0 * m[0][0] +
              t1 * m[1][0] +
              t2 * m[2][0]);

    dst[1] = (t0 * m[0][1] +
              t1 * m[1][1] +
              t2 * m[2][1]);

    dst[2] = (t0 * m[0][2] +
              t1 * m[1][2] +
              t2 * m[2][2]);

}

/* dst = transpose(m) - dst may be same as m */
void uTransposeMat(uMatrix dst, const uMatrix m)
{
    long             i, j;
    float           tmp;

    for (i = 0; i < 4; i++)
    {
        dst[i][i] = m[i][i];
        for (j = 0; j < i; j++)
        {
            tmp = m[i][j];
            dst[i][j] = m[j][i];
            dst[j][i] = tmp;
        }
    }
}

void uSinCos(float d,float *s, float *c)
{
   *s=sinf(d);
   *c=cosf(d);
}
/* GL 5.0 like rotation, about an arbitrary vector */
void uMakeRotMat(uMatrix m, float degrees, float x, float y, float z)
{
    float           s, c, t, f;
    float xyzsq = _SQUARE(x)+_SQUARE(y)+_SQUARE(z);
    f = 1.0f / pfSqrt(xyzsq); /*pfsqrt.s*/

    x *= f;
    y *= f;
    z *= f;

    /*pfSinCos(degrees, &s, &c);*/ /*pfsincos.s*/
    uSinCos(degrees, &s, &c);
    t = 1 - c;

    m[0][0] = t * x * x + c;
    m[0][1] = t * x * y + s * z;
    m[0][2] = t * x * z - s * y;
    m[0][3] = 0.0f;
    m[1][0] = t * y * x - s * z;
    m[1][1] = t * y * y + c;
    m[1][2] = t * y * z + s * x;
    m[1][3] = 0.0f;
    m[2][0] = t * z * x + s * y;
    m[2][1] = t * z * y - s * x;
    m[2][2] = t * z * z + c;
    m[2][3] = 0.0f;
    m[3][0] = 0.0f;
    m[3][1] = 0.0f;
    m[3][2] = 0.0f;
    m[3][3] = 1.0f;
}

/* dst = dst * m */
void uPostMultMat(uMatrix dst, const uMatrix m)
{
    long i;
    register float r0, r1, r2, r3;


    if ((FloatIsNonZero(m[0][3]) |
         FloatIsNonZero(m[1][3]) |
         FloatIsNonZero(m[2][3]) |
         FloatIsNonZero(dst[0][3]) |
         FloatIsNonZero(dst[1][3]) |
         FloatIsNonZero(dst[2][3])) == 0x0 &&
        m[3][3] == 1.0f &&
        dst[3][3] == 1.0f)
    {
        for (i = 0; i < 3; i++)
        {
            r0 = (dst[i][0] * m[0][0] +
                  dst[i][1] * m[1][0] +
                  dst[i][2] * m[2][0]);

            r1 = (dst[i][0] * m[0][1] +
                  dst[i][1] * m[1][1] +
                  dst[i][2] * m[2][1]);

            r2 = (dst[i][0] * m[0][2] +
                  dst[i][1] * m[1][2] +
                  dst[i][2] * m[2][2]);

            dst[i][0] = r0;     /* copy temp into result row */
            dst[i][1] = r1;
            dst[i][2] = r2;
        }
        r0 = (dst[3][0] * m[0][0] +
              dst[3][1] * m[1][0] +
              dst[3][2] * m[2][0] +
              m[3][0]);

        r1 = (dst[3][0] * m[0][1] +
              dst[3][1] * m[1][1] +
              dst[3][2] * m[2][1] +
              m[3][1]);

        r2 = (dst[3][0] * m[0][2] +
              dst[3][1] * m[1][2] +
              dst[3][2] * m[2][2] +
              m[3][2]);

        dst[3][0] = r0; /* copy temp into result row */
        dst[3][1] = r1;
        dst[3][2] = r2;

        dst[0][3] = 0.0f;
        dst[1][3] = 0.0f;
        dst[2][3] = 0.0f;
        dst[3][3] = 1.0f;
        return;
    }
    else
    {
        for (i = 0; i < 4; i++)
        {
            r0 = (dst[i][0] * m[0][0] +
                  dst[i][1] * m[1][0] +
                  dst[i][2] * m[2][0] +
                  dst[i][3] * m[3][0]);

            r1 = (dst[i][0] * m[0][1] +
                  dst[i][1] * m[1][1] +
                  dst[i][2] * m[2][1] +
                  dst[i][3] * m[3][1]);

            r2 = (dst[i][0] * m[0][2] +
                  dst[i][1] * m[1][2] +
                  dst[i][2] * m[2][2] +
                  dst[i][3] * m[3][2]);

            r3 = (dst[i][0] * m[0][3] +
                  dst[i][1] * m[1][3] +
                  dst[i][2] * m[2][3] +
                  dst[i][3] * m[3][3]);

            dst[i][0] = r0;     /* copy temp into result row */
            dst[i][1] = r1;
            dst[i][2] = r2;
            dst[i][3] = r3;
        }
    }
}


/*
 * make rotation that rotates v1 onto v2 through v1 v2 plane
 * with no scaling
 *
 * long contruction from Turkowski, Graphics Gems, p. 522
 * should run through algebra to see if more explicit
 * form is easily expressed
 */
void uMakeRotOntoMat(uMatrix m, const uVec3 v1, const uVec3 v2)
{
    uVec3          v, w;
    uVec3          n1, axis;
    uVec3          wp;
    uMatrix        q, r, zrot90;
    float          fscale, dot;

    /* create normalized copy of v1 */
    uCopyVec3(v, v1);
    fscale = uNormalizeVec3(v);

    /* create normalized copy of v2 */
    uCopyVec3(w, v2);
    fscale = uNormalizeVec3(w) / fscale;

    /* check for colinear vectors */
    dot = uDotVec3(v, w);
    if ((dot >= -1.0f && dot < (-1.0f + 1e-6)) ||
        (dot <= 1.0f && dot > (1.0f - 1e-6)))
    {
        uMakeScaleMat(m, fscale, fscale, fscale);
        return;
    }

    uCrossVec3(axis, v, w);
    uNormalizeVec3(axis);

    /* w, n1 & axis form orthonormal basis */
    uCrossVec3(n1, axis, v);

    q[3][0] = 0.0f;
    q[3][1] = 0.0f;
    q[3][2] = 0.0f;
    q[3][3] = 1.0f;
    q[0][3] = 0.0f;
    q[1][3] = 0.0f;
    q[2][3] = 0.0f;
    uSetMatRowVec3(q, 0, v);    /* v*qt is x axis */
    uSetMatRowVec3(q, 1, n1);
    uSetMatRowVec3(q, 2, axis); /* axis*qt is z axis */

    uTransposeMat(m, q);        /* m is qt */
    uXformVec3(wp, w, m);       /* find image of destination vector under qt */
    uMakeIdentMat(r);
    uSetMatRowVec3(r, 0, wp);   /* rotate x axis onto image of w under qt */

    uMakeRotMat(zrot90, 90.0f, 0.0f, 0.0f, 1.0f);
    uXformVec3(wp, wp, zrot90); /* ortho vec by rotating wp about z axis */
    uSetMatRowVec3(r, 1, wp);   /* complete basis with Z axis */
    /* r is now a valid rotation matrix */

    uPostMultMat(m, r);
    uPostMultMat(m, q);                 /* construct tr(q)*r*q */
}

/* dst = v * m, dst may be same vector as v */
void uXformPt3(uVec3 dst, const uVec3 v, const uMatrix m)
{
    register float  t0, t1, t2, w;

    t0 = v[0];
    t1 = v[1];
    t2 = v[2];

    dst[0] = (t0 * m[0][0] +
              t1 * m[1][0] +
              t2 * m[2][0] +
              m[3][0]);

    dst[1] = (t0 * m[0][1] +
              t1 * m[1][1] +
              t2 * m[2][1] +
              m[3][1]);

    dst[2] = (t0 * m[0][2] +
              t1 * m[1][2] +
              t2 * m[2][2] +
              m[3][2]);

    w = (t0*m[0][3] +
         t1*m[1][3] +
         t2*m[2][3] +
         m[3][3]);
    if (w != 1.0f)
    {
        if (uABSlt(w, _ETA))
            w = _ETA;
        w = 1.0f/w;
        dst[0] *= w;
        dst[1] *= w;
        dst[2] *= w;
    }
}


