/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
// ExportImport.c++

#include "ExportImport.h"
#include "Database.h"
#include "CardInfo.h"
#include "ListView.h"

#include "OkArgs.h"

// Motif includes
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/SelectioB.h>

// Vk includes.
#include <Vk/VkErrorDialog.h>
#include <Vk/VkQuestionDialog.h>
#include <Vk/VkFileSelectionDialog.h>
#include <Vk/VkInterruptDialog.h>
#include <Vk/VkApp.h>




ExportImportDialog::ExportImportDialog() :
  VkGenericDialog( "ExportImportDialog" )
{
  _selectFieldLabel = NULL;
}


ExportImportDialog::~ExportImportDialog()
{
  delete _selectFieldLabel;
  for ( int i=0; i<SetupFields::TOTAL_FIELDS; i++ )
    delete _fieldToggle[ i ];
  delete _delimiter;
}


Widget
ExportImportDialog::createDialog( Widget parent )
{
  OkArgs args( 5 );
  args.set( XmNautoUnmanage, FALSE );
  Widget dialog = XmCreatePromptDialog( parent, _name,
                                        args.list(), args.count() );

  // Remove unneeded children.
  XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_TEXT));
  XtUnmanageChild(XmSelectionBoxGetChild(dialog,XmDIALOG_SELECTION_LABEL));

  Widget form = XmCreateForm( dialog, "form", NULL, 0 );
  XtManageChild( form );

  Widget frame = XmCreateFrame( form, "frame", NULL, 0 );
  XtManageChild( frame );

  _selectFieldLabel = new OkLabel( "selectFieldLabel", frame );
  _selectFieldLabel->setLabelString( _nextSelectFieldLabel );
  XtVaSetValues( *_selectFieldLabel,
    XmNchildType, XmFRAME_TITLE_CHILD,
    NULL );
  _selectFieldLabel->show();

  args.reset();
  args.set( XmNorientation, XmHORIZONTAL );
  args.set( XmNpacking, XmPACK_COLUMN );
  Widget rowColumn = XmCreateRowColumn( frame, "rowColumn",
					args.list(), args.count() );
  
  for ( int i=0; i<SetupFields::TOTAL_FIELDS; i++ ) {
    _fieldToggle[ i ] = new OkToggleButton( 
	theSetupFields->fieldLabel( (SetupFields::Field) i ), rowColumn );
    _fieldToggle[ i ]->setValue( TRUE );
    _fieldToggle[ i ]->show();
  }
 
  // Toggle for Attributes.
  _fieldToggle[ i ] = new OkToggleButton( "Attributes", rowColumn );
  _fieldToggle[ i ]->setValue( TRUE );
  _fieldToggle[ i ]->show();

  XtManageChild( rowColumn );

  _delimiter = new OkLabeledComponent<OkText>( "delimiter", form );
  XtVaSetValues( _delimiter->baseWidget(),
    XmNleftAttachment, XmATTACH_FORM,
    XmNtopAttachment, XmATTACH_WIDGET,
    XmNtopWidget, frame,
    NULL );
  XtVaSetValues( _delimiter->component()->baseWidget(),
    XmNcolumns, 1,
    XmNmaxLength, 1,
    NULL );
  _delimiter->component()->setValue( "," );  // Default delimiter is comma.
  _delimiter->show();

  return dialog;
}


char
ExportImportDialog::delimiter() const
{
  OkStr s = _delimiter->component()->getValue();
  return ( s==""?' ':s[ 0 ] );
}


void
ExportImportDialog::export()
{
  _op = EXPORT;
  setTitle( "Export" );
  setButtonLabels( "Export" );
  if ( _selectFieldLabel )
    _selectFieldLabel->setLabelString( "Select fields to export:" );
  else
    _nextSelectFieldLabel = "Select fields to export:";

  postBlocked();
}


void
ExportImportDialog::import()
{
  _op = IMPORT;
  setTitle( "Import" );
  setButtonLabels( "Import" );
  if ( _selectFieldLabel )
    _selectFieldLabel->setLabelString( "Select fields to import:" );
  else 
    _nextSelectFieldLabel = "Select fields to import:";
  postBlocked();
}

void
ExportImportDialog::ok( Widget, XtPointer )
{
  // Get the file name from user.
  theFileSelectionDialog->setFilterPattern( "*" );
  theFileSelectionDialog->setTitle( (_op==EXPORT?"Export to:":"Import from:") );
  if ( theFileSelectionDialog->postAndWait() == VkDialogManager::CANCEL )
	return;

  if ( _op==EXPORT && OkFile::Exists( theFileSelectionDialog->fileName() ) ) {
    OkStr s = "Overwrite existing file: ";
    s.append( theFileSelectionDialog->fileName() );
    if ( theQuestionDialog->postAndWait( s ) == VkDialogManager::CANCEL )
      return;
  }

  _file.open( theFileSelectionDialog->fileName(), 
              (_op==EXPORT?ios::out:ios::in) );
  if ( ! _file.file() ) {
      OkStr s;
      if ( _op==EXPORT )
        s = "Cannot create file: ";
      else
        s = "Cannot open file: ";
  
      s.append( theFileSelectionDialog->fileName() );
      theErrorDialog->postBlocked( s );
      return;
  }

  // Shutdown dialog, and start exporting/importing.
  unpost();

  // File is ready.

  theApplication->setBusyDialog( theInterruptDialog );

  if ( _op==EXPORT ) {

    theApplication->busy( "Exporting ..." );

    for( register int i=1; i<=theDatabase->totalCards(); i++ ) {

      // Check user interrupt.
      if ( theInterruptDialog->wasInterrupted() ) {
       theApplication->notBusy();
       if ( theQuestionDialog->postAndWait( "Stop exporting ?" )==
						VkDialogManager::OK ) {
		theApplication->busy();
       		break;
	}
       theApplication->busy( "Exporting ..." );
      }

      CardInfo cInfo( theDatabase->_cardDirectory->objAddr( i ) );
      assert( cInfo.ObjectExists() );

      if ( _fieldToggle[ SetupFields::NAME ]->getValue() )
        _file.file() << (char *) cInfo.name() << delimiter();
      if ( _fieldToggle[ SetupFields::PHONE1 ]->getValue() )
        _file.file() << (char *) cInfo.phone1() << delimiter();
      if ( _fieldToggle[ SetupFields::PHONE2 ]->getValue() )
        _file.file() << (char *) cInfo.phone2() << delimiter();
      if ( _fieldToggle[ SetupFields::FAX ]->getValue() )
        _file.file() << (char *) cInfo.fax() << delimiter();
      if ( _fieldToggle[ SetupFields::MOBILE_PHONE ]->getValue() )
        _file.file() << (char *) cInfo.mobilePhone() << delimiter();
      if ( _fieldToggle[ SetupFields::EMAIL ]->getValue() )
        _file.file() << (char *) cInfo.email() << delimiter();

      if ( _fieldToggle[ SetupFields::ADDRESS_LINE1 ]->getValue() )
        _file.file() << (char *) cInfo.addressLine1() << delimiter();
      if ( _fieldToggle[ SetupFields::ADDRESS_LINE2 ]->getValue() )
        _file.file() << (char *) cInfo.addressLine2() << delimiter();
      if ( _fieldToggle[ SetupFields::CITY ]->getValue() )
        _file.file() << (char *) cInfo.city() << delimiter();
      if ( _fieldToggle[ SetupFields::STATE ]->getValue() )
        _file.file() << (char *) cInfo.state() << delimiter();
      if ( _fieldToggle[ SetupFields::ZIP ]->getValue() )
        _file.file() << (char *) cInfo.zip() << delimiter();
      if ( _fieldToggle[ SetupFields::COUNTRY ]->getValue() )
        _file.file() << (char *) cInfo.country() << delimiter();

      if ( _fieldToggle[ SetupFields::TOTAL_FIELDS ]->getValue() ) 
        for ( register int j=0; j<ATTR_COUNT; j++ )
          _file.file() << (int) cInfo.attribute( j ) << delimiter();

      // Remove the last delimiter.
      _file.file().seekg( -1, ios::cur );
      _file.file() << endl;

    }


  } else {
   // IMPORT

   theApplication->busy( "Importing ..." );

   char buf[ BUFSIZ ];

   for (;;) {

     // Check user interrupt.
     if ( theInterruptDialog->wasInterrupted() ) {
       theApplication->notBusy();
       if ( theQuestionDialog->postAndWait( "Stop importing ?" )==
						VkDialogManager::OK ) {
		theApplication->busy();
       		break;
	}
       theApplication->busy( "Importing ..." );
     }


     // For each line in the file:
     _file.file().getline( buf, BUFSIZ );
     if ( _file.file().eof() )
       break;

     OkStr line = buf;
     u_int pos = 0;

     CardInfo cInfo;

     if ( pos<line.length() && _fieldToggle[ SetupFields::NAME ]->getValue() )
       cInfo.setName( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::PHONE1 ]->getValue() )
       cInfo.setPhone1( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::PHONE2 ]->getValue() )
       cInfo.setPhone2( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::FAX ]->getValue() )
       cInfo.setFax( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && 
	  _fieldToggle[ SetupFields::MOBILE_PHONE ]->getValue() )
       cInfo.setMobilePhone( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::EMAIL ]->getValue() )
       cInfo.setEmail( line.ptoken( pos, delimiter() ) );

     if ( pos<line.length() && 
	  _fieldToggle[ SetupFields::ADDRESS_LINE1 ]->getValue() )
       cInfo.setAddressLine1( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && 
	  _fieldToggle[ SetupFields::ADDRESS_LINE2 ]->getValue() )
       cInfo.setAddressLine2( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::CITY ]->getValue() )
       cInfo.setCity( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::STATE ]->getValue() )
       cInfo.setState( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && _fieldToggle[ SetupFields::ZIP ]->getValue() )
       cInfo.setZip( line.ptoken( pos, delimiter() ) );
     if ( pos<line.length() && 
	  _fieldToggle[ SetupFields::COUNTRY ]->getValue() )
       cInfo.setCountry( line.ptoken( pos, delimiter() ) );

     if ( _fieldToggle[ SetupFields::TOTAL_FIELDS ]->getValue() )
       for ( int j=0; j<ATTR_COUNT; j++ ) 
	 if ( pos<line.length() )
           cInfo.setAttribute( j, int( line.ptoken( pos, delimiter() ) ) );

     CardInfo* cInfoPtr = &cInfo;
     OkStr item;
     Database::Load( item, cInfoPtr );
     theListView->addCard( item, &cInfoPtr );

     // Updating each time could slow down import, but gives a nice
     // feedback to user of import progress.
     theListView->updateIndicatorLabel();
    }
  }

  _file.close();

  theApplication->notBusy();
  theApplication->setBusyDialog( NULL );


}

