/*
 * Copyright (C) 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
// ListView.c++

#include "ListView.h"
#include "SetupFields.h"
#include "SetupAttributes.h"
#include "GeneralPrefs.h"
#include "CardList.h"
#include "CardView.h"
#include "CardInfo.h"
#include "Buttons.h"
#include "SrchDialog.h"
#include "Database.h"
#include "PrefFile.h"
#include "ExportImport.h"

#include "OkFile.h"
#include "OkArgs.h"

#include <Vk/VkApp.h>
#include <Vk/VkWarningDialog.h>
#include <Vk/VkQuestionDialog.h>
#include <Vk/VkErrorDialog.h>
#include <Vk/VkFileSelectionDialog.h>
#include <Vk/VkPromptDialog.h>
#include <Vk/VkInfoDialog.h>

#include <Xm/Form.h>
#include <Xm/Separator.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>

#include <X11/keysym.h>

#include <stdlib.h>
#include <assert.h>


ListView* theListView = NULL;


VkMenuDesc
ListView::fileMenuPane[] = {
  { ACTION, "new", &ListView::newFileCB },
  { ACTION, "open", &ListView::openFileCB },
  { ACTION, "save", &ListView::saveCB },
  { SEPARATOR },
  { ACTION, "import", &ListView::importCB },
  { ACTION, "export", &ListView::exportCB },
  { SEPARATOR },
  { ACTION, "printAll", &ListView::printAllCB },
  { SEPARATOR },
  { ACTION, "quit", &ListView::quitCB },
  { END },
};

VkMenuDesc
ListView::cardMenuPane[] = {
  { ACTION, "search", &ListView::searchCB },
  { SEPARATOR },
  { ACTION, "show", &ListView::showCB },
  { ACTION, "delete", &ListView::deleteCB },
  { ACTION, "deleteAll", &ListView::deleteAllCB },
  { ACTION, "new", &ListView::newCB },
//  { SEPARATOR },
//  { ACTION, "print", &ListView::printCB },
  { SEPARATOR },
  { ACTION, "closeAll", &ListView::closeAllCardViewsCB },
  { END },
};

VkMenuDesc
ListView::sortMenuPane[] = {
  { TOGGLE, "ascending", &ListView::sortAscendingCB },
  { TOGGLE, "descending", &ListView::sortDescendingCB },
  { END },
};

VkMenuDesc
ListView::prefsMenuPane[] = {
  { ACTION, "setupFields", &ListView::setupFieldsCB },
  { ACTION, "setupAttributes", &ListView::setupAttributesCB },
  { RADIOSUBMENU, "sort", NULL, ListView::sortMenuPane },
  { SEPARATOR },
  { ACTION, "general", &ListView::generalCB },
  { END },
};

VkMenuDesc
ListView::mainMenuPane[] = {
  { SUBMENU, "file", NULL, ListView::fileMenuPane },
  { SUBMENU, "card", NULL, ListView::cardMenuPane },
  { SUBMENU, "prefs", NULL, ListView::prefsMenuPane },
  { END },
};


ListView::ListView() :
  VkWindow( "listView" ),
  _exportImportDialog( NULL )
{
  assert( !theListView );
  theListView = this;

  // Instantiate Per User General Preferences dialogs.
  _generalPrefDialog = new GeneralPrefDialog;
  _generalPrefDialog->addCallback( VkPrefDialog::prefCallback,
		this, (VkCallbackMethod) &ListView::generalPrefChanged );
  
  _currCardNo = 0;

  // No card viewer yet.
  _activeCardViewer = NO_ACTIVE_VIEWER;
  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    _cardViewTable[ i ] = NULL;


  // Setup UI.

  setMenuBar( mainMenuPane );

  OkArgs args( 6 );
  if ( thePrefFile->listViewWidth() && thePrefFile->listViewHeight() ) {
    args.set( XmNwidth, thePrefFile->listViewWidth() );
    args.set( XmNheight, thePrefFile->listViewHeight() );
  }
  Widget form = XmCreateForm( mainWindowWidget(), "form",
					args.list(), args.count() );
  addView( form );

  _srchText = new OkLabeledComponent<OkText>( "srchText", form );
  assert( _srchText );
  XtVaSetValues( *_srchText,
    XmNtopAttachment, XmATTACH_FORM,
    XmNleftAttachment, XmATTACH_FORM,
    XmNrightAttachment, XmATTACH_FORM,
    NULL );

  _srchText->show();

  args.reset();
  args.set( XmNtopAttachment, XmATTACH_WIDGET );
  args.set( XmNtopWidget, *_srchText );
  args.set( XmNleftAttachment, XmATTACH_FORM );
  args.set( XmNrightAttachment, XmATTACH_FORM );
  Widget separator = XmCreateSeparator( form, "separator", 
					args.list(), args.count() );
  XtManageChild( separator );

  // Create list header.
  args.reset();
  args.set( XmNtopAttachment, XmATTACH_WIDGET );
  args.set( XmNtopWidget, separator );
  args.set( XmNleftAttachment, XmATTACH_FORM );
  args.set( XmNrightAttachment, XmATTACH_FORM );
  _listHeader = new OkLabel( "listHeader", form, args.list(), args.count() );
  _listHeader->setLabelString( "" );
  _listHeader->show();


  // Put up list widget.
  OkStrArray strArray;
  _cardList = new CardList( form, strArray, _srchText->t()->baseWidget() );
  assert( _cardList );
  XtVaSetValues( XtParent( *_cardList ),
    XmNtopAttachment, XmATTACH_WIDGET,
    XmNtopWidget, _listHeader->baseWidget(),
    XmNleftAttachment, XmATTACH_FORM,
    XmNleftOffset, 5,
    XmNrightAttachment, XmATTACH_FORM,
    XmNrightOffset, 5,
    XmNbottomAttachment, XmATTACH_FORM,
    XmNbottomOffset, 45,
    NULL );
  _cardList->show();


  args.reset();
  args.set( XmNtopAttachment, XmATTACH_WIDGET );
  args.set( XmNtopWidget, *_cardList );
  args.set( XmNleftAttachment, XmATTACH_FORM );
  args.set( XmNrightAttachment, XmATTACH_FORM );
  separator = XmCreateSeparator( form, "separator", 
					args.list(), args.count() );
  XtManageChild( separator );


  // Buttons.
  args.reset();
  args.set( XmNorientation, XmHORIZONTAL );
  args.set( XmNpacking, XmPACK_COLUMN );
  args.set( XmNtopAttachment, XmATTACH_WIDGET );
  args.set( XmNtopWidget, separator );
  args.set( XmNrightAttachment, XmATTACH_FORM );
  Widget rowColumn = XmCreateRowColumn( form, "rowColumn",
					args.list(), args.count() );
  XtManageChild( rowColumn );

  _showBtn = new ShowBtn( rowColumn );
  _showBtn->show();
  _deleteBtn = new DeleteBtn( rowColumn );
  _deleteBtn->show();

  _newBtn = new NewBtn( rowColumn );
  _newBtn->show();


  args.reset();
  args.set( XmNbottomAttachment, XmATTACH_FORM );
  args.set( XmNleftAttachment, XmATTACH_FORM );
  _indicatorLabel = XmCreateLabel( form, "indicatorLabel", 
					args.list(), args.count() );
  XtManageChild( _indicatorLabel );


  // Adjust visual to represent app state.
  setMenuAndBtnSensitivity();
  updateIndicatorLabel();

  // Direct all keyboard inputs to text widget.
  XtSetKeyboardFocus( mainWindowWidget(), srchText()->baseWidget() );

  // Add checking for ESC key being hit.
  XtAddEventHandler( srchText()->baseWidget(), KeyPressMask, FALSE, 
    &ListView::checkEscKeyPressCB, XtPointer( this ) );

}


ListView::~ListView()
{
}


void
ListView::checkEscKeyPress( XKeyEvent* e, Boolean* b )
{
  if ( XLookupKeysym( e, 0 )==XK_Escape ) {
    // ESC key pressed, deselect list item and clear the text.
    _cardList->deselectAllItems();
    srchText()->setValue( "" );
    *b = FALSE;

    updateIndicatorLabel();
    setMenuAndBtnSensitivity();
  }
}

void
ListView::setCardSensitivity()
{
  if ( _cardList->selectedItemCount() ) {
    XtSetSensitive( *_showBtn, TRUE );
    XtSetSensitive( *_deleteBtn, TRUE );
    ((VkMenu *)menu())->activateItem( "show" );
    ((VkMenu *)menu())->activateItem( "delete" );
//    ((VkMenu *)menu())->activateItem( "print" );
  } else {
    XtSetSensitive( *_showBtn, FALSE );
    XtSetSensitive( *_deleteBtn, FALSE );
    ((VkMenu *)menu())->deactivateItem( "show" );
    ((VkMenu *)menu())->deactivateItem( "delete" );
//    ((VkMenu *)menu())->deactivateItem( "print" );
  }


}

/////////////////////////////////////////////////////////////////
// Show card using the given Card Info object.
void
ListView::showCard( unsigned pos, CardInfo* cInfo )
{
  theApplication->busy();

  // Make this card the current one.
  _currCardNo = pos;

  // Visually show it in the list.
  _cardList->selectItem( _currCardNo );

  // Is this card already shown by one of the viewers ?
  // If so, just activate it, and return.
  if ( ! cInfo )
   for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    if ( _cardViewTable[ i ] && _cardViewTable[ i ]->cardPos()==pos ) {
      // Make this one active, and raise it.

      setActiveCardViewer( _cardViewTable[ i ] );
      _cardViewTable[ i ]->open();
      _cardViewTable[ i ]->show();

      theApplication->notBusy();
      return;
    }

  // Card is not currently being shown.  Dig it.

  // We may need to load info from db.
  if ( ! cInfo ) {
    // Get the card info from db.
    // Card View will take care deletion of cInfo.
    cInfo = theDatabase->cardInfo( pos );
  }

  // We better have the card info at this point.
  assert( cInfo );


  // Do we need to create a viewer ?
  if ( noActiveCardViewer() ) {
    // No active viewer.  Make one.
    for ( int i=0; i<MAX_CARD_VIEWS; i++ )
      if ( _cardViewTable[ i ] == NULL )
	break;  // Found one.
    if ( i>=MAX_CARD_VIEWS ) {
      // No more viewer available.
      theApplication->notBusy();
      theWarningDialog->postBlocked( 
		"Sorry no more card viewer available.", this );
      return;
    }

    _cardViewTable[ i ] = new CardView( i );
    assert( _cardViewTable[ i ] );
    setAttrPrefChangedView( _cardViewTable[ i ] );

    // Make this new one the active viewer.
    setActiveCardViewer( _cardViewTable[ i ] );
  }

  // Populate the fields with new card info.
  _cardViewTable[ activeCardViewer() ]->setCardInfo( cInfo );
  // Set the position AFTER setting the card.
  // The previous position is needed to update info on List View.
  _cardViewTable[ activeCardViewer() ]->setCardPos( pos );

  theApplication->notBusy();
}



void
ListView::setActiveCardViewer( CardView* cardViewer )
{
  assert( cardViewer );

  setNoActiveCardViewer();
  cardViewer->makeActive( TRUE );

  _activeCardViewer = cardViewer->viewerNo();
}


void
ListView::setNoActiveCardViewer()
{
  if ( ! noActiveCardViewer() )
    // Deactivate current active viewer.
    _cardViewTable[ activeCardViewer() ]->makeActive( FALSE );

  _activeCardViewer = NO_ACTIVE_VIEWER;
}


void
ListView::closeAllCardViews()
{
  theApplication->busy();

  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    if ( _cardViewTable[ i ] ) _cardViewTable[ i ]->hide();

  theApplication->notBusy();
}


void
ListView::updateIndicatorLabel( int selectedListItemPos )
{
  OkStr s;

  if ( selectedListItemPos ) {
    s = OkStr( selectedListItemPos, "No:  %d/" );
    assert( theDatabase );
    s.append( OkStr( theDatabase->totalCards(), "%d" ) );
  } else {
    s =( theDatabase ? OkStr( theDatabase->totalCards(), "Total:  %d" ) : 
		       OkStr( "Total:  0" ) );
  }

  XtVaSetValues( _indicatorLabel, 
    XmNlabelString, XmString( OkCompoundString( s ) ), 
    NULL );

  XmUpdateDisplay( _indicatorLabel );
}


void
ListView::searchCB( Widget, XtPointer, XtPointer )
{
  theSrchDialog->post( NULL, theListView );
}


void
ListView::setAttrPrefChanged( VkComponent *, void*, void* reason )
{
  if ( (VkDialogManager::VkDialogReason)reason == VkDialogManager::CANCEL )
    return;

  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    if ( _cardViewTable[ i ] ) 
      setAttrPrefChangedView( _cardViewTable[ i ] );

  theSetupAttributes->save();
}

void
ListView::setAttrPrefChangedView( CardView* view )
{
  assert( view );

  theApplication->busy();
  Boolean noAttribute=TRUE;

  // Must do this before setting other attibute properties.
  view->attributes()->setRadioStyle( theSetupAttributes->radioStyle() );

  for ( int j=0; j<ATTR_COUNT; j++ ) {
        if ( theSetupAttributes->showAttr( j ) ) {
	  noAttribute = FALSE;
          view->attributes()->showAttr( j );
          view->attributes()->
			setAttrLabel( j, theSetupAttributes->attrLabel( j ) );
        } else {
          view->attributes()->hideAttr( j );
        }
       
  }

  if ( noAttribute )
	XtSetSensitive( 
	  XtParent( view->attributes()->baseWidget() ), FALSE );
  else
	XtSetSensitive( 
	  XtParent( view->attributes()->baseWidget() ), TRUE );



  XmUpdateDisplay( *view );
  theApplication->notBusy();
}


void
ListView::generalPrefChanged( VkComponent *, void*, void* reason )
{
  if ( ((VkDialogManager::VkDialogReason) reason) == VkDialogManager::CANCEL )
    return;  // Do nothing.

  theApplication->busy();

  // Tell Pref about the changes.
  thePrefFile->setDataFilePath( _generalPrefDialog->defRolodexLoc() );

  thePrefFile->setShowFaceImg( _generalPrefDialog->showFaceImg() );
  thePrefFile->setFaceImgLoc( _generalPrefDialog->faceImgLoc() );
  for ( int i=0; i<MAX_CARD_VIEWS; i++ ) {
    if ( ! _cardViewTable[ i ] )
      continue;

    if ( _generalPrefDialog->showFaceImg() )
      _cardViewTable[ i ]->showFaceImg();
    else
      _cardViewTable[ i ]->hideFaceImg();
  }

  thePrefFile->
	setAlwaysUseIntrDialog( _generalPrefDialog->alwaysUseIntrDialog() );

  theApplication->notBusy();
}


void
ListView::sort( SortOrder sortOrder )
{
  // There will be 2 callbacks coming in.  
  // The first from previous radio button, the second from the new
  // radio button.  Ignores the first one.
  static Boolean firstOne=TRUE;
  if ( firstOne ) {
    firstOne = FALSE;
    return; // This is the first one, ignore it.
  } else {
    firstOne = TRUE;
  }

  theApplication->busy( "Re-sorting ..." );

  closeAllCardViews();
  cardList()->deleteAllItems();

  theDatabase->sort( sortOrder );

  theApplication->notBusy();

  setMenuAndBtnSensitivity();
  updateIndicatorLabel();
}


void
ListView::updateCardListFromCardInfo( unsigned pos, const CardInfo* cInfo )
{
  if ( ! theDatabase )
    return;

  theApplication->busy();

  OkStr item;
  theDatabase->Load( item, cInfo );

  int selectedItemPos = _cardList->getSelectedItem();
  _cardList->replaceItem( item, pos );
  if ( pos==selectedItemPos )
    _cardList->selectItem( pos, FALSE );

  setMenuAndBtnSensitivity();

  theApplication->notBusy();
}

void
ListView::updateCardListFromDb( const char* busyMsg )
{
  if ( ! theDatabase )
    return;

  if ( totalCards()>500 )
    theApplication->busy( (busyMsg?busyMsg:"Updating List View ...") );
  else
    theApplication->busy();

  closeAllCardViews();

  OkStr listHeader;
  OkStrArray listItems;
  theDatabase->load( listHeader, listItems );
  _listHeader->setLabelString( listHeader );
  _cardList->deleteAllItems();
  _cardList->addItems( listItems );
  setMenuAndBtnSensitivity();
  updateIndicatorLabel();

  theApplication->notBusy();
}


VkDialogManager::VkDialogReason
ListView::openFileDialog( OkStr& fileName, const char* dialogTitle )
{
  theFileSelectionDialog->setTitle( dialogTitle );

  theFileSelectionDialog->setDirectory( thePrefFile->dataFilePath() );

  theFileSelectionDialog->setFilterPattern( "*.DAT" );

  if ( OkStr( dialogTitle ) == "New Contact Book" ) {
    OkStr defFileName = thePrefFile->dataFilePath() | OkStr( "MYCONTACTS.DAT" );
    theFileSelectionDialog->setSelection( defFileName );
  }

  VkDialogManager::VkDialogReason r = theFileSelectionDialog->postAndWait();
  if ( r == VkDialogManager::OK ) {

    // Remove trailing ".DAT" from file name.
    OkStr tmp = theFileSelectionDialog->fileName();
    if ( tmp.tail( 4 ) == ".DAT" )
      fileName = tmp.head( tmp.length()-4 );
    else
      fileName = tmp;
  }

  return r;
}



void
ListView::save()
{
  theApplication->busy();

  // Save view changes.  This will also save the database.
  saveAllCardViewChanges();

  // Save user pref file.
  thePrefFile->save();

  theApplication->notBusy();
}



void
ListView::openFile( const char* f, Boolean newRolodex )
{
  OkStr fileName = (f?f:"");

  // Is file name supplied ?
  if ( fileName == "" )
    // No.  Get it from user.
    if ( openFileDialog( fileName, ( newRolodex ? "New Contact Book" : 
						  "Open Contact Book" ) ) == 
	  VkDialogManager::CANCEL )
	return; // Change mind.

  if ( ! newRolodex ) {
    // Unless it's new.  The rolodex must exist.
    if ( ! OkFile::Exists( fileName|OkStr( ".DAT" ) ) ||
	 ! OkFile::Exists( fileName|OkStr( ".NDX" ) ) ) {
      OkStr s = "Cannot open: " | fileName;
      theInfoDialog->postBlocked( s );
      return;
    }
  }


  // Shutdown current one, and recreate everything.
  theApplication->busy();

  closeAllCardViews();

  new Database( fileName );
  assert( theDatabase );

  // If this is a new database, flush won't work.
  // So, the trick is to close it and re-open immediately.
  if ( newRolodex ) {
    new Database( fileName );
    assert( theDatabase );
  }


  showRolodexNameInTitle( fileName );

  // Set sort radio btns to show the correct sort order.
  if ( theDatabase->sortOrder()==ASCENDING ) {
    ((VkMenuToggle *) ((VkMenu *)menu())->findNamedItem( "ascending" ))->
						  setVisualState( TRUE );
    ((VkMenuToggle *) ((VkMenu *)menu())->findNamedItem( "descending" ))->
						  setVisualState( FALSE );
  } else {
    ((VkMenuToggle *) ((VkMenu *)menu())->findNamedItem( "descending" ))->
						  setVisualState( TRUE );
    ((VkMenuToggle *) ((VkMenu *)menu())->findNamedItem( "ascending" ))->
						  setVisualState( FALSE );
  }

  // Connect SetupAttribute dialog with setAttrPrefChanged.
  theSetupAttributes->addCallback( VkPrefDialog::prefCallback,
		this, (VkCallbackMethod) &ListView::setAttrPrefChanged );

  setAttrPrefChanged( NULL, NULL, (void *) VkDialogManager::OK );

  updateCardListFromDb( "Loading Contact Book ..." );

  setMenuAndBtnSensitivity();

  theApplication->notBusy();

  if ( newRolodex ) theInfoDialog->
    postBlocked( "Use Setup Fields and Attributes to customize new rolodex." );

}


void
ListView::saveAllCardViewChanges()
{
  theApplication->busy();

  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
      if ( _cardViewTable[ i ] ) 
	_cardViewTable[ i ]->save();

  theApplication->notBusy();
}


void
ListView::setMenuAndBtnSensitivity() 
{
  if ( theDatabase ) {

    ((VkMenu *)menu())->activateItem( "save" );
    ((VkMenu *)menu())->activateItem( "import" );
    ((VkMenu *)menu())->activateItem( "export" );
    ((VkMenu *)menu())->activateItem( "printAll" );

    ((VkMenu *)menu())->activateItem( "card" );
    ((VkMenu *)menu())->activateItem( "prefs" );

    XtSetSensitive( *_newBtn, TRUE );
    XtSetSensitive( _srchText->component()->baseWidget(), TRUE );

  } else {

    ((VkMenu *)menu())->deactivateItem( "save" );
    ((VkMenu *)menu())->deactivateItem( "import" );
    ((VkMenu *)menu())->deactivateItem( "export" );
    ((VkMenu *)menu())->deactivateItem( "printAll" );

    ((VkMenu *)menu())->deactivateItem( "card" );
    ((VkMenu *)menu())->deactivateItem( "prefs" );

    XtSetSensitive( *_newBtn, FALSE );
    XtSetSensitive( _srchText->component()->baseWidget(), FALSE );
  }

  setCardSensitivity();

}


Boolean
ListView::deleteCard( unsigned pos, CardInfo* cInfo, VkSimpleWindow* win )
{
  assert( cInfo );

  OkStr s = "Delete '" | cInfo->name() | "' ?";
  theQuestionDialog->setButtonLabels( "Yes", "No" );
  if ( theQuestionDialog->postAndWait( s, (win?win:this) ) == 
						VkDialogManager::CANCEL )  
	return FALSE;	
 
  theApplication->busy();

  theDatabase->deleteCard( pos, cInfo );
  cardList()->deleteItem( pos );
  adjustViewerCardPos( pos, DELETE );
  setMenuAndBtnSensitivity();
  updateIndicatorLabel();

  theApplication->notBusy();

  return TRUE;
}

Boolean
ListView::deleteCard( unsigned pos )
{
  // Is the card with this id being shown ?
  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    if ( _cardViewTable[ i ] && 
	 _cardViewTable[ i ]->cardPos()==pos ) {
      return _cardViewTable[ i ]->deleteCard();
    }

  CardInfo *cInfo = theDatabase->cardInfo( pos );
  assert( cInfo );
  Boolean b = deleteCard( pos, cInfo );
  delete cInfo;

  return b;
}


void
ListView::deleteAll()
{
  if ( theQuestionDialog->postAndWait( "Delete all cards ?" ) ==
						VkDialogManager::CANCEL )
       return;

  theApplication->busy();

  // Destroy all card viewers.
  setNoActiveCardViewer();
  closeAllCardViews();

  theDatabase->deleteAllCards();
  cardList()->deleteAllItems();
  updateIndicatorLabel();
  setMenuAndBtnSensitivity();

  theApplication->notBusy();
}

unsigned
ListView::addCard( VkSimpleWindow* win )
{
  OkStr name = "Enter the person's name for the new card:";
  thePromptDialog->setTitle( "Add Card" );
  thePromptDialog->setButtonLabels( "Add" );
  if ( thePromptDialog->postAndWait( name, (win?win:this) ) == 
						VkDialogManager::CANCEL )
	return 0;
			
				

  name = ( thePromptDialog->text() ? thePromptDialog->text() : "" );
  CardInfo* cInfo=NULL;
  unsigned pos = addCard( name, &cInfo );

  // Bring up the CardView for this new interactively added card.
  showCard( pos, cInfo );

  // The name is not in the database yet.  So, needs to tell the
  // card viewer that it IS dirty, even though it's just shown.
  _cardViewTable[ activeCardViewer() ]->setDirty();

  return pos;
}

unsigned
ListView::addCard( const char* name, CardInfo** cInfo )
{
  theApplication->busy();

  unsigned pos = _cardList->add( name );
  _cardList->lastVisibleItem( pos );

  if ( *cInfo==NULL ) {
    *cInfo = theDatabase->addCard( name, pos );
    assert( *cInfo );
  } else {
    theDatabase->addCard( *cInfo, pos );
  }

  adjustViewerCardPos( pos, ADD );

  theApplication->notBusy();

  return pos;
}


void
ListView::adjustViewerCardPos( unsigned pos, AdjustReason reason )
{
  for ( int i=0; i<MAX_CARD_VIEWS; i++ )
    if ( _cardViewTable[ i ] && _cardViewTable[ i ]->cardPos()>pos )
      if ( reason==ADD )
	_cardViewTable[ i ]->setCardPos( _cardViewTable[ i ]->cardPos()+1 );
      else
	_cardViewTable[ i ]->setCardPos( _cardViewTable[ i ]->cardPos()-1 );
}


void
ListView::showRolodexNameInTitle( const char* fileName )
{
    OkStr newTitle = theApplication->className();
    newTitle.append( " - " );

    char *tmps = OkFile::PathBaseName( fileName );
    newTitle.append( tmps );
    delete tmps;

    setTitle( newTitle );
}



Boolean
ListView::okToQuit()
{
  theApplication->busy();

  // Ask active viewer if ok to quit.
  if ( ! noActiveCardViewer() &&
       ! _cardViewTable[ activeCardViewer() ]->okToQuit() )
    return FALSE;

  // Shutdown all viewers.
  for ( int i=0; i<MAX_CARD_VIEWS; i++ ) {
    delete _cardViewTable[ i ];
  }

  // Shutdown the database.
  delete theDatabase;

  // Save window information.
  Dimension w,h;
  XtVaGetValues( viewWidget(),
    XmNwidth, &w,
    XmNheight, &h,
    NULL );

  thePrefFile->setListViewWidth( w );
  thePrefFile->setListViewHeight( h );

  theApplication->notBusy();

  return TRUE;
}


void
ListView::export()
{
  if ( ! _exportImportDialog )
    _exportImportDialog = new ExportImportDialog;

  _exportImportDialog->export(); 
}


void
ListView::import()
{
  if ( ! _exportImportDialog )
    _exportImportDialog = new ExportImportDialog;

  _exportImportDialog->import(); 
}


void
ListView::setupFields()
{
  assert( theSetupFields );
  if ( theSetupFields->postAndWait() == VkDialogManager::CANCEL ) 
    return;

  // Destroy all viewers, and reload list view.
  setNoActiveCardViewer();
  updateCardListFromDb();

  static Boolean firstTime = TRUE;
  if ( firstTime ) {
    theInfoDialog->postModal( "All card viewers will be closed.\nResize the main window to see more or less fields." );
    firstTime = FALSE;
  }

  // Need to make new export import dialog next time to reflect the change
  // in field labels.
  delete _exportImportDialog;
  _exportImportDialog = NULL;
}

void
ListView::setupAttributes()
{
  assert( theSetupAttributes );
  theSetupAttributes->postBlocked();
}

int
ListView::totalCards() const
{
  return theDatabase->totalCards();
}
